#!/bin/bash

usage()
{
    echo
    echo "Options:"
    echo "  -c, --cli      List included CLI tools"
    echo "  -h, --help     Display help"
    echo "  -q, --quiet    Answer all prompts with default choice"
    echo
    echo "Variables:"
    echo "  CLITOOLS=0  (Skip CLI tools installation)"
    echo "  CLITOOLS=1  (Install CLI tools to default location)"
    echo "  CLITOOLS=/path/to/install/cli/tools"

    if [ "${DRVR_TYPE}" == "gpu" ]; then
        echo "  NVIDIA_DRIVER=/path/to/driver/source"
    fi

    if [ ${EXP_IS_PRESENT} -eq 1 ]; then
        echo "  ATTOVIEW=0  (Skip ATTOview support installation)"
        echo "  ATTOVIEW=1  (Install ATTOview support)"
    fi

    echo
}

## Display Header ##

display_header()
{
    printf "${RED}\n"
    printf " ______________________________ \n"
    printf "|     _    _____ _____   ___   |\n"
    printf "|    / \  |_   _|_   _| / _ \  |\n"
    printf "|   / _ \   | |   | |  | | | | |\n"
    printf "|  / ___ \  | |   | |  | |_| | |\n"
    printf "| /_/   \_\ |_|   |_|   \___/  |\n"
    printf "|______________________________|\n"
    printf "${ESC}\n"
    echo "ATTO ${PRODUCT_DISP} Installer ${PRODUCT_VER}"
    echo
    echo "Display help with -h"
    echo
}


## Initialize Variables ##

init_variables()
{
    # The following are patched by the build system to be real values
    DRIVER_NAME="esas5hba"
    DRVR_TYPE="std"
    PRODUCT_ID="esashba5"
    PRODUCT_DISP="ExpressSAS 5"
    PRODUCT_VER="1.10.1f1"

    BATCH=0
    IS_RHEL_DERIVED=0
    IS_SUSE_DERIVED=0
    IS_DEBIAN_DERIVED=0
    IS_SLES=0
    IS_OPENSUSE=0
    IS_RHEL=0
    IS_CENTOS=0
    IS_ROCKY=0
    IS_FEDORA=0
    IS_DEBIAN=0
    IS_UBUNTU=0
    MODULE_ROOT="/lib/modules/$(uname -r)"
    OP_CLI=0
    PROMPT=1

    # See req_software for tool variables

    # Architecture

    OS_ARCH=`uname -m | sed -e s/i.86/i386/`

    if [ "${OS_ARCH}" == "x86_64" ]; then
        GEN_ARCH="x64"
    fi

    # Exporter (derived from architecture)

    EXPORTER_INSTALLER="${DIR}/scripts/${GEN_ARCH}/${PRODUCT_ID}_exporter.run"

    if [ -f "${EXPORTER_INSTALLER}" ]; then
        EXP_IS_PRESENT=1
        EXP_INST=${ATTOVIEW:-2}
    else
        EXP_IS_PRESENT=0
        EXP_INST=0
    fi

    # CLI tools (derived from architecture)

    CLI_DIR="${DIR}/cli/${GEN_ARCH}"
    CLI_TOOLS="$(find "${CLI_DIR}" -type f -exec basename -a {} + | sort | xargs)"
    CLI_INSTALLDIR="${CLITOOLS:-/usr/local/sbin}"

    if [[ -v CLITOOLS ]]; then
        case "${CLITOOLS}" in
            "0")
                CLI_INSTALLDIR=
                CLI_INST=0
                ;;

            "1")
                CLI_INSTALLDIR="/usr/local/sbin"
                CLI_INST=1
                ;;

            *)
                CLI_INST=1
                ;;
        esac
    else
        CLI_INST=2
    fi

    # Other derived variables

    UNINSTALLER="/usr/local/sbin/${DRIVER_NAME}-uninstall"

    if [ "${DRVR_TYPE}" == "gpu" ]; then
        LEGACY_INSTALL=1
    else
        LEGACY_INSTALL=0
    fi

    # Process arguments

    for arg; do
        case "${arg}" in
            --cli | -c)
                OP_CLI=1
                ;;

            --help | -h)
                usage
                exit 0
                ;;

            --quiet | -q)
                PROMPT=0
                BATCH=1
                ;;
            *)
                usage
                exit 1
        esac
    done

    # Pretty Colors

    RED="\033[1;31m"
    GRE="\033[1;32m"
    BLU="\033[1;34m"
    ORA="\033[1;38;5;202m"
    YEL="\033[1;33m"
    CYN="\033[1;36m"
    BLD="\033[1m"
    ESC="\033[0m"
}


## Helper Functions ##

prompt_user()
{
    printf "${CYN}$1${ESC}" && read RSP
    eval "$2=${RSP}"
}

cmd_not_found()
{
    if [[ "$1" == "dkms" ]]; then
        echo
        echo "If you would like to install DKMS, please"
        echo "run the following command:"
        echo
        if [[ ${IS_RHEL_DERIVED} == 1 ]]; then
            echo " # sudo ${PACMAN} install epel-release $1"
        else
            echo " # sudo ${PACMAN} install $1"
        fi
    else
        printf "\n${RED}ERROR:${ESC} Unable to find '$1'.\n"
        echo
        echo "       You must have '$1' to install this driver."
        echo
        if [[ -n ${PACMAN} ]]; then
            echo "       You should be able to correct this problem by"
            echo "       running the following command:"
            echo
            echo "       # sudo ${PACMAN} install $1"
        else
            echo "       Please refer to your system documentation for"
            echo "       instructions to install $1"
        fi
    fi
    echo
    exit 1
}

exit_failure()
{
    printf "\n${RED}ERROR:${ESC} The installer encountered an error.\n"
    printf "You can try to manually install the driver by running\n\n"
    printf " # make clean install\n\n"
    printf "from the src directory.\n\n"

    exit 1
}

clean_exit()
{
    trap - 0
    exit 0
}

clean_failure()
{
    trap - 0
    exit 1
}

## Perform system checks before attempting install ##

system_checks()
{
    # Verify that the script is being executed as root

    if [[ "${EUID}" -ne 0 ]]; then
        printf "\n${RED}This installer must be run as superuser!${ESC}\n\n"
        exit 1
    fi

    # Verify that the path doesn't contain whitespace

    if [[ "${DIR}" == *" "* ]]; then
        printf "\n${RED}ERROR:${ESC} The installer is being run in a path that contains spaces\n"
        printf "which is not supported by GNU make.  Please copy the installation\n"
        printf "package to a path without spaces.\n\n"
        exit 1
    fi

    # Check for supported architecture

    if [[ "${OS_ARCH}" != "x86_64" ]]; then
        printf "\n${RED}ERROR:${ESC} Unsupported architecture ${OS_ARCH}\n\n"
        exit 1
    fi
}

## Determine the running distribution ##

find_distro()
{
    if [ -e /etc/os-release ]; then

        source /etc/os-release
        DISTRO=${ID}
        OS_VERSION="${VERSION_ID}"

        KERNEL_VERSION=$(uname -r)
        MAIN_KERN_VER=$(uname -r | awk -F. '{print $1}')
        MAJOR_REL_VER=$(uname -r | awk -F. '{print $2}')
        MINOR_REL_VER=$(uname -r | awk -F. '{print $3}' | awk -F- '{print $1}')
        PATCH_LVL_VER=$(uname -r | awk -F. '{print $3}' | awk -F- '{print $2}')

        # Assume the distro is supported until it isn't...

        local IS_SUPPORTED=1

        # We can expand this to limit install to certain versions as well if we want

        case "${DISTRO}" in
        sles | sled)
            IS_SUSE_DERIVED=1
            IS_SLES=1
            PACMAN="zypper"
            ;;
        opensuse-leap | opensuse)
            IS_SUSE_DERIVED=1
            IS_OPENSUSE=1
            PACMAN="zypper"
            ;;
        rhel)
            IS_RHEL_DERIVED=1
            IS_RHEL=1
            PACMAN="yum"
            ;;
        centos)
            IS_RHEL_DERIVED=1
            IS_CENTOS=1
            PACMAN="yum"
            ;;
        rocky)
            IS_RHEL_DERIVED=1
            IS_ROCKY=1
            PACMAN="yum"
            ;;
        fedora)
            IS_RHEL_DERIVED=1
            IS_FEDORA=1
            PACMAN="yum"
            ;;
        ubuntu)
            IS_DEBIAN_DERIVED=1
            IS_UBUNTU=1
            PACMAN="apt-get"
            ;;
        debian)
            IS_DEBIAN_DERIVED=1
            IS_DEBIAN=1
            PACMAN="apt-get"

            # Debian is not officially supported or tested, so still warn the user
            IS_SUPPORTED=0
            ;;
        *)
            IS_SUPPORTED=0
            ;;
        esac
    else
        IS_SUPPORTED=0
    fi

    if [[ ${IS_SUPPORTED} -eq 0 ]]; then
        printf "${ORA}WARNING:${ESC}"
        printf "This is an unsupported distribution!\n"
        echo
        echo "         Installation and software may work, but is not officially maintained"
        echo "         for your currently running distribution."
        echo
        if [[ ${PROMPT} -eq 1 ]]; then
            while true; do
                prompt_user "Would you like to install anyway? [Y/n]: " yn
                case ${yn} in
                    '' | [Yy]* )
                        break
                        ;;
                    [Nn]* )
                        printf "\n${RED}Installation cancelled${ESC}\n\n"
                        exit 1
                        ;;
                    * )
                        echo "Please respond [Y/n]"
                        ;;
                esac
            done
        fi
    fi
}


## Verify that the user has software required for installation ##

req_software()
{
    if [[ ${IS_DEBIAN} -eq 1 ]]; then
       if [[ "${PATH}" != *"/sbin"* ]]; then
           export PATH="${PATH}:/sbin"
       fi
    fi

    set +e
    DKMS=`type -P dkms`
    MAKE=`type -P make`
    GCC=`type -P gcc`
    RMMOD=`type -P rmmod`
    MODPROBE=`type -P modprobe`
    MOKUTIL=`type -P mokutil`
    set -e

    if [[ -z ${MAKE} ]]; then
        cmd_not_found make
    elif [[ -z ${GCC} ]]; then
        cmd_not_found gcc
    fi
}


## Determine the method for installation ##

install_method()
{
    # A recent kernel change to Modpost has rendered DKMS broken for
    # proprietary modules in kernel 5.8+

    if [[ ${MAIN_KERN_VER} -eq 5 && ${MAJOR_REL_VER} -ge 8 ]] || [[ ${MAIN_KERN_VER} -ge 6 ]]; then
        LEGACY_INSTALL=1
    elif [[ ${PROMPT} -eq 1 && -n ${DKMS} ]]; then
        while true; do
            echo "DKMS can be useful on systems that receive frequent kernel updates."
            prompt_user "Would you like to install with DKMS? [N/y]: " yn
            case ${yn} in
                [Yy]* )
                    LEGACY_INSTALL=0
                    break
                    ;;
                '' | [Nn]* )
                    LEGACY_INSTALL=1
                    break
                    ;;
                * )
                    echo "Please respond [Y/n]"
                    ;;
            esac
        done
    else
        LEGACY_INSTALL=1
    fi
}


## Verify that kernel headers are installed ##

verify_headers()
{
    KDIR="/lib/modules/$(uname -r)/build"

    if [ ${IS_DEBIAN_DERIVED} -eq 1 ]; then
        OS_HEADERS="linux-headers-$(uname -r)"
    else
        OS_HEADERS="kernel-devel"
    fi

    if [ ! -d ${KDIR} ]; then
        printf "\n${RED}ERROR:${ESC} Unable to find the kernel header directory.\n"
        echo
        echo "       The kernel headers were expected in"
        echo "       ${KDIR}"
        if [[ -n ${PACMAN} ]]; then
            echo
            echo "       You should be able to correct this problem by"
            echo "       running the following command:"
            echo
            echo "       # sudo ${PACMAN} install ${OS_HEADERS}"
            echo
            echo "       Note: If the kernel headers installed by ${PACMAN}"
            echo "       don't match the version of your installation, you may"
            echo "       need to execute the following command and reboot:"
            echo
            echo "       # sudo ${PACMAN} update"
        else
            echo
            echo "       Please refer to your system documentation for"
            echo "       instructions on obtaining the kernel headers."
            echo "       This may be called the \"kernel-devel\" package."
            echo
            echo "       If the kernel headers are installed in a different"
            echo "       location, you can build the driver manually from the src/"
            echo "       directory by specifying the correct kernel header directory"
            echo "       on the make command line with the KDIR variable, eg:"
            echo
            echo "       # make clean install KDIR=/usr/src/kernel"
        fi
        echo
        clean_exit
    fi
}


## Prep for Install ##

prep_install()
{
    # Set distro version in atto_distro.h if needed.

    if [[ "${IS_SUSE_DERIVED}" == "1" ]]; then
        if [[ "${OS_VERSION}" =~ "15" ]]; then
            SLE_VERSION="#define SLE_VERSION_CODE SLE_VERSION(15,${OS_VERSION#15.},0)"
        fi

        if ! grep -q "${SLE_VERSION}" "${DIR}/src/atto_distro.h"; then
            LINE="\/\* \*\*\*SUSE Install script overrides\*\*\* \*\/"
            sed -i "s/${LINE}/${LINE}\n\n${SLE_VERSION}/" "${DIR}/src/atto_distro.h"
        fi
    fi

    if [[ "${IS_DEBIAN}" == "1" ]]; then
        DEB_VERSION="#define DEB_VERSION(a,b,c) KERNEL_VERSION(a,b,c)\n#define DEB_VERSION_CODE DEB_VERSION(${OS_VERSION},0,0)"

        if ! grep -q "${DEB_VERSION}" "${DIR}/src/atto_distro.h"; then
            LINE="\/\* \*\*\*Debian Install script overrides\*\*\* \*\/"
            sed -i "s/${LINE}/${LINE}\n\n${DEB_VERSION}/" "${DIR}/src/atto_distro.h"
        fi
    fi
}


## Remove and delete any conflicting drivers ##

blacklist_module()
{
    # NOTE: DKMS will automatically find and backup previous installs of kernel
    #       modules of the same name as what we will be installing. No need to
    #       manually locate and uninstall.

    case ${PRODUCT_ID} in
    esashba6 )
        CONFLICT_NAME=("mpi3mr")
        ;;
    esashba5 )
        CONFLICT_NAME=("mpt3sas")
        ;;
    esashba4 )
        CONFLICT_NAME=("pm80xx")
        ;;
    celerity5 )
        CONFLICT_NAME=("lpfc")
        ;;
    esac

    # Blacklist drivers that conflict with nvidia.ko

    if [ "${DRVR_TYPE}" == "gpu" ]; then
        CONFLICT_NAME+=("nouveau" "rivafb" "nvidiafb" "rivatv")
    fi

    # Blacklist the conflicting drivers

    if [[ -n "${CONFLICT_NAME}" ]]; then
        echo -e "# This file was automatically generated by the ATTO Linux installer\n" \
             > /etc/modprobe.d/${DRIVER_NAME}-blacklist.conf

        for conflict in ${CONFLICT_NAME[@]}; do
            if [ "${conflict}" == "nouveau" ]; then
                echo "options nouveau modeset=0" >> /etc/modprobe.d/${DRIVER_NAME}-blacklist.conf
                echo "blacklist ${conflict}" >> /etc/modprobe.d/${DRIVER_NAME}-blacklist.conf
            elif [ "${conflict}" == "mpt3sas" ] || [ "${conflict}" == "lpfc" ]; then
                echo "softdep ${conflict} pre: ${DRIVER_NAME}" >> /etc/modprobe.d/${DRIVER_NAME}-blacklist.conf
            else
                echo "blacklist ${conflict}" >> /etc/modprobe.d/${DRIVER_NAME}-blacklist.conf
            fi
        done

        # Workaround for HPE iLO virtual display driver

        if [[ $(lsmod | grep mgag200) ]]; then
            echo "install ${DRIVER_NAME} /sbin/modprobe mgag200; /sbin/modprobe --ignore-install ${DRIVER_NAME}" \
                >> /etc/modprobe.d/${DRIVER_NAME}-blacklist.conf
        fi
    fi
}


## Direct2Gpu Requirements and Setup ##

direct2gpu_init()
{
    # Check for NVIDIA driver

    if ! lspci | grep -i 'controller: nvidia' > /dev/null; then
        printf "No NVIDIA GPUs were found.\n"
        printf "The ${DRIVER_NAME}-gpu driver requires an NVIDIA GPU to be installed in the system.\n\n"
        clean_failure
    fi

    # Acquire the path if not already specified

    if [ -z "${NVIDIA_DRIVER}" ]; then
        if [ ${PROMPT} -eq 1 ]; then
            printf "Building the Direct2GPU-enabled ${DRIVER_NAME} driver requires specifying the\n"
            printf "location of NVIDIA's driver header files and module symbols.\n"
            printf "Refer to the readme-direct2gpu.txt for setup instructions.\n\n"

            printf "${CYN}Please enter the full path to the NVIDIA driver directory:${ESC}\n"
            printf "Example: /usr/src/Linux-x86_64-<version>/kernel\n"
            read -e RESPONSE
        else
            # Gotta have the path for automated installs

            printf "${RED}ERROR:${ESC} NVIDIA_DRIVER must be specified.\n"
            clean_failure
        fi
    fi

    # Validate the directory

    if [ ! -d "${RESPONSE}" ]; then
        printf "${RED}ERROR:${ESC} ${RESPONSE} is not a valid directory\n"
        clean_failure
    fi

    # Check for the symbols

    if [ ! -e "${RESPONSE}/Module.symvers" ]; then
        printf "${RED}ERROR:${ESC} Module.symvers was not found in ${RESPONSE}\n"
        clean_failure
    fi

    NVIDIA_DRIVER=${RESPONSE}
}


## Confirm system modification ##

confirm_install()
{
    if [[ ${PROMPT} -eq 1 ]]; then
        while true; do
            # Check if Secure Boot is enabled

            if [[ -z ${MOKUTIL} ]]; then
                printf "\n${ORA}WARNING:${ESC} Unable to determine Secure Boot state."
                printf "\nPlease install mokutil to properly detect Secure Boot state."
            elif ${MOKUTIL} --sb-state | grep -q enabled > /dev/null 2>&1; then
                printf "\n${ORA}WARNING:${ESC} Secure boot is enabled on this system."
                printf "\nIt is recommended to disable secure boot in your system's BIOS."
                printf "\nIf Secure Boot is necessary, please follow directions in the README.\n"

                printf "\n${ORA}WARNING:${ESC} ATTO drivers will not load until properly signed or Secure Boot is disabled.\n\n"
            fi

            prompt_user "Your system is about to be modified.  Continue? [Y/n]: " yn
            case ${yn} in
                '' | [Yy]* )
                    break
                    ;;
                [Nn]* )
                    printf "\n${RED}Installation cancelled${ESC}\n\n"
                    clean_exit
                    ;;
                * )
                    echo "Please respond [Y/n]"
                    ;;
            esac
        done

        echo
    fi
}


## Uninstall drivers ##

list_dkms_versions()
{
    if [ -n "${DKMS}" ]; then 
        # Extract just <module>/<version> by stripping off the first comma
        # and everything after

        ${DKMS} status | grep "${DRIVER_NAME}" | sed 's/,.*//'
    fi
}

uninstall_drivers()
{
    local SUCCESS=YES

    # Uninstall DKMS-installed drivers
    
    local DRIVERS=($(list_dkms_versions | xargs))

    for driver in ${DRIVERS[@]}; do
        local PARTS=($(echo ${driver} | tr "/" " "))
        local VER=${PARTS[1]}
        echo
        echo "Uninstalling ${DRIVER_NAME} ${VER} via DKMS..."
        echo

        ${DKMS} remove "${driver}" --all || SUCCESS=

        # Delete the source

        if [ -d "/usr/src/${DRIVER_NAME}-${VER}" ]; then
            echo
            echo "Removing /usr/src/${DRIVER_NAME}-${VER}"

            rm -rf "/usr/src/${DRIVER_NAME}-${VER}" || SUCCESS=
        fi
    done

    # Uninstall kernel modules

    local DRIVER_INSTALL=`find ${MODULE_ROOT} -name "${DRIVER_NAME}*"`

    if [ -n "${DRIVER_INSTALL}" ]; then
        echo
        echo "Removing installed ${DRIVER_NAME} module..."

        for driver in ${DRIVER_INSTALL}; do
            rm -rf ${driver} || SUCCESS=
        done
    fi

    # Display results

    if [ -z "${SUCCESS}" ]; then
        printf "${RED}Failed to fully remove the existing ${DRIVER_NAME} driver from the system${ESC}\n"
        printf "Continuing with installation anyway.\n"
    fi
}


## Generate the dkms.conf file ##

gen_dkms_conf()
{
    if [ ${IS_UBUNTU} -eq 1 ]; then
        # Ubuntu requires us to install to updates in order to override built-in modules

        local MOD_DEST="updates"
    else
        local MOD_DEST="extra"
    fi
            
# Do not indent!

cat > "${DIR}/src/dkms.conf" <<EOF
PACKAGE_NAME="${DRIVER_NAME}"
PACKAGE_VERSION="${PRODUCT_VER}"
BUILT_MODULE_NAME[0]="${DRIVER_NAME}"
BUILT_MODULE_NAME[1]="${DRIVER_NAME}_cfg"
DEST_MODULE_LOCATION[0]="/${MOD_DEST}"
DEST_MODULE_LOCATION[1]="/${MOD_DEST}"
AUTOINSTALL="yes"
EOF
}


## Generate initramfs if needed ##

gen_initramfs()
{
    if [[ -n "${CONFLICT_NAME}" ]]; then
        if [ ${IS_DEBIAN_DERIVED} -eq 1 ]; then
            update-initramfs -u
        elif [ ${IS_SUSE_DERIVED} -eq 1 ]; then
            dracut -f -q
        elif [ ${IS_RHEL_DERIVED} -eq 1 ]; then
            dracut -f
        fi
    fi
}


## Install using DKMS ##

dkms_install()
{
    printf "\nInstalling the ${DRIVER_NAME} driver via DKMS...\n\n"

    # Move source to the dkms install location
    cp -r src /usr/src/${DRIVER_NAME}-${PRODUCT_VER}

    # The --force is to avoid version conflicts since our
    # version notation is non standard.
    ${DKMS} install ${DRIVER_NAME}/${PRODUCT_VER} --directive ARCH=${OS_ARCH} --force

    gen_initramfs

    set +e
    printf "\n${GRE}Installed the ${DRIVER_NAME} driver via DKMS:${ESC}\n\n"
    ${DKMS} status ${DRIVER_NAME}/${PRODUCT_VER}
    echo
    set -e
}


legacy_install()
{
    cd ${DIR}/src

    printf "Building and installing the ${DRIVER_NAME} driver, please wait...\n"

    if [ "${DRVR_TYPE}" == "gpu" ]; then
        OPTIONS="NVIDIA_INCDIR=${NVIDIA_DRIVER} NVIDIA_SYMS=${NVIDIA_DRIVER}/Module.symvers"
    fi

    set +e
    if ${MAKE} clean install ${OPTIONS} >make.log 2>make.err; then
        set -e
        gen_initramfs
        printf "\n${GRE}Installed the ${DRIVER_NAME} driver successfully${ESC}\n\n"
    else
        echo
        printf "\n${RED}An error occurred while building the driver${ESC}\n"
        echo
        echo "---make output---"
        cat make.log
        echo
        echo "---make errors---"
        cat make.err
        echo "---------"
        echo
        clean_failure
    fi
}


## Run uninstaller script ##

run_uninstaller()
{
    if [ -x "${UNINSTALLER}" ]; then
        INSTALLER=1 BATCH=${BATCH} "${UNINSTALLER}"
        echo
    fi
}


## Load Module ##

load_module()
{
    printf "Loading the ${DRIVER_NAME} driver...\n\n"

    if [[ -n ${CONFLICT_NAME} ]]; then
        if lsmod | grep ${CONFLICT_NAME} > /dev/null \
        && ! ${RMMOD} ${CONFLICT_NAME}; then
            printf "${RED}ERROR:${ESC} Unable to unload conflicting builtin module '${CONFLICT_NAME}'.\n"
            printf "${RED}A system reboot may be required to complete the driver installation${ESC}\n\n"
        fi
    fi

    if [ "${DRVR_TYPE}" == "gpu" ]; then
        if ! lsmod | grep -w nvidia > /dev/null; then
            if ! modprobe nvidia; then
                printf "${RED}ERROR:${ESC} Could not insert module.\n"
                printf "${RED}A system reboot may be required to complete the driver installation${ESC}\n\n"
            fi
        fi
    fi

    if lsmod | grep -w ${DRIVER_NAME} > /dev/null; then
        if ! ${RMMOD} ${DRIVER_NAME}; then
            printf "${RED}ERROR:${ESC} Module could not be removed.\n"
            printf "${RED}A system reboot may be required to complete the driver installation${ESC}\n\n"
        elif lsmod | grep ${DRIVER_NAME}_cfg > /dev/null && ! ${RMMOD} ${DRIVER_NAME}_cfg; then
            printf "${RED}ERROR:${ESC} Module could not be removed.\n"
            printf "${RED}A system reboot may be required to complete the driver installation${ESC}\n\n"
        elif ! ${MODPROBE} ${DRIVER_NAME}; then
            printf "${RED}ERROR:${ESC} Could not insert module.\n"
            printf "${RED}A system reboot may be required to complete the driver installation${ESC}\n\n"
        else
            printf "${GRE}Loaded ${DRIVER_NAME} successfully${ESC}\n\n"
        fi
    else
        if lsmod | grep -w ${DRIVER_NAME}_cfg > /dev/null; then
            if ! ${RMMOD} ${DRIVER_NAME}_cfg; then
                printf "${RED}ERROR:${ESC} An outdated ${DRIVER_NAME}_cfg Module existis on the system,\n"
                printf "but could not be removed.\n"
                printf "${RED}A system reboot may be required to complete the driver installation${ESC}\n\n"
                break
            fi
        fi

        if ! ${MODPROBE} ${DRIVER_NAME}; then
            printf "${RED}ERROR:${ESC} Could not insert module.\n"
            printf "${RED}A system reboot may be required to complete the driver installation${ESC}\n\n"
        fi
    fi
}


## List CLI tools

list_cli_tools()
{
    for tool in ${CLI_TOOLS}; do
        local NVERSION=`${CLI_DIR}/${tool} -v | awk '{print $NF}' | sed '/./,$!d'`
        local DETECTED=""

        # Check if this tool is already installed

        local INSTALLED=`type -P ${tool}`

        if [[ -n ${INSTALLED} ]]; then
            # If installed, get the version

            local IVERSION=`${tool} -v | awk '{print $NF}' | sed '/./,$!d'`
            local IDIR=`dirname ${INSTALLED}`

            DETECTED=" (detected ${IVERSION} in ${IDIR})"
        fi

        echo "${tool} ${NVERSION}${DETECTED}"
    done

    echo
}


## Install CLI Tools ##

install_cli_tools()
{
    if [ ${CLI_INST} -eq 0 ]; then
        return
    elif [ ${CLI_INST} -eq 2 -a ${PROMPT} -eq 1 ]; then
        # Prompt for a different path

        echo "The following CLI tools will be installed:"

        list_cli_tools

        while true; do
            printf "${CYN}Enter the CLI tool installation path (delete path to skip installation):${ESC}\n"
            read -e -i "${CLI_INSTALLDIR}" RESPONSE

            if [ -z "${RESPONSE}" ]; then
                CLI_INSTALLDIR=""
                return
            fi

            # Validate the directory

            if [[ -d ${RESPONSE} ]]; then
                CLI_INSTALLDIR=${RESPONSE}
                break
            else
                prompt_user "${RESPONSE} was not found.  Create? [Y/n]: " yn

                case ${yn} in
                    '' | [Yy]* )
                        mkdir -p ${RESPONSE}
                        CLI_INSTALLDIR=${RESPONSE}
                        break
                        ;;
                esac
            fi
        done

        echo
    fi

    # Copy tools to install directory

    for TOOL in ${CLI_TOOLS}; do
        install -D -m755 ${CLI_DIR}/${TOOL} ${CLI_INSTALLDIR}/${TOOL}
        echo "Installed ${CLI_INSTALLDIR}/${TOOL}"
    done

    printf "\n${GRE}Installed CLI tools successfully${ESC}\n\n"
}


## Install Metric Exporter ##

install_exporter()
{
    if [ ${EXP_INST} -eq 0 ]; then
        return
    elif [ ${EXP_INST} -eq 2 -a ${PROMPT} -eq 1 ]; then
        while true; do
            prompt_user "Install ATTOview support (data exporter)? [Y/n]: " yn
            case ${yn} in
                '' | [Yy]* )
                    break
                    ;;
                [Nn]* )
                    return
                    ;;
                * )
                    echo "Please respond [Y/n]"
                    ;;
            esac
        done
    fi

    printf "Installing ATTOview support...\n\n"

    "${EXPORTER_INSTALLER}"

    echo
}


## Install service ##

install_service()
{
    mkdir -p /usr/local/libexec
    local SVC_NAME="${DRIVER_NAME}d"
    local DST="/usr/local/libexec/${SVC_NAME}"
    local SRC="${DIR}/service/${GEN_ARCH}"

    if [ ! -d "${SRC}" ]; then
        return
    fi

    # Install the service files

    if [ -d "${DST}" ]; then
        rm -rf "${DST}"
    fi

    cp -r ${SRC} ${DST}
    chmod 755 ${DST}
    find ${DST} -type d -exec chmod 755 {} +
    find ${DST} -name "*.sh" -exec chmod 755 {} +
    chmod 755 ${DST}/${SVC_NAME}

    # Write the service config

    cat >/etc/systemd/system/${SVC_NAME}.service <<EOF
[Unit]
Description = ${PRODUCT_DISP} service
After = network-online.target

[Service]
Type = simple
Restart = on-failure
WorkingDirectory = ${DST}
ExecStart = ${DST}/${SVC_NAME}

[Install]
WantedBy = multi-user.target
EOF

    # Fire up the service

    systemctl enable ${SVC_NAME}
    systemctl daemon-reload
    systemctl start ${SVC_NAME}

    printf "\n${GRE}Installed service successfully${ESC}\n\n"
}


## Install Uninstaller Script ##

install_uninstaller()
{
    local SCRIPT="${DIR}/scripts/uninstall.sh"
    local INCVARS="${DIR}/scripts/uninstall.inc"

    cat > "${INCVARS}" <<EOF
CLI_INSTALLDIR="${CLI_INSTALLDIR}"
CLI_TOOLS="${CLI_TOOLS}"
EOF

    install -D -m755 "${SCRIPT}" "${UNINSTALLER}"
    install -D -m644 "${INCVARS}" "${UNINSTALLER}.inc"

    printf "Uninstaller located at ${BLU}${UNINSTALLER}${ESC}\n\n"
}


## Main ##

main()
{
    set -eo pipefail

    DIR=$(dirname "$(readlink -f "$0")")

    init_variables "$@"
    display_header

    if [ ${OP_CLI} -eq 1 ]; then
        list_cli_tools
        exit 0
    fi

    system_checks
    find_distro
    verify_headers
    req_software
    install_method

    if [ "${DRVR_TYPE}" == "gpu" ]; then
        direct2gpu_init
    fi

    # One last chance to bail
    confirm_install

    # System modification starts now
    trap 'exit_failure' 0

    # Ideally, we want any installed package to uninstall itself
    run_uninstaller

    # Uninstall drivers ourselves just in case
    uninstall_drivers

    prep_install
    blacklist_module

    if [[ ${LEGACY_INSTALL} -eq 0 ]]; then
        gen_dkms_conf
        dkms_install
    else
        legacy_install
    fi

    load_module

    install_service
    install_exporter
    install_cli_tools
    install_uninstaller

    printf "${BLU}${PRODUCT_DISP} installation complete${ESC}\n\n"
    clean_exit
}

main "$@"
