#!/bin/bash

list_dkms_versions()
{
    if [ -n "${DKMS}" ]; then 
        # Extract just <module>/<version> by stripping off the first comma
        # and everything after

        ${DKMS} status | grep "${DRIVER_NAME}" | sed 's/,.*//'
    fi
}


driver_installed()
{
    # Check DKMS

    if [ -n "$(list_dkms_versions)" ]; then
        return
    fi

    # Check installed kernel modules

    local FILES="$(find ${MODULE_ROOT} -name "${DRIVER_NAME}*" -type f)"

    if [ -n "${FILES}" ]; then
        return
    fi

    # Nothing found

    return 1
}


uninstall_drivers()
{
    local SUCCESS=YES

    # Uninstall DKMS-installed drivers
    
    local DRIVERS=($(list_dkms_versions | xargs))

    for driver in ${DRIVERS[@]}; do
        local PARTS=($(echo ${driver} | tr "/" " "))
        local VER=${PARTS[1]}
        echo "Uninstalling ${DRIVER_NAME} ${VER} via DKMS..."
        echo

        ${DKMS} remove "${driver}" --all || SUCCESS=

        # Delete the source

        if [ -d "/usr/src/${DRIVER_NAME}-${VER}" ]; then
            echo "Removing /usr/src/${DRIVER_NAME}-${VER}"
            echo
            rm -rf "/usr/src/${DRIVER_NAME}-${VER}" || SUCCESS=
        fi
    done

    # Uninstall kernel modules

    local DRIVER_INSTALL=`find ${MODULE_ROOT} -name "${DRIVER_NAME}*"`

    if [ -n "${DRIVER_INSTALL}" ]; then
        echo "Removing installed ${DRIVER_NAME} module..."
        echo

        for driver in ${DRIVER_INSTALL}; do
            rm -rf ${driver} || SUCCESS=
        done
    fi

    # Display results

    if [ "${SUCCESS}" ]; then
        printf "${GRE}Removed the ${DRIVER_NAME} driver successfully${ESC}\n"

        # No need to do this if we're installing

        if [ -z "${INSTALLER}" ]; then
            printf "${GRE}Reboot your system for changes to take effect${ESC}\n"
        fi

        printf "\n"
    else
        printf "${RED}Failed to fully remove the ${DRIVER_NAME} driver from the system${ESC}\n\n"
    fi
}


remove_blacklist()
{
    # Remove any blacklist files we created during install

    local FILE="/etc/modprobe.d/${DRIVER_NAME}-blacklist.conf"

    if [ -e "${FILE}" ]; then
        rm -f "${FILE}"
    fi
}


## Uninstall CLI Tools ##

uninstall_cli_tools() {
    local REMOVED=0

    for TOOL in ${CLI_TOOLS}; do
        if [ -e "${CLI_INSTALLDIR}/${TOOL}" ]; then
            echo "Removing ${CLI_INSTALLDIR}/${TOOL}..."
            rm -f "${CLI_INSTALLDIR}/${TOOL}"
            REMOVED=1
        fi
    done

    if [ ${REMOVED} -eq 1 ]; then
        printf "\n${GRE}Removed CLI tools successfully${ESC}\n\n"
    else
        printf "\nNo CLI tools removed\n\n"
    fi
}


## Uninstall Metric Exporter ##

uninstall_exporter()
{
    local EXPORTER="/usr/local/libexec/${PRODUCT_ID}_exporter/uninstall.sh"

    if [ -f "${EXPORTER}" ]; then
        printf "Removing ATTOview support...\n\n"

        "${EXPORTER}"
    fi
}


## Uninstall service ##

uninstall_service()
{
    local SVC_NAME="${DRIVER_NAME}d"
    local DST="/usr/local/libexec/${SVC_NAME}"

    if [ -d "${DST}" ]; then
        # Stop and delete service config

        systemctl stop ${SVC_NAME} || true
        systemctl disable ${SVC_NAME} || true

        rm -f /etc/systemd/system/${SVC_NAME}.service

        systemctl daemon-reload

        # Delete service files

        rm -rf ${DST}

        printf "\n${GRE}Removed service successfully${ESC}\n\n"
    fi
}


## Determine the running distribution ##

find_distro()
{
    if [ -e /etc/os-release ]; then
        source /etc/os-release
        DISTRO=${ID}

        case "${DISTRO}" in
        sles | sled | opensuse-leap | opensuse)
            IS_SUSE_DERIVED=1
            ;;
        rhel | centos | rocky | fedora)
            IS_RHEL_DERIVED=1
            ;;
        ubuntu)
            IS_DEBIAN_DERIVED=1
            ;;
        esac
    fi
}

## Omit driver from initrd ##

omit_driver()
{
    if [ ${IS_RHEL_DERIVED} -eq 1 ]; then
        printf "Removing driver from initrd...\n\n"

        dracut -f --omit-drivers ${DRIVER_NAME}
    fi
}


## Rebuild Initrd ##

gen_initramfs()
{
    if [ ${IS_DEBIAN_DERIVED} -eq 1 ]; then
        update-initramfs -u
    elif [ ${IS_SUSE_DERIVED} -eq 1 ]; then
        dracut -f -q
    fi
}


## Remove Uninstaller ##

remove_uninstaller()
{
    for file in "${UNINSTALLER}" "${UNINSTALL_INC}"; do
        if [ -e ${file} ]; then
            rm -f "${file}"
        fi
    done
}


## Initialize Variables ##

# The following are patched by the build system to be real values
DRIVER_NAME="esas5hba"
DRVR_TYPE="std"
PRODUCT_ID="esashba5"
PRODUCT_DISP="ExpressSAS 5"
PRODUCT_VER="1.10.1f1"

UNINSTALLER="$(readlink -f "$0")"
UNINSTALL_INC="${UNINSTALLER%.sh}.inc"

if [ ! -f "${UNINSTALL_INC}" ]; then
    printf "${UNINSTALL_INC} was not found and is required to uninstall\n\n"
    exit 1
fi

source "${UNINSTALL_INC}"

# Historically, 'BATCH' is passed by the installer to the uninstaller.  We're
# not using it anymore, but keep it around in case we ever need a prompt.

BATCH="${BATCH:-0}"

MODULE_ROOT="/lib/modules/$(uname -r)"
IS_SUSE_DERIVED=0
IS_RHEL_DERIVED=0
IS_DEBIAN_DERIVED=0

## Pretty Colors ##

RED="\033[1;31m"
GRE="\033[1;32m"
BLU="\033[1;34m"
ORA="\033[1;38;5;202m"
YEL="\033[1;33m"
CYN="\033[1;36m"
BLD="\033[1m"
ESC="\033[0m"

echo

## Verify that the script is being executed as root ##

if [[ "${EUID}" -ne 0 ]]; then
    printf "\n${RED}This uninstaller must be run as superuser!${ESC}\n\n"
    exit 1
fi

DKMS=`type -P dkms`

find_distro

uninstall_exporter
uninstall_service

if [ -z "${INSTALLER}" ]; then
    # For RHEL variants, this must be done before the driver is removed to avoid an error

    omit_driver
fi

uninstall_drivers

if ! driver_installed; then
    remove_blacklist
fi

# If the package had CLI tools and the user decided to install them...

if [ "${CLI_TOOLS}" -a "${CLI_INSTALLDIR}" ]; then
    uninstall_cli_tools
fi

if [ -z "${INSTALLER}" ]; then
    # For non-RHEL variants, this must be done after the driver is removed to be effective

    gen_initramfs
fi

remove_uninstaller

printf "${BLU}${PRODUCT_DISP} removal complete${ESC}\n\n"

exit 0
