/****************************************************************************

DESCRIPTION: Driver framework functions - wrappers for kernel services
needed by the core layer.

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#ifndef ATDF_H
#define ATDF_H


#include "oshwcmn.h"


/* Misc. */

void    atdf_assert( void );

/* Lookaside list */

void*   atdf_kmem_cache_alloc( struct kmem_cache* cache );
void*   atdf_kmem_cache_alloc_node( struct kmem_cache* cache, int cpu_id );
void    atdf_kmem_cache_free( struct kmem_cache* cache, void* object );

/* System time */

u32     atdf_curr_time_secs( void );
u32     atdf_curr_time_ms( void );
u64     atdf_curr_time_us( void );

/* CPU */

void    atdf_cpu_barrier( void );
void    atdf_cpu_barrier_read( void );
void    atdf_cpu_barrier_write( void );
void    atdf_cpu_yield( void );
void    atdf_cpu_stall( u32 delay_us );
int     atdf_cpu_get( void );

/* Timer interface */

#ifdef __KERNEL__

/* Timer context that wraps timer_list.  Starting with the 4.15 kernel, timer_list
is expected to be embedded any user context since the 'data' field was removed.
The API now passes a pointer to timer_list to the timer callback.  To maintain
compatibility with previous kernels, we set timer_list as the callback data.
*/

struct atdf_timer {
    struct timer_list   os_timer;
    void*               data;
    void                (*handler)( void* data );
    unsigned long       period_jif;
    u8                  started;
    u8 volatile         stopped;
};

#else
struct atdf_timer;
#endif

void    atdf_timer_cancel( struct atdf_timer* timer );
void    atdf_timer_done( struct atdf_timer* timer );
void    atdf_timer_init( struct atdf_timer* timer, void* context, void(*handler)(void*) );
void    atdf_timer_restart( struct atdf_timer* timer );
void    atdf_timer_start( struct atdf_timer* timer, u32 expire_ms );
void    atdf_timer_stop( struct atdf_timer* timer );

/* Completion */

void    atdf_compl_init( struct completion* cmpl );
void    atdf_compl_reinit( struct completion* cmpl );
void    atdf_compl_wait( struct completion* cmpl );
void    atdf_compl_signal( struct completion* cmpl );
int     atdf_compl_wait_tmo( struct completion* cmpl, u32 tmo_ms );
int     atdf_compl_wait_tmo_intr( struct completion* cmpl, u32 tmo_ms );

/* SGL */

u32     atdf_sgl_get_sge_data( struct scatterlist* sge, u64* addr );
void    atdf_sgl_get_next_sge( struct scatterlist** sge );

/* String */

u32     atdf_sprintf( char* str, const char* format, ... ) __attribute__( (format( printf, 2, 3 )) );
int     atdf_sscanf( const char* str, const char* format, ... ) __attribute__( (format( scanf, 2, 3 )) );

/* Memory */

u32     atdf_mem_copy_from_user( void* src, void* dst, u32 size );
void*   atdf_mem_dma_alloc( struct device* dev, u32 size, u64* dma_addr );
void*   atdf_mem_dma_alloc_node( struct device* dev, u32 size, u64* dma_addr, int cpu );
void    atdf_mem_dma_flush( struct device* dev, u32 size, void* cpu_addr, u64 dma_addr );
void    atdf_mem_dma_free( struct device* dev, u32 size, void* cpu_addr, u64 dma_addr );
void    atdf_mem_dma_invalidate( struct device* dev, u32 size, void* cpu_addr, u64 dma_addr );
void*   atdf_mem_virt_alloc( u32 size );
void*   atdf_mem_virt_alloc_node( u32 size, u32 node );
void    atdf_mem_virt_free( void* addr );

/* Mapped SGL - Non-contiguous DMA memory mapped into virtual kernel space.
This API can be used in two ways:
- Map/unmap user buffer into kernel space with map/unmap functions.
- Allocate segmented DMA memory and map into kernel address space.
In both cases, atdf_msgl_get_sge() can get an SGE for an offset into the MSGL.
*/

struct atdf_mapd_sgl {
    struct device*  dev;
    void*           kern_addr;
    struct page**   dma_pages;
    u64*            dma_maps;
    u32             dma_offset;
    u32             num_pages;
};

struct atdf_mapd_sgl*
        atdf_msgl_alloc( struct atsi_adap* adap, u32 size, void** virt );
void    atdf_msgl_free( struct atdf_mapd_sgl* msgl );
u32     atdf_msgl_get_sge( struct atdf_mapd_sgl* msgl, u32 offset, u64* dma_addr );
void    atdf_msgl_init( struct atdf_mapd_sgl* msgl, struct atsi_adap* adap );
int     atdf_msgl_map( struct atdf_mapd_sgl* msgl, void* userptr, u32 size );
void    atdf_msgl_unmap( struct atdf_mapd_sgl* msgl );

/* PCI */

int     atdf_pci_cfg_read( struct pci_dev* pcidev, u16 offset, void* ptr, int size );
int     atdf_pci_cfg_write( struct pci_dev* pcidev, u16 offset, u32 val, int size );


#endif /* ATDF_H */
