/****************************************************************************

DESCRIPTION: Linux interface ATA functions

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2010-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#include "atdl.h"


/****************************************************************************

    FUNCTION:   OsAtaSendCommand()

    PURPOSE:    Sends a SCSI ATA_PASSTHROUGH command via the Linux kernel.

    ARGUMENTS:  cmd     - ATA_PASSTHROUGH command
                buffer  - Data buffer attached to the request
                size    - Length of the data buffer

    RETURNS:    0 if the command is sent successfully, else an errno value.

****************************************************************************/

static int OsAtaSendCommand( struct scsi_device* sd,
                             unsigned char* cmd,
                             unsigned char* buffer,
                             int size )
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6,3,0)
    if( scsi_execute_cmd( sd, cmd,
                          REQ_OP_DRV_IN,
                          buffer, size,
                          20 * HZ, 5, NULL ) )
#else
    if( scsi_execute_req( sd, cmd,
                          (size) ? DMA_FROM_DEVICE : DMA_NONE,
                          buffer, size,
                          NULL, 20 * HZ, 5, NULL ) )
#endif
    {
        atci_dbg_str( "scsi_execute failed\n" );
        return( -EIO );
    }

    return( 0 );
}


/****************************************************************************

    FUNCTION:   OsAtaSetupDrvCmd()

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_DRIVE_CMD
                Ioctl.

    ARGUMENTS:  ata_passthru_args - Arguments to the Ioctl
                args    - Pointer to the user-space ATA Ioctl arguments buffer

    RETURNS:    0 if the command is sent successfully, else an errno value.

****************************************************************************/

static int OsAtaSetupDrvCmd( unsigned char* ata_passthru_args,
                             struct scsi_device* sd,
                             void* args )
{
    unsigned char   cmd_buf[MAX_COMMAND_SIZE] = {0};
    int             rez;
    int             buf_size = ata_passthru_args [3] * 512;
    unsigned char*  ata_pass_thru_buffer = NULL;

    /* Set up the command structure */

    cmd_buf [0] = 0xa1; /* ATA PASS-THROUGH 12 */

    /* ATA Register:  */

    cmd_buf [9] = ata_passthru_args [0]; /* Command Code   */
    cmd_buf [5] = ata_passthru_args [1]; /* Sector # (LBA) */
    cmd_buf [3] = ata_passthru_args [2]; /* Features       */
    cmd_buf [4] = ata_passthru_args [3]; /* Sector count   */

    if (ata_passthru_args [0] == 0xb0)
    {
        /* This is a SMART command, put     */
        /* magic numbers in LBA Med, LBA Hi */

        cmd_buf [6] = 0x4f;
        cmd_buf [7] = 0xc2;
    }

    /* Allocate a buffer */

    if (buf_size)
    {
        ata_pass_thru_buffer = kmalloc (buf_size,
                                        GFP_KERNEL);

        if (ata_pass_thru_buffer == NULL)
        {
            rez = -ENOMEM;
            atci_dbg_str( "kmalloc (ata_pass_thru_buffer) failed\n" );
AtaPassThruError:
            if (ata_pass_thru_buffer) kfree (ata_pass_thru_buffer);
            return rez;
        }

        cmd_buf [1] = 0x08;  /* Protocol - PIO Data In */
        cmd_buf [2] = 0x0E;
    }
    else
        cmd_buf [1] = 0x06;  /* Protocol - Non-data */

    rez = OsAtaSendCommand( sd, cmd_buf, ata_pass_thru_buffer, buf_size );

    if ( rez < 0 )
        goto AtaPassThruError;
    else if ( buf_size )
    {
        /* Save the response data back to the user's buffer. */

        if (__copy_to_user ((void *) (args + 4),
                            ata_pass_thru_buffer,
                            buf_size) != 0)
        {
            atci_dbg_str( "ioctl_handler copy_to_user didn't copy it all\n" );
            rez = -EFAULT;
            goto AtaPassThruError;
        }

        kfree (ata_pass_thru_buffer);
    }

    return ( rez );
}


/****************************************************************************

    FUNCTION:   OsAtaSetupDrvTask()

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_DRIVE_TASK
                Ioctl.

    ARGUMENTS:  ata_passthru_args - Arguments to the Ioctl

    RETURNS:    0 if the command is sent successfully, else an errno value.

****************************************************************************/

static int OsAtaSetupDrvTask( unsigned char* ata_passthru_args,
                              struct scsi_device* sd )
{
    unsigned char   cmd_buf[MAX_COMMAND_SIZE] = {0};
    int             rez;

    /* ATA Register:  */

    cmd_buf [0] = 0xa1; /* ATA PASS-THROUGH 12 */
    cmd_buf [9] = ata_passthru_args [0]; /* Command Code   */
    cmd_buf [3] = ata_passthru_args [1]; /* Features       */
    cmd_buf [4] = ata_passthru_args [2]; /* Sector count   */
    cmd_buf [5] = ata_passthru_args [3]; /* Sector # (LBA) */
    cmd_buf [6] = ata_passthru_args [4]; /* Cyl low        */
    cmd_buf [7] = ata_passthru_args [5]; /* Cyl high       */
    cmd_buf [8] = ata_passthru_args [6]; /* Device head    */
    cmd_buf [1] = 0x06;     /* Protocol - Non-data */

    rez = OsAtaSendCommand( sd, cmd_buf, NULL, 0 );

    return ( rez );
}


/****************************************************************************

    FUNCTION:   OsAtaSetupSetPIOMode()

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_SET_PIO_MODE
                Ioctl.

    ARGUMENTS:  val - Value the PIO mode is to be set to.

    RETURNS:    0 if the command is sent successfully, else an errno value.

****************************************************************************/

static int OsAtaSetupSetPIOMode( long val, struct scsi_device* sd )
{
    unsigned char   cmd_buf[MAX_COMMAND_SIZE] = {0};
    int             rez;

    if ( val > 255
      || val < 0 )
        return ( -EINVAL );

    /* ATA Register:  */

    cmd_buf [0] = 0xa1; /* ATA PASS-THROUGH 12                      */
    cmd_buf [9] = 0xef; /* Command Code- SET FEATURES               */
    cmd_buf [3] = 3;    /* Features- Set transfer mode              */
    cmd_buf [4] = (unsigned char) val;  /* Sector count- PIO mode   */
    cmd_buf [1] = 0x06;     /* Protocol - Non-data */

    rez = OsAtaSendCommand( sd, cmd_buf, NULL, 0 );

    return ( rez );
}


/****************************************************************************

    FUNCTION:   OsAtaSetupSetMultcount()

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_SET_MULTCOUNT
                Ioctl.

    ARGUMENTS:  val - Value the multcount is to be set to.

    RETURNS:    0 if the command is sent successfully, else an errno value.

****************************************************************************/

static int OsAtaSetupSetMultcount( int val, struct scsi_device* sd )
{
    unsigned char   cmd_buf[MAX_COMMAND_SIZE] = {0};
    int             rez;

    if ( val > 255
      || val < 0 )
        return ( -EINVAL );

    /* ATA Register:  */

    cmd_buf [0] = 0xa1; /* ATA PASS-THROUGH 12              */
    cmd_buf [9] = 0xc6; /* Command Code- SET MULTIPLE MODE  */
    cmd_buf [3] = 3;    /* Features- Set transfer mode      */
    cmd_buf [4] = (unsigned char) val;  /* Count- Block count in this case  */
    cmd_buf [1] = 0x06;     /* Protocol - Non-data */

    rez = OsAtaSendCommand( sd, cmd_buf, NULL, 0 );

    return ( rez );
}


/****************************************************************************

    FUNCTION:   OsAtaSetupDrvReset()

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_DRIVE_RESET
                Ioctl.

    RETURNS:    0 if the command is sent successfully, else an errno value.

****************************************************************************/

static int OsAtaSetupDrvReset( struct scsi_device* sd )
{
    unsigned char   cmd_buf[MAX_COMMAND_SIZE] = {0};
    int             rez;

    /* ATA Register:  */
    /* Fill in a DEVICE RESET command */

    cmd_buf [0] = 0xA1; /* ATA PASS-THROUGH 12 */
    cmd_buf [1] = 0x02; /* Protocol - Soft Reset */

    rez = OsAtaSendCommand( sd, cmd_buf, NULL, 0 );

    return ( rez );
}


/****************************************************************************

    FUNCTION:   OsAtaSetupGetMultcount()

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_GET_MULTICOUNT
                Ioctl.

    ARGUMENTS:  args    - Pointer to the user-space ATA Ioctl arguments buffer

    RETURNS:    0 if the command is sent successfully, else an errno value.

****************************************************************************/

static int OsAtaSetupGetMultcount( struct scsi_device* sd, void* args )
{
    unsigned char   cmd_buf[MAX_COMMAND_SIZE] = {0};
    int  rez;
    int  buf_size   = 512;
    long multcount;
    u16* ata_pass_thru_buffer = kmalloc (buf_size, GFP_KERNEL);

    if (ata_pass_thru_buffer == NULL)
    {
        rez = -ENOMEM;
        atci_dbg_str( "kmalloc (ata_pass_thru_buffer) failed\n" );
AtaPassThruError:
        if (ata_pass_thru_buffer) kfree (ata_pass_thru_buffer);
        return rez;
    }

    /* All we do here is send out an IDENTIFY DEVICE command, and return    */
    /* the byte that specifies the MultCount.                               */

    /* Set up the command structure */

    cmd_buf [0] = 0xa1; /* ATA PASS-THROUGH 12 */

    /* ATA Register:  */

    cmd_buf [9] = 0xEC; /* Command Code- Identify Device */
    cmd_buf [1] = 0x08; /* Protocol - PIO Data In */
    cmd_buf [2] = 0x0E; /* Flags- we are sending blocks from drive to adap */
    cmd_buf [4] = 1;    /* Sector Count- Block count expected in this case */

    rez = OsAtaSendCommand( sd,
                            cmd_buf,
                            (unsigned char*) ata_pass_thru_buffer,
                            buf_size );

    if ( rez < 0 )
        goto AtaPassThruError;
    else
    {
        /* Copy the number of logical sectors transferred on a READ/WRITE   */
        /* MULTIPLE command back to the user's buffer.                      */

        /* First check that the count is valid */

        if ( (ata_pass_thru_buffer[59] & 0xff00) == 0x100 )
            multcount = ata_pass_thru_buffer[59] & 0xff;
        else
            multcount = 0;

        atci_dbg_print( "field: %x  multcount: %ld\n", ata_pass_thru_buffer[59], multcount );

        /* Then copy the value to user-space */

        if (__copy_to_user ( args,
                             &multcount,
                             sizeof(long) ) != 0)
        {
            atci_dbg_str( "ioctl_handler copy_to_user didn't copy it all\n" );
            rez = -EFAULT;
            goto AtaPassThruError;
        }

        kfree (ata_pass_thru_buffer);

        return ( rez );
    }
}


/**
 * atdl_ata_ioctl_handler() - Handle ATA IOCTLs
 * 
 * @sd:     Scsi device to handle the IOCTL
 * @cmd:    Command number for the IOCTL
 * @arg:    User buffer for the IOCTL data
 * 
 * Returns: System error code
 */

int atdl_ata_ioctl_handler( struct scsi_device* sd, u32 cmd, void* arg )
{
    void*   ioctl_data = NULL;
    int     rslt;
    int     size;
    struct  atdf_mapd_sgl userbuf;

    atto_log_info( ATTO_LOG_IOCTL, "ATA ioctl (%p, %x, %p)\n", sd, cmd, arg );

    atdf_msgl_init( &userbuf, &atdl_sh_to_hadap( sd->host )->adap );

    /* Some processes (eg. hdparm) may send us ioctls that are only meant   */
    /* for IDE devices, or are so old that they are not supported by        */
    /* libata.  Our policy is to support SG_IO, all the HDIO ioctls that    */
    /* libata supports, and all the ioctls that can be sent through an ATA  */
    /* PASSTHROUGH command.                                                 */

    switch( cmd )
    {
        /* The first three do have data passed in arg, but for the sake of  */
        /* mapping, the size is 0.                                          */

        case HDIO_GET_MULTCOUNT:
        case HDIO_SET_MULTCOUNT:
        case HDIO_SET_PIO_MODE:
        case HDIO_DRIVE_RESET:      size = 0; break;
        case HDIO_DRIVE_CMD:        size = 4; break;
        case HDIO_DRIVE_TASK:       size = 7; break;

        default:                    return -EINVAL;
    }

    if( size )
    {
        rslt = atdf_msgl_map( &userbuf, arg, size );

        if( rslt )
            return rslt;

        ioctl_data = userbuf.kern_addr + userbuf.dma_offset;
    }

    switch( cmd )
    {
        case HDIO_GET_MULTCOUNT:
            rslt = OsAtaSetupGetMultcount( sd, arg );
            break;

        case HDIO_SET_MULTCOUNT:
            rslt = OsAtaSetupSetMultcount( (int)(long) arg, sd );
            break;

        case HDIO_SET_PIO_MODE:
            rslt = OsAtaSetupSetPIOMode( (long) arg, sd );
            break;

        case HDIO_DRIVE_RESET:
            rslt = OsAtaSetupDrvReset( sd );
            break;

        case HDIO_DRIVE_CMD:
            rslt = OsAtaSetupDrvCmd( ioctl_data, sd, arg );
            break;

        case HDIO_DRIVE_TASK: /* Also covers HDIO_DRIVE_CMD_AEB */
            rslt = OsAtaSetupDrvTask( ioctl_data, sd );
            break;

        default:
            rslt = -ENOSYS;
            break;
    }

    atdf_msgl_unmap( &userbuf );

    return rslt;
}

