/****************************************************************************

DESCRIPTION: Driver Parameter Management

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2021
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#include "atdl.h"
#include <hwdpm.h>


/* Helpers */

static int atdl_is_hex_str( const char* buf, size_t len )
{
    size_t index;

    for( index = 0; index < len; index++ )
    {
        if( !isxdigit( buf[index] ) )
            return -1;
    }

    return 0;
}

static int atdl_is_white( char c )
{
    return c == ' ' || c == '\r' || c == '\n';
}

static int atdl_str_trim( const char** buf, size_t len )
{
    const char* start = *buf;
    const char* end = *buf + len;

    while( start < end )
    {
        if( !atdl_is_white( *start ) )
            break;

        start++;
    }

    while( start < end )
    {
        if( !atdl_is_white( *(end - 1) ) )
            break;

        end--;
    }

    *buf = start;

    return end - start;
}


/* Module parameter creation macros */

#define ATDI_MOD_PARAM_ULONG(name, default, desc) \
    unsigned long name = default; \
    module_param( name, ulong, S_IRUGO | S_IRUSR ); \
    MODULE_PARM_DESC( name, desc )

#define ATDI_MOD_PARAM_U32(name, default, desc) \
    unsigned int name = default; \
    module_param( name, uint, S_IRUGO | S_IRUSR ); \
    MODULE_PARM_DESC( name, desc )


/**
 * atdl_dev_attr_show_dec() - Show a device attribute in decimal 
 */

static ssize_t atdl_dev_attr_show_dec( char* buf, u32 val )
{
    return snprintf( buf, PAGE_SIZE, "%d\n", val );
}


/**
 * atdl_dev_attr_show_x32() - Show a device attribute in hex
 */

static ssize_t atdl_dev_attr_show_x32( char* buf, u32 val )
{
    return snprintf( buf, PAGE_SIZE, "0x%08x\n", val );
}


/**
 * atdl_dev_attr_show_str() - Show a device attribute string (custom)
 */

static ssize_t atdl_dev_attr_show_str( struct device* dev, struct device_attribute* attr, char* buf )
{
    struct Scsi_Host*   sh = class_to_shost( dev );
    const char*         name = attr->attr.name;

    if( strcmp( name, "cdev_path" ) == 0 )
        return snprintf( buf, PAGE_SIZE, "/dev/" ATTO_DRVR_NAME "/%d\n", sh->unique_id );
    else
    {
#ifdef CELERITY5
        struct atdl_adap* hostadap = atdl_sh_to_hadap( sh );
        struct atcl_adap* coreadap = hostadap->coreadap;
        int rslt = 0;

        if( strcmp( name, "port_address" ) == 0 )
            rslt = atci_adap_addr_str_get( coreadap, 1, buf );
        else if( strcmp( name, "node_address" ) == 0 )
            rslt = atci_adap_addr_str_get( coreadap, 0, buf );
        else
            return -EINVAL;

        strcat( buf, "\n" );
        
        return rslt + 1;
#endif
    }

    return -EINVAL;
}


/**
 * atdl_dev_attr_store_dec() - Store a device attribute from decimal 
 */

static ssize_t atdl_dev_attr_store_dec( struct device *dev, struct device_attribute *attr,
                                        const char *buf, size_t count, u32* val )
{
    u32         newval = 0;
    const char* p = buf;
    int         len = atdl_str_trim( &p, count );
    size_t      index;

    if( len == 0 )
        return -EINVAL;

    for( index = 0; index < len; index++ )
    {
        if( !isdigit( p[index] ) )
            return -EINVAL;
    }

    if( sscanf( p, "%u", &newval ) != 1 )
        return -EINVAL;

    *val = newval;

    return count;
}


/**
 * atdl_dev_attr_store_x32() - Store a device attribute from hex
 */

static ssize_t atdl_dev_attr_store_x32( struct device *dev, struct device_attribute *attr,
                                        const char *buf, size_t count, u32* val )
{
    u32         newval = 0;
    const char* p = buf;
    int         len = atdl_str_trim( &p, count );
#ifdef SF_DEBUG
    const char* name = attr->attr.name;
#endif

    if( len < 3
     || p[0] != '0'
     || p[1] != 'x' )
        return -EINVAL;

    if( atdl_is_hex_str( p + 2, len - 2 ) )
        return -EINVAL;

    if( sscanf( p, "%x", &newval ) != 1 )
        return -EINVAL;

    *val = newval;

#ifdef SF_DEBUG
    if( strcmp( name, "atto_sflib_mask" ) == 0 )
        atci_sflib_mask_set( newval );
#endif

    return count;
}


/**
 * atdl_dev_attr_store_str() - Store a device attribute from string (custom)
 */

#ifdef CELERITY5

static ssize_t atdl_dev_attr_store_str( struct device *dev, struct device_attribute *attr,
                                        const char *buf, size_t count )
{
    struct Scsi_Host*   sh = class_to_shost( dev );
    const char*         p = buf;
    int                 len = atdl_str_trim( &p, count );
    const char*         name = attr->attr.name;
    struct atdl_adap*   hostadap = atdl_sh_to_hadap( sh );
    struct atcl_adap*   coreadap = hostadap->coreadap;
    int rslt = -1;

    if( atdl_is_hex_str( p, len ) )
        return -EINVAL;

    if( strcmp( name, "port_address" ) == 0 )
        rslt = atci_adap_addr_str_set( coreadap, 1, p, len );
    else if( strcmp( name, "node_address" ) == 0 )
        rslt = atci_adap_addr_str_set( coreadap, 0, p, len );

    if( rslt )
        return -EINVAL;

    return count;
}

#endif


/* Macros to create wrappers for device attribute show/store methods */

#define ATDL_DEV_ATTR_SHOW_DEC(name) \
static ssize_t name##_show( struct device *dev, struct device_attribute *attr, char *buf ) \
{ \
    return atdl_dev_attr_show_dec( buf, name ); \
}

#define ATDL_DEV_ATTR_STORE_DEC(name) \
static ssize_t name##_store( struct device *dev, struct device_attribute *attr, const char *buf, size_t count ) \
{ \
    return atdl_dev_attr_store_dec( dev, attr, buf, count, &(name) ); \
}

#define ATDL_DEV_ATTR_SHOW_X32(name) \
static ssize_t name##_show( struct device *dev, struct device_attribute *attr, char *buf ) \
{ \
    return atdl_dev_attr_show_x32( buf, name ); \
}

#define ATDL_DEV_ATTR_STORE_X32(name) \
static ssize_t name##_store( struct device *dev, struct device_attribute *attr, const char *buf, size_t count ) \
{ \
    return atdl_dev_attr_store_x32( dev, attr, buf, count, &(name) ); \
}

#define ATDL_DEV_ATTR_SHOW_STR(name) \
static ssize_t name##_show( struct device *dev, struct device_attribute *attr, char *buf ) \
{ \
    return atdl_dev_attr_show_str( dev, attr, buf ); \
}

#define ATDL_DEV_ATTR_STORE_STR(name) \
static ssize_t name##_store( struct device *dev, struct device_attribute *attr, const char *buf, size_t count ) \
{ \
    return atdl_dev_attr_store_str( dev, attr, buf, count ); \
}


/* Device attribute creation macros */

/* These macros are used to hide the differences across kernel versions.    */
/* Each attribute that is readable needs a 'show' fn. Each attribute that   */
/* is writeable needs a 'store' fn. These must follow the naming convention */
/* of name##_show and name##_store.                                         */

#define ATDL_DEV_ATTR_R(name) static DEVICE_ATTR(name, S_IRUGO, name##_show, NULL)
#define ATDL_DEV_ATTR_RW(name) static DEVICE_ATTR(name, S_IRUGO | S_IWUSR, name##_show, name##_store)
#define ATDL_DEV_ATTR_NAME(name) dev_attr_##name
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,16,0) && !IS_SLES_RELEASE_OR_LATER(15,5,0)
#define ATDL_DEV_ATTR_PTR( name ) &ATDL_DEV_ATTR_NAME(name)
#else
#define ATDL_DEV_ATTR_PTR( name ) &ATDL_DEV_ATTR_NAME(name).attr
#endif


#define ATDL_DEV_ATTR_DEC_RW(name) \
    ATDL_DEV_ATTR_SHOW_DEC(name); \
    ATDL_DEV_ATTR_STORE_DEC(name); \
    ATDL_DEV_ATTR_RW(name)

#define ATDL_DEV_ATTR_DEC_R(name) \
    ATDL_DEV_ATTR_SHOW_DEC(name); \
    ATDL_DEV_ATTR_R(name)

#define ATDL_DEV_ATTR_X32_RW(name) \
    ATDL_DEV_ATTR_SHOW_X32(name); \
    ATDL_DEV_ATTR_STORE_X32(name); \
    ATDL_DEV_ATTR_RW(name)

#define ATDL_DEV_ATTR_X32_R(name) \
    ATDL_DEV_ATTR_SHOW_X32(name); \
    ATDL_DEV_ATTR_R(name)

#define ATDL_DEV_ATTR_STR_RW(name) \
    ATDL_DEV_ATTR_SHOW_STR(name); \
    ATDL_DEV_ATTR_STORE_STR(name); \
    ATDL_DEV_ATTR_RW(name)

#define ATDL_DEV_ATTR_STR_R(name) \
    ATDL_DEV_ATTR_SHOW_STR(name); \
    ATDL_DEV_ATTR_R(name)


/* Module parameter and/or device attribute definitions */

#ifdef HW_VSA_REBUILD
ATDI_MOD_PARAM_U32( raid_rebuild_mode, 0, HW_DPM_HELP_RAID_RBLD_MODE );
#endif

#ifdef HW_TRACING
ATDI_MOD_PARAM_U32( trace_buf_size, 0x100000, HW_DPM_HELP_TRACE_BUF_SIZE );
ATDI_MOD_PARAM_U32( trace_buf_wrap, 1, HW_DPM_HELP_TRACE_BUF_WRAP );
#endif

#if defined(_MP_)
/* quick_init always on for multipathing since device      */
/* discovery is delayed until the paths can be configured. */

ATDI_MOD_PARAM_U32( quick_init, 1, OS_DPM_HELP_QUICK_INIT );
ATDI_MOD_PARAM_U32( max_luns, 1024, OS_DPM_HELP_MAX_LUNS );
ATDI_MOD_PARAM_U32( mp_init_wait, 60, OS_DPM_HELP_MP_INIT_WAIT );
#elif defined(PROTO_SAS) || defined(PROTO_FC)
ATDI_MOD_PARAM_U32( quick_init, 0, OS_DPM_HELP_QUICK_INIT );
#endif /* _MP_ */

#if defined(PROTO_FC)
ATDI_MOD_PARAM_U32( max_logins, 0, HW_DPM_HELP_TARGETS );
#endif

#if defined(CELERITY5)
ATDI_MOD_PARAM_U32( multi_tgt_queue, 0xFFFFFFFF, HW_DPM_HELP_TGT_MULTI_QUEUE );
ATDI_MOD_PARAM_U32( reg_sgl_percent, 0xFFFFFFFF, HW_DPM_HELP_REG_SGL_PERCENT );
ATDI_MOD_PARAM_U32( eq_delay, 0xFFFFFFFF, HW_DPM_HELP_INT_COAL_DELAY );
ATDL_DEV_ATTR_STR_RW( port_address );
ATDL_DEV_ATTR_STR_RW( node_address );
#elif defined(CELERITY4)
ATDI_MOD_PARAM_U32( force_low_dma, 0, OS_DPM_HELP_FORCE_LOW_DMA );
ATDI_MOD_PARAM_U32( fab_alternate_exchanges, 0, HW_DPM_HELP_FAB_ALT_XCHG );
#endif

#ifdef HW_NPIV
ATDI_MOD_PARAM_U32( npiv_enable, 0, HW_DPM_HELP_NPIV );
ATDI_MOD_PARAM_U32( max_npiv_vports, 16, HW_DPM_HELP_NPIV_VIRT_PORTS );
ATDI_MOD_PARAM_U32( max_npiv_logins, 64, HW_DPM_HELP_NPIV_TARGETS );
#endif

#ifdef FEAT_SRIOV
ATDI_MOD_PARAM_U32( sriov_mode, 1, OS_DPM_HELP_SRIOV_MODE );
ATDI_MOD_PARAM_U32( sriov_numvfs, 0, HW_DPM_HELP_SRIOV_NUM_VFS );
#endif

#if defined(HW_SCSI_TGT) || defined(HW_NVME_TGT)

/* Enable target mode on all ports.  For FibreChannel adapters that support it,
 * either FCP or NVMe target mode can be enabled (assuming kernel support for
 * NVMe).
 *
 *     0 = Disabled target mode                 ATTO_TARGET_MODE_DISABLED
 *     1 = Enable (FCP/SAS) target mode         ATTO_TARGET_MODE_SCSI
 *     2 = Enable NVMe target mode (FC only)    ATTO_TARGET_MODE_NVME
 *
 * Default: 0
 */

ATDI_MOD_PARAM_U32( target_mode, ATTO_TARGET_MODE_DISABLED, HW_DPM_HELP_TGT_MODE );
#ifndef ESASHBA6
ATDI_MOD_PARAM_U32( multimode_ports, 0xFFFFFFFF, HW_DPM_HELP_MULTI_MODE_PORTS );
#endif
ATDI_MOD_PARAM_U32( target_mode_link_delay, 1, HW_DPM_HELP_TGT_LINK_DELAY );

#if defined(PROTO_FC)

ATDI_MOD_PARAM_U32( target_mode_mask, 0, OS_DPM_HELP_TGT_MODE_MASK );
ATDI_MOD_PARAM_U32( initiator_mode_mask, 0xffffffff, OS_DPM_HELP_INI_MODE_MASK );
ATDI_MOD_PARAM_U32( num_xcb, 0, HW_DPM_HELP_TGT_CMDS );
ATDI_MOD_PARAM_U32( num_scb, 0, HW_DPM_HELP_TGT_SEQUENCES );

#elif defined(PROTO_SAS)
ATDI_MOD_PARAM_U32( initiator_phy_mask, 0xFFFFFFFF, HW_DPM_HELP_INI_PHY_MASK );
ATDI_MOD_PARAM_U32( target_phy_mask, 0xFFFFFFFF, HW_DPM_HELP_TGT_PHY_MASK );
ATDI_MOD_PARAM_U32( num_tgt_req, 0, HW_DPM_HELP_TGT_CMDS );
#endif
#endif /* HW_SCSI_TGT || HW_NVME_TGT */

#if defined(PROTO_SAS)
ATDI_MOD_PARAM_U32( num_targets, 0, HW_DPM_HELP_TARGETS );
ATDI_MOD_PARAM_U32( fw_log_size, 0xFFFFFFFF, HW_DPM_HELP_FW_LOG_SIZE );
ATDI_MOD_PARAM_U32( num_sas_addr, 0xFFFFFFFF, HW_DPM_HELP_NUM_SAS_ADDR );
#endif

#if defined(FEAT_ZERO_EXP_CHANGE_CNT)
ATDI_MOD_PARAM_U32( zero_eecc, 0, HW_DPM_HELP_ZERO_EECC );
#endif

#ifdef FEAT_SAS_DEV_ORDER
ATDI_MOD_PARAM_U32( sas_dev_order, 0, OS_DPM_HELP_SAS_DEV_ORDER );
ATDL_DEV_ATTR_DEC_RW( sas_dev_order );
#endif

ATDI_MOD_PARAM_U32( log_level, 0xFFFFFFFF, HW_DPM_HELP_LOG_LEVEL );
ATDL_DEV_ATTR_DEC_RW( log_level );

#if defined(PROTO_SAS) || defined(PROTO_FC)
ATDI_MOD_PARAM_U32( can_queue, 0, HW_DPM_HELP_INI_CMDS );
ATDI_MOD_PARAM_U32( cmd_per_lun, 1024, OS_DPM_HELP_CMD_PER_LUN );
ATDI_MOD_PARAM_U32( num_sg_lists, 0, HW_DPM_HELP_SGL_SEGMENTS );
ATDI_MOD_PARAM_U32( sgl_page_size, 0, HW_DPM_HELP_SGL_SEG_DATA_SGE );
ATDI_MOD_PARAM_U32( cmd_retry_count, 0xFFFFFFFF, OS_DPM_HELP_CMD_RETRY_CNT );
#endif

ATDI_MOD_PARAM_U32( change_notification, 1, OS_DPM_HELP_DEV_CHG_NOTIFY );
ATDI_MOD_PARAM_U32( max_xfer_size, 0, OS_DPM_HELP_MAX_XFER_SIZE );
ATDI_MOD_PARAM_U32( cpu_affinity, 1, OS_DPM_HELP_CPU_AFFINITY );

#ifdef FEAT_DISABLE_DEV_SCAN
ATDI_MOD_PARAM_U32( disable_device_scan, 0, HW_DPM_HELP_DIS_DEV_SCAN );
#endif

#if defined(FEAT_TRANSPORT)
ATDI_MOD_PARAM_U32( use_transport_layer, 1, OS_DPM_HELP_XPORT_LAYER );

#if defined(PROTO_FC)
ATDI_MOD_PARAM_U32( rport_timeout, 30, OS_DPM_HELP_RPORT_TIMEOUT );
#endif
#endif /* FEAT_TRANSPORT */

ATDI_MOD_PARAM_U32( atto_log_mask, ATTO_LOG_MASK_DEFAULT, OS_DPM_HELP_EVT_TYPE_MASK );
ATDL_DEV_ATTR_X32_RW( atto_log_mask );

#ifdef SF_DEBUG
ATDI_MOD_PARAM_U32( atto_sflib_mask, 0, OS_DPM_HELP_SF_DBG_MASK );
ATDL_DEV_ATTR_X32_RW( atto_sflib_mask );
#endif

#ifdef FEAT_CDEV_IOCTL
ATDL_DEV_ATTR_STR_R( cdev_path );
#endif


/* Full attributes list */

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,16,0) && !IS_SLES_RELEASE_OR_LATER(15,5,0)
struct device_attribute* g_atdl_host_attrs[] = {
#else
struct attribute* g_atdl_host_attrs[] = {
#endif
#ifdef FEAT_CDEV_IOCTL
    ATDL_DEV_ATTR_PTR( cdev_path ),
#endif
    ATDL_DEV_ATTR_PTR( log_level ),
    ATDL_DEV_ATTR_PTR( atto_log_mask ),
#ifdef FEAT_SAS_DEV_ORDER
    ATDL_DEV_ATTR_PTR( sas_dev_order ),
#endif
#ifdef SF_DEBUG
    ATDL_DEV_ATTR_PTR( atto_sflib_mask ),
#endif
#ifdef CELERITY5
    ATDL_DEV_ATTR_PTR( port_address ),
    ATDL_DEV_ATTR_PTR( node_address ),
#endif
    NULL,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,16,0) || IS_SLES_RELEASE_OR_LATER(15,5,0)
static const struct attribute_group g_atdl_host_attr_group = {
    .attrs = g_atdl_host_attrs
};

const struct attribute_group* g_atdl_host_groups[] = {
	&g_atdl_host_attr_group,
	NULL
};
#endif

