/****************************************************************************

DESCRIPTION: Initialization functions

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#include "atdl.h"


MODULE_DESCRIPTION( ATDL_ADAP_NAME " Driver" );
MODULE_AUTHOR( "ATTO Technology" );
MODULE_LICENSE( "Proprietary" );
MODULE_VERSION( AV_FULL_VERSION_STR );


/* Globals */

struct atdl_globals     g_atdl_vars;
struct atgi_connect_ops g_atgi_ops;


static struct pci_device_id g_atdl_pci_id_table[] = {
    #include <ospciids.h>
    { 0 } /* The required NULL terminator */
};

MODULE_DEVICE_TABLE( pci, g_atdl_pci_id_table );


/**
 * atdl_init_intr_affinity() - Setup initial affinity hint and notification callback
 */

static void atdl_init_intr_affinity( struct atdl_adap* hostadap, int first_init )
{
    struct atsi_adap*       adap = &hostadap->adap;
    u8                      msi_cnt_io = adap->msi_cfg.MsiCntIo;
    struct atdl_msi_vector* msi_vect;
    int                     cpu;
    u32                     vect;

    if( !msi_cnt_io )
        msi_cnt_io++;

    while( first_init )
    {
        int nodes;
        int cpus;
        int node;
        u32 nodes_per_vect;

        /* Round robin available I/O vectors to online CPUs.  All possible  */
        /* CPUs are initialized in case one is hot plugged, although we're  */
        /* not fully supporting CPU hot plugging.                           */

        for_each_possible_cpu( cpu )
            adap->cpu_to_msi_vect[cpu] = cpu % msi_cnt_io;

        /* From here, we only care about online CPUs */

        nodes = num_online_nodes();
        cpus = num_online_cpus();

        /* Defaults are fine for a single node/vector */

        if( nodes <= 1
         || msi_cnt_io <= 1 )
            break;

        /* Defaults are also fine if we have enough vectors to cover all CPUs */

        if( cpus <= msi_cnt_io )
            break;

        /* We have a NUMA system without enough vectors for all CPUs.  Since*/
        /* vector memory is allocated out of some node, the CPUs mapped to  */
        /* it should be out of the same node.                               */

        vect = 0;
        nodes_per_vect = nodes / msi_cnt_io;

        for_each_node_with_cpus( node )
        {
            u32                 vect_per_node = msi_cnt_io / nodes;
            u32                 cpus_per_vect = cpus / msi_cnt_io;
            const cpumask_t*    mask = cpumask_of_node( node );

            for_each_cpu( cpu, mask )
            {
                adap->cpu_to_msi_vect[cpu] = vect;

                if( cpus_per_vect )
                    cpus_per_vect--;

                if( !cpus_per_vect )
                {
                    /* CPU limit reached for current vector */

                    if( vect_per_node )
                        vect_per_node--;

                    if( vect_per_node )
                    {
                        /* Still have vectors to use on this node */
                        
                        cpus_per_vect = cpus / msi_cnt_io;

                        /* Make sure we have another vector to use.  If not */
                        /* we'll fill the node with the last used.  This    */
                        /* covers cases where counts are not evenly         */
                        /* divisible or there are more nodes than vectors   */

                        if( vect < msi_cnt_io - 1 )
                            vect++;
                    }
                }
            }

            /* See if we can advance the vector when switching nodes */

            if( nodes_per_vect )
                nodes_per_vect--;

            if( !nodes_per_vect
             && vect < msi_cnt_io - 1 )
            {
                nodes_per_vect = nodes / msi_cnt_io;
                vect++;
            }
        }

        break;
    }

    /* Now build up the affinity mask for each vector */

    if( first_init )
    {
        for_each_online_cpu( cpu )
        {
            msi_vect = hostadap->msi_vectors + adap->cpu_to_msi_vect[cpu];

            cpumask_set_cpu( cpu, &msi_vect->aff_mask );
        }
    }

    /* Bail out if not doing ATTO balancing */

    if( cpu_affinity != 1 )
        return;

    /* Set the affinity and hint */

    for( vect = 0; vect < hostadap->msi_cnt; vect++ )
    {
        msi_vect = hostadap->msi_vectors + vect;

        if( cpumask_empty( &msi_vect->aff_mask ) )
            continue;

        if( g_atgi_ops.irq_set_affinity_hint( msi_vect->irq, &msi_vect->aff_mask ) )
        {
            atdl_log_err( hostadap, ATTO_LOG_INIT,
                          "Failed to set affinity hint on vector %u (%u)\n",
                          vect,
                          cpumask_weight( &msi_vect->aff_mask ) );
        }
    }
}


/**
 * atdi_free_interrupts() - Disable and release interrupts
 */

void atdi_free_interrupts( struct atdl_adap* hostadap )
{
    struct atsi_adap* adap = &hostadap->adap;

    if( adap->intr_mode == ATSI_INTR_MODE_MSIX )
    {
        atdl_intr_msix_disable( hostadap );
        atci_dbg_str( " MSI-X vectors released  MSI-X disabled  " );
    }
    else if( adap->intr_mode == ATSI_INTR_MODE_MSI )
    {
        atdl_intr_msi_disable( hostadap );
        atci_dbg_str( " MSI disabled  " );
    }
#ifndef HW_NO_LEGACY_INTS
    else if( adap->intr_mode == ATSI_INTR_MODE_LEGACY )
    {
        atdl_intr_legacy_disable( hostadap );
        atci_dbg_str( " IRQ released  " );
    }
#endif

    adap->intr_mode = ATSI_INTR_MODE_UNCLAIMED;
}


/**
 * atdi_init_interrupts() - Reserve, claim, and initialize interrupts
 * 
 * Returns: The current interrupt mode
 */

u8 atdi_init_interrupts( struct atdl_adap* hostadap, u8 max_vectors )
{
    struct atsi_adap*   adap = &hostadap->adap;
    PHWINIT_MSI         msicfg = &hostadap->adap.msi_cfg;    
    int                 first_init = 0;
    u8                  msicnt;

    /* On the first call, allocate the vector table */

    if( !hostadap->msi_vectors )
    {
        u32 vect;

        hostadap->msi_vectors = kcalloc( max_vectors, sizeof(*hostadap->msi_vectors), GFP_KERNEL );

        if( !hostadap->msi_vectors )
        {
            atdl_log_err( hostadap, ATTO_LOG_INIT,
                          "Failed to allocate MSI vector context table\n" );

            return ATSI_INTR_MODE_UNCLAIMED;
        }

        hostadap->msi_cnt = max_vectors;

        for( vect = 0; vect < max_vectors; vect++ )
        {
            struct atdl_msi_vector* msi_vect = hostadap->msi_vectors + vect;

            msi_vect->irq = -1;
            msi_vect->hostadap = hostadap;
            msi_vect->vector = vect;
        }

        first_init = 1;
    }

    memset( msicfg, 0, sizeof(*msicfg) );

    /* Try MSI-X first.  Enable desired vectors */

    msicnt = atdl_intr_msix_enable( hostadap, max_vectors );

    if( msicnt == max_vectors )
    {
        adap->intr_mode = ATSI_INTR_MODE_MSIX;
        goto init_hw;
    }

    atdl_intr_msix_disable( hostadap );

    atdl_log_err( hostadap, ATTO_LOG_INIT,
                  "Failed to enable MSI-X, falling back to MSI.\n" );

    /* Try MSI next */

    msicnt = atdl_intr_msi_enable( hostadap );

    if( msicnt )
    {
        adap->intr_mode = ATSI_INTR_MODE_MSI;
        goto init_hw;
    }

    atdl_log_err( hostadap, ATTO_LOG_INIT,
                  "Failed to enable MSI, falling back to legacy interrupts)\n" );

#ifndef HW_NO_LEGACY_INTS
    /* Last gasp.  Try legacy interrupts */

    if( !atdl_intr_legacy_enable( hostadap ) )
        adap->intr_mode = ATSI_INTR_MODE_LEGACY;
    else
#endif
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT,
                      "Failed to claim interrupts, failing init.\n" );

        return ATSI_INTR_MODE_UNCLAIMED;
    }

init_hw:

    /* Now initialize the hardware layer.  It will figure out the enabled mode */

    msicfg->MsiCnt = msicnt;

    if( msicnt > 1 )
        msicfg->Options |= HWINITMSI_MULTI_VECTOR;

    /* Verify the mode matches for the hell of it */

    if( atci_init_adap_cfg( hostadap->coreadap )
     || (adap->intr_mode == ATSI_INTR_MODE_MSIX
      && !(msicfg->Results & HWINITMSI_MSIX_ENB))
     || (adap->intr_mode == ATSI_INTR_MODE_MSI
      && !(msicfg->Results & HWINITMSI_MSI_ENB)) )
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT,
                      "Failed to init interrupts in HW, mode %u, result %d, failing init.\n",
                      adap->intr_mode, msicfg->Results );

        atdi_free_interrupts( hostadap );
    }
    else
    {
        atdl_log_info( hostadap, ATTO_LOG_INIT, "Interrupts successfully initialized\n" );

        if( adap->intr_mode == ATSI_INTR_MODE_MSIX )
            atdl_init_intr_affinity( hostadap, first_init );
        else
        {
            /* For MSI and legacy interrupts, HwrqSetMsiVector() will just  */
            /* set vector 0 for all CPUs.                                   */
        }
    }

    return adap->intr_mode;
}


/**
 * atdi_adap_reference() - Add a reference to the specified adapter
 * 
 * Returns:  0  the adapter is valid, reference added
 *          -1  the adapter is not present or terminating
 *          -2  the adapter is present, but no work is allowed
 */

int atdi_adap_reference( struct atsi_adap* adap )
{
    int             rslt = -1;
    int             i;
    unsigned long   lockflags;

    if( adap == NULL )
        return -1;

    spin_lock_irqsave( &g_atdl_vars.adap_lock, lockflags );

    for( i = 0; i < ATSI_MAX_ADAPTERS; i++ )
    {
        if( g_atsi_vars.adap_list[i] != adap )
            continue;

        if( adap->flags & ATSI_AF_TERMINATING )
            break;

        if( adap->flags & ATSI_AF_ALLOW_WORK )
        {
            /* The adapter is not terminating.  Note we don't reference the */
            /* SCSI host since it is released after we quiesce our services */
            /* on teardown.                                                 */

            adap->ref_cnt++;
            rslt = 0;
        }
        else
            rslt = -2;

        break;
    }

    spin_unlock_irqrestore( &g_atdl_vars.adap_lock, lockflags );

    return rslt;
}


/**
 * atdi_adap_dereference() - Remove a reference to the specified adapter
 */

void atdi_adap_dereference( struct atsi_adap* adap )
{
    unsigned long   lockflags;

    spin_lock_irqsave( &g_atdl_vars.adap_lock, lockflags );

    if( adap->ref_cnt )
        adap->ref_cnt--;

    spin_unlock_irqrestore( &g_atdl_vars.adap_lock, lockflags );
}


/**
 * atdl_init_dma_mask() - Set the PCI DMA mask
 *
 * @pcidev: PCI device
 * @size:   Number of valid mask bits (32 or 64)
 * 
 * Returns: 0 on success, non-zero on failure
 */

static int atdl_init_dma_mask( struct atdl_adap* hostadap, struct pci_dev* pcidev, int size )
{
    u64 dma_mask;

    /* First set the coherent mask to the requested value */

    dma_mask = DMA_BIT_MASK( size );

    if( dma_set_coherent_mask( &pcidev->dev, dma_mask ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT, "Error setting %d-bit coherent DMA mask!\n", size );
        return -1;
    }

    /* Set the DMA mask to 64 bits, unless forced down to 32 bits via a     */
    /* module parameter (Celerity 8Gb can only access 45 bits of memory).   */

    dma_mask = DMA_BIT_MASK( 64 );

#ifdef CELERITY4
    if( force_low_dma )
        dma_mask = DMA_BIT_MASK( 32 );
#endif

    if( dma_set_mask( &pcidev->dev, dma_mask ) )
        atdl_log_err( hostadap, ATTO_LOG_INIT, "Error setting 0x%llx DMA mask!\n", dma_mask );

    return 0;
}


/**
 * atdl_start_services() - Start the timer, DPC and worker
 */

static void atdl_start_services( struct atdl_adap* hostadap )
{
    struct atsi_adap* adap = &hostadap->adap;

    /* The DPC was already enabled to prevent a mismatch with the HW layer  */
    /* scheduling semaphore.                                                */

    if( adap->flags & ATSI_AF_HW_INIT_FAIL )
        return;

    /* Start the timer */

    atdl_log_info( hostadap, ATTO_LOG_INIT, "Starting timer\n" );

    atdf_timer_start( &hostadap->timer, adap->timer_period );

    /* Start the worker */

    atdf_set_bits( &adap->flags, ATSI_AF_ALLOW_WORK );

    atdi_worker_schedule( 0 );
}


/**
 * atdl_block_services() - Block new work from being scheduled
 */

static void atdl_block_services( struct atdl_adap* hostadap )
{
    u32 starttime;

    /* Prevent new work from being scheduled */

    atdf_clear_bits( &hostadap->adap.flags, ATSI_AF_ALLOW_WORK );

    /* Quiesce worker thread activity - seems best to do this now since     */
    /* removing the SCSI host deregisters the transport, etc.               */

    atdl_log_dbg( hostadap, ATTO_LOG_INIT, "Quiescing work items\n" );

    starttime = atdf_curr_time_ms();

    while( hostadap->adap.ref_cnt )
    {
        atdf_cpu_yield();

        if( atdf_curr_time_ms() - starttime > 5000 )
        {
            atdl_log_err( hostadap, ATTO_LOG_INIT, "Failed to clear adapter references\n" );
            break;
        }
    }

    /* ...let the timer and DPC continue to run... */
}


/**
 * atdl_stop_services() - Stop the timer, DPC and worker
 */

static void atdl_stop_services( struct atdl_adap* hostadap )
{
    /* Stop the timer */

    atdl_log_info( hostadap, ATTO_LOG_INIT, "Stopping timer\n" );

    atdf_timer_stop( &hostadap->timer );

    /* Block future DPC scheduling */

    atdl_log_info( hostadap, ATTO_LOG_INIT, "Stopping DPC\n" );

    atdf_clear_bits( &hostadap->flags, ATDL_AF_ALLOW_DPC );
}


/**
 * atdl_free_adapter() - Free adapter resources
 */

static void atdl_free_adapter( struct atdl_adap* hostadap )
{
    struct atsi_adap*   adap = &hostadap->adap;
    struct pci_dev*     pcid = hostadap->adap.pcidev;
    unsigned long       lockflags;

    /* Stop the DPC and wait for it to stop executing */

    if( hostadap->dpc_work_q )
    {
        g_atgi_ops.destroy_workqueue( hostadap->dpc_work_q );
        hostadap->dpc_work_q = NULL;
    }

#if defined(FEAT_TRANSPORT)
    if( use_transport_layer )
    {
        /* Must be called after scsi_remove_host() */

        atdl_xport_adap_free( hostadap );
    }
#endif

    /* Free the core adapter */

    atci_free_adapter( hostadap->coreadap );

    /* Deregister PCI */

#ifdef FEAT_SRIOV
    if( adap->pci.vfcnt )
        g_atgi_ops.pci_disable_sriov( pcid );
#endif

    atdl_log_info( hostadap, ATTO_LOG_PCI,
                   "Calling pci_disable_device(), enable_cnt:%d\n",
                   pcid->enable_cnt.counter );

    pci_disable_device( pcid );

    atdl_log_info( hostadap, ATTO_LOG_INIT,
                   "pci_disable_device() done, enable_cnt:%d\n", pcid->enable_cnt.counter );

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_set_drv_data(%p, NULL) called\n", pcid );

    pci_set_drvdata( pcid, NULL );

    atci_dbg_str( " pci unregistered " );

    /* The core is fully cleaned up so we can clear this now */

    hostadap->coreadap = NULL;

    /* Free resources */

    if( hostadap->ioreq_pool )
        kmem_cache_destroy( hostadap->ioreq_pool );

    if( adap->scratch_buf )
        vfree( adap->scratch_buf );

    if( adap->cpu_to_msi_vect )
        kfree( adap->cpu_to_msi_vect );

    if( hostadap->msi_vectors )
        kfree( hostadap->msi_vectors );

    /* Remove the adapter list entry */

    spin_lock_irqsave( &g_atdl_vars.adap_lock, lockflags );
    g_atsi_vars.adap_list[hostadap->adap.scsihost->unique_id] = NULL;
    g_atsi_vars.adap_cnt--;
    spin_unlock_irqrestore( &g_atdl_vars.adap_lock, lockflags );

    atci_dbg_print( "Adapter killed, %d remaining ", g_atsi_vars.adap_cnt );
}


/**
 * atde_pcidev_probe_one() - pci_driver.probe entry point
 *
 * Returns: System error code
 */

static int atde_pcidev_probe_one( struct pci_dev* pcid, const struct pci_device_id* pcidid )
{
    struct scsi_host_template* sht = &g_atdl_scsi_templ;
    struct Scsi_Host*   sh = NULL;
    struct atdl_adap*   hostadap = NULL;
    struct atsi_adap*   adap;
    struct atcl_adap*   a;
    int                 err;
    int                 i;
    u32                 max_sect;
    unsigned long       lockflags;
    struct attohba_ops  hba_ops;
    char                dpc_name[64];

#ifdef FEAT_SRIOV
    if( pcid->is_virtfn
     && sriov_mode == ATSI_SRIOV_MODE_PF )
    {
        atto_log_err( ATTO_LOG_INIT,
                      "Skipping VF at %d:%d:%d.  sriov_mode %d disables driver loading on VFs\n",
                      pcid->bus->number,
                      (pcid->devfn >> 3) & 0x1F,
                      pcid->devfn & 7,
                      ATSI_SRIOV_MODE_PF );

        return -EACCES;
    }
#endif

    atto_log_info( ATTO_LOG_INIT,
                   "Initializing adapter at %d:%d:%d, PCI IDs 0x%02x 0x%02x 0x%02x 0x%02x\n",
                   pcid->bus->number,
                   (pcid->devfn >> 3) & 0x1F,
                   pcid->devfn & 7,
                   pcid->vendor,
                   pcid->device,
                   pcid->subsystem_vendor,
                   pcid->subsystem_device );

    max_sect = max_xfer_size / 512;

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,17,0)
    if( max_sect > 0xFFFF )
        max_sect = 0xFFFF;
#endif

    sht->max_sectors = max_sect;

    /* Allocate the SCSI host - log_level is not valid yet */

    if( !(sh = scsi_host_alloc( sht, sizeof(*hostadap)
                                     + atci_obj_size( ATSI_OBJ_ID_ADAPTER ) )) )
    {
        atto_log_err( ATTO_LOG_SCSI, "scsi_host_alloc() FAIL\n" );
        return -ENODEV;
    }

    hostadap = atdl_sh_to_hadap( sh );
    adap = &hostadap->adap;
    a = (struct atcl_adap*) (hostadap + 1);

    sprintf( adap->log_str, "host%d", sh->host_no );

    atdl_log_info( hostadap, ATTO_LOG_INIT, "Allocation OK (%p)\n", sh );

    /* Assign an adapter table entry */

    sh->unique_id = ATSI_MAX_ADAPTERS;

    spin_lock_irqsave( &g_atdl_vars.adap_lock, lockflags );

    for( i = 0; i < ATSI_MAX_ADAPTERS; i++ )
    {
        if( g_atsi_vars.adap_list[i] == NULL )
        {
            g_atsi_vars.adap_cnt++;
            g_atsi_vars.adap_list[i] = adap;
            sh->unique_id = i;
            break;
        }
    }

    spin_unlock_irqrestore( &g_atdl_vars.adap_lock, lockflags );

    if( sh->unique_id >= ATSI_MAX_ADAPTERS )
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT, "Max adapters reached.\n" );
        goto fail_free_host;
    }

    /* Maximum supported CDB length */

#if defined(ESASHBA4) || defined(ESASHBA5) || defined(ESASHBA6)
    sh->max_cmd_len     = 32;
#else
    sh->max_cmd_len     = 16;
#endif
#if defined(FEAT_TRANSPORT)
    /* The Kernel defaults to a blank template, so leave it be unless we    */
    /* are using the transport layer.                                       */

    if( use_transport_layer )
        sh->transportt  = g_atdl_vars.xport_templ;
#endif

    /* Initialize the SCSI host data */

    hostadap->hba_host.hostadap = hostadap;
    hostadap->coreadap = a;
    hostadap->flags = ATDL_AF_ALLOW_DPC;

    atdf_timer_init( &hostadap->timer, a, atci_timer_execute );
    sprintf( hostadap->rsrc_name, ATTO_DRVR_NAME "_%02d", sh->unique_id );

    /* Initialize the shared adapter */

    adap->pci.bus = pcid->bus->number;
    adap->pci.devfunc = pcid->devfn;
    adap->pci.dev = (pcid->devfn >> 3) & 0x1F;
    adap->pci.func = pcid->devfn & 7;
    adap->pci.domain = pci_domain_nr( pcid->bus );
    adap->pcidev = pcid;
    adap->coreadap = a;
    adap->hostadap = hostadap;
    adap->scsihost = sh;
    adap->hostnum = sh->host_no;
    adap->timer = &hostadap->timer;
    adap->intr_mode = ATSI_INTR_MODE_UNCLAIMED;
    adap->pci_id.ven_id = pcidid->vendor;
    adap->pci_id.dev_id = pcidid->device;
    adap->pci_id.ss_ven_id = pcidid->subvendor;
    adap->pci_id.ss_dev_id = pcidid->subdevice;
    adap->pci_id.class_code = pcidid->class;
    adap->pci_id.class_mask = pcidid->class_mask;
    adap->pci_id.rev_id = pcid->revision;

    /* General purpose memory lock that protects the scratch buffer and     */
    /* adap/bus reset counters                                              */

    spin_lock_init( &hostadap->mem_lock );

    adap->mem_lock = &hostadap->mem_lock;
    
    /* Initialize completion objects */

    init_completion( &hostadap->compl_shutdown );
    init_completion( &hostadap->compl_adap_rst );
    init_completion( &hostadap->compl_bus_rst );

    adap->compl_adap_rst = &hostadap->compl_adap_rst;
    adap->compl_bus_rst = &hostadap->compl_bus_rst;

    /* Initialize MSI-X vector map  */

    adap->cpu_to_msi_vect = kcalloc( num_possible_cpus(),
                                     sizeof(*adap->cpu_to_msi_vect),
                                     GFP_KERNEL );

    if( !adap->cpu_to_msi_vect )
    {
        atto_log_err( ATTO_LOG_INIT, "Unable to initialize vector map\n" );
        goto fail_free_adap;
    }

    /* Initialize core memory */

    if( atci_init_adap_mem( a, hostadap, adap, sh->unique_id ) )
    {
        atto_log_err( ATTO_LOG_INIT, "Unable to initialize core memory\n" );
        goto fail_free_adap;
    }

    /* Init the DPC */

    INIT_WORK( &hostadap->dpc_work_s, atde_dpc_work_q );
    sprintf( dpc_name, "%s-dpc", hostadap->rsrc_name );

    hostadap->dpc_work_q = g_atgi_ops.alloc_workqueue_w( dpc_name, 0, 1 );

    if( !hostadap->dpc_work_q )
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT, "Error allocating DPC workqueue\n" );
        goto fail_free_adap;
    }

    /* Create a lookaside cache for requests */

    hostadap->ioreq_pool = kmem_cache_create( ATTO_DRVR_NAME"_ioreq_pool",
                                              sizeof(struct atsi_req)
                                            + atci_obj_size( ATSI_OBJ_ID_REQUEST )
                                            + sizeof(struct atdl_req),
                                              8, 0, NULL );

    if( !hostadap->ioreq_pool )
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT, "Failed to allocate request pool!\n" );
        goto fail_free_adap;
    }

    /* Allocate scratch buffer */

    adap->scratch_len = 0x10000;
    adap->scratch_buf = vmalloc( adap->scratch_len );

    if( !adap->scratch_buf )
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT, "Failed to allocate scratch buffer!\n" );
        goto fail_free_adap;
    }

    /* Enable the PCI device */

    atdl_log_info( hostadap, ATTO_LOG_PCI,
                   "before pci_enable_device() enable_cnt:%d\n",
                   pcid->enable_cnt.counter );

    if( (err = pci_enable_device( pcid )) != 0 )
    {
        atdl_log_err( hostadap, ATTO_LOG_PCI, "pci_enable_device() FAIL (%d)\n", err );
        goto fail_free_adap;
    }

    atdl_log_info( hostadap, ATTO_LOG_PCI,
                   "after pci_device_enable() enable_cnt:%d\n",
                   pcid->enable_cnt.counter );

    /* Set to bus master for BIOSes that don't do it for us */

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_set_master() called\n" );
    pci_set_master( pcid );

    /* Set up 64 bit addressing or fall back to 32.  8Gb only supports 45 bits */

#if !defined(CELERITY4)
    if( atdl_init_dma_mask( hostadap, pcid, 64 ) )
#endif
    {
        if( atdl_init_dma_mask( hostadap, pcid, 32 ) )
        {
            atdl_log_err( hostadap, ATTO_LOG_INIT, "failed to set DMA mask, giving up.\n" );
            goto fail_free_pci_dev;
        }
    }

    /* Initialize the core adapter */

    if( atci_init_adap_operational( a ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT, "unable to initalize device at PCI bus %x:%x\n", pcid->bus->number, pcid->devfn );
        goto fail_free_pci_dev;
    }

    /* Now set the SCSI host properties from validated params */

    sh->max_channel     = 0;
    sh->max_id          = ATSI_MAX_ID;
    sh->max_lun         = ATSI_MAX_LUN;
#if defined(PROTO_SAS) || defined(PROTO_FC)
    sh->can_queue       = can_queue;
    sh->cmd_per_lun     = cmd_per_lun;
    /* Set sg_tablesize based off max_xfer_size */
    sh->sg_tablesize    = max_xfer_size / g_atsi_vars.page_size;
#endif

    sh->this_id = sh->max_id + 1;

#ifdef FEAT_CDEV_IOCTL
    /* Create the IOCTL handler */

    if( atdl_cdev_create( hostadap, sh->unique_id ) )
        atdl_log_err( hostadap, ATTO_LOG_INIT, "unable to create ioctl handler\n" );
#endif

    /* Add this channel as a SCSI host */

    atdl_log_info( hostadap, ATTO_LOG_SCSI, "scsi_add_host() called\n" );

    pci_set_drvdata( pcid, sh );

    err = scsi_add_host( sh, &pcid->dev );

    if( err )
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT, "scsi_add_host returned %d.\n", err );

        goto fail_free_pci_data;
    }

    adap->flags |= ATSI_AF_SCSI_HOST_ADDED;

    /* Connect to attocfg */

    memset( &hba_ops, 0, sizeof( hba_ops ) );

    hba_ops.attohba = &hostadap->hba_host;

    if( g_atgi_ops.register_host( &hba_ops,
                                  &sh->shost_gendev.kobj,
                                  &hostadap->cfg_host ) )
    {
        atdl_log_warn( hostadap, ATTO_LOG_INIT, "Failed to register host with attocfg.\n" );
    }

    /* Even if these fail, continue to load without the feature */

#if defined(FEAT_NVME_TGT)
    atdl_nvme_tgt_tport_create( hostadap, &pcid->dev );
#endif
#if defined(FEAT_NVME_INI)
    /* Note this requires the realized MSI-X vector count */

    atdl_nvme_ini_lport_create( hostadap, &pcid->dev );
#endif

#if defined(FEAT_TRANSPORT)
    if( use_transport_layer )
    {
        if( !atdl_xport_adap_init( hostadap ) )
            atdf_set_bits( &hostadap->flags, ATDL_AF_XPORT_STARTED );
    }
#endif

    /* Start adapter services (timer, DPC, etc.) */

    atdl_start_services( hostadap );

    sh->irq     = pcid->irq;
    sh->io_port = pcid->resource[0].start;

#ifdef FEAT_SRIOV
    /* Do the user a favor and enable the VFs specified in NVRAM */

    if( adap->pci.vfcnt )
        g_atgi_ops.pci_enable_sriov( pcid, adap->pci.vfcnt );
#endif

    return 0;

fail_free_pci_data:

    pci_set_drvdata( pcid, NULL );

fail_free_pci_dev:

    pci_disable_device( pcid );

fail_free_adap:

    atdf_set_bits( &adap->flags, ATSI_AF_TERMINATING );

#ifdef FEAT_CDEV_IOCTL
    /* Destroy the IOCTL handler */

    atdl_cdev_destroy( hostadap );
#endif

    atdl_stop_services( hostadap );

    atci_stop_adapter( hostadap->coreadap );

    atdl_free_adapter( hostadap );

fail_free_host:

    atdl_log_info( hostadap, ATTO_LOG_SCSI, "scsi_host_put() called\n" );
    scsi_host_put( sh );

    return -ENODEV;
}


/**
 * atde_pcidev_remode_one() - pci_driver.remove entry point
 *
 * Returns: System error code
 */

static void atde_pcidev_remove_one( struct pci_dev* pdev )
{
    struct Scsi_Host*   sh;
    struct atdl_adap*   hostadap;

    if( !pdev )
        return;

    sh = pci_get_drvdata( pdev );

    if( !sh )
    {
        /* This can happen if pci_set_drvdata was already called */
        /* to clear the host pointer.  If this is the case, we   */
        /* are okay; this channel has already been cleaned up.   */

        return;
    }

    hostadap = atdl_sh_to_hadap( sh );

    atdl_log_info( hostadap, ATTO_LOG_INIT, "Removing PCI dev %p\n", pdev );

    atdf_set_bits( &hostadap->adap.flags, ATSI_AF_TERMINATING );

    /* The shared /proc node must be cleaned up prior to releasing the last
     * host instance, otherwise a warning with stacktrace about leaking entries
     * is printed from scsi_host_dev_release() when cleaning up the /proc
     * directory.
     */

#ifdef FEAT_PROC_IOCTL
    if( g_atsi_vars.adap_cnt == 1 && g_atdl_vars.proc_major > 0 )
    {
        atdl_log_info( hostadap, ATTO_LOG_INIT, "unregister proc\n" );

        remove_proc_entry( ATTONODE_NAME, g_atdl_vars.proc_host->hostt->proc_dir );

        unregister_chrdev( g_atdl_vars.proc_major, ATTO_DRVR_NAME );

        g_atdl_vars.proc_major = 0;
    }
#endif

    /* Detach from attocfg */

    if( g_atgi_ops.unregister_host( hostadap->cfg_host ) )
        atdl_log_err( hostadap, ATTO_LOG_INIT, "Failed to unregister host with attocfg.\n" );

    /* Block new work */

    atdl_block_services( hostadap );

#if defined(FEAT_TRANSPORT)
    /* Remove the attached targets */

    if( hostadap->flags & ATDL_AF_XPORT_STARTED )
    {
        atdf_clear_bits( &hostadap->flags, ATDL_AF_XPORT_STARTED );

        atdl_log_info( hostadap, ATTO_LOG_XPORT, "Stopping transport layer\n" );

        /* Must be called before scsi_remove_host() */

        atdl_xport_adap_stop( hostadap );
    }
#endif

    /* Start NVMe teardown */

#if defined(FEAT_NVME_TGT)
    atdl_nvme_tgt_tport_destroy( hostadap );
#endif

#if defined(FEAT_NVME_INI)
    atdl_nvme_ini_lport_destroy( hostadap );
#endif

    /* Remove the host from the system */

    atdl_log_info( hostadap, ATTO_LOG_SCSI, "scsi_remove_host() called\n" );

    scsi_remove_host( sh );

#ifdef FEAT_CDEV_IOCTL
    /* Destroy the IOCTL handler */

    atdl_cdev_destroy( hostadap );
#endif

    /* Stop services */

    atdl_stop_services( hostadap );

    /* Shut down the hardware */

    atci_stop_adapter( hostadap->coreadap );

    /* Wait for NVMe teardown - both waits must be first before any subsequent cleanup */

#if defined(FEAT_NVME_INI)
    atdl_log_info( hostadap, ATTO_LOG_INIT, "Waiting for NVMe initiator termination\n" );

    wait_for_completion( &hostadap->compl_shutdown );
#endif
#if defined(FEAT_NVME_TGT)
    atdl_log_info( hostadap, ATTO_LOG_INIT, "Waiting for NVMe target termination\n" );

    wait_for_completion( &hostadap->compl_shutdown );
#endif

    /* Free resources */

#if defined(FEAT_NVME_TGT)
    atdl_nvme_tgt_tport_free( hostadap );
#endif
#if defined(FEAT_NVME_INI)
    atdl_nvme_ini_lport_free( hostadap );
#endif

    atdl_free_adapter( hostadap );

    atdl_log_info( hostadap, ATTO_LOG_SCSI, "scsi_host_put() called\n" );

    scsi_host_put( sh );

    /* The global cleanup after the last adapter is handled in _exit() */
}


#ifdef FEAT_SRIOV

/**
 * atde_sriov_configure() - Change the number of VFs
 */

static int atde_sriov_configure( struct pci_dev* pcidev, int numvfs )
{
    struct Scsi_Host*   sh = pci_get_drvdata( pcidev );
    struct atdl_adap*   hostadap;

    if( !sh )
        return -ENOMEM;

    hostadap = atdl_sh_to_hadap( sh );

    if( numvfs )
    {
        if( numvfs > hostadap->adap.pci.vfcnt )
            return -ERANGE;

        return g_atgi_ops.pci_enable_sriov( pcidev, numvfs );
    }

    g_atgi_ops.pci_disable_sriov( pcidev );

    return 0;
}

#endif /* FEAT_SRIOV */


#ifdef CONFIG_PM

/**
 * atde_pcidev_suspend() - Power management - quiesce the adapter
 * 
 * Note: These are a pain to debug!  By the time you are this far along into
 * power management, the serial port is probably shut down, so debug output
 * AND consoles redirected to ttySn are no longer working.  The only way 
 * to get KDB functional is to do something like:
 * 
 *    # sleep 3; sync; echo disk>/sys/power/state
 * 
 * ...and then quickly switch over to whatever the console
 * window is (eg. tty10).  Then if suspend or resume crashes,
 * you'll get a usable KDB prompt.
 * 
 * Also, check out the SER_DEFER stuff for potential way around this problem.
 */

static int atde_pcidev_suspend( struct pci_dev* pdev, pm_message_t state )
{
    struct Scsi_Host*   sh = pci_get_drvdata( pdev );
    struct atdl_adap*   hostadap;
    u32                 device_state;

    if( !sh )
        return 0;

    hostadap = atdl_sh_to_hadap( sh );

    atci_serdbg_cache();

    atdl_log_info( hostadap, ATTO_LOG_PCI, "suspending adapter()\n" );

    atdl_block_services( hostadap );
    atdl_stop_services( hostadap );

    atci_stop_adapter( hostadap->coreadap );

    device_state = pci_choose_state( pdev, state );

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_save_state() called\n" );

    pci_save_state( pdev );

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_disable_device() called\n" );

    pci_disable_device( pdev );

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_set_power_state() called\n" );

    pci_set_power_state( pdev, device_state );

    return 0;
}


static int atde_pcidev_resume( struct pci_dev* pdev )
{
    struct Scsi_Host*   sh = pci_get_drvdata( pdev );
    struct atdl_adap*   hostadap;
    int                 rslt;

    if( !sh )
        return 0;

    hostadap = atdl_sh_to_hadap( sh );

    atci_serdbg_init();

    atdl_log_info( hostadap, ATTO_LOG_PCI, "Resuming adapter()\n" );

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_set_power_state(PCI_D0) called\n" );

    pci_set_power_state( pdev, PCI_D0 );

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_enable_wake(PCI_D0, 0) called\n" );

    pci_enable_wake( pdev, PCI_D0, 0 );

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_restore_state() called\n" );

    pci_restore_state( pdev );

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_enable_device() called\n" );

    rslt = pci_enable_device( pdev );

    if( rslt )
        return rslt;

    pci_set_master( pdev );

    atdl_log_info( hostadap, ATTO_LOG_SCSI, "calling atci_init_adap_resume()\n" );

    atdf_set_bits( &hostadap->flags, ATDL_AF_ALLOW_DPC );

    if( atci_init_adap_resume( hostadap->coreadap ) )
        rslt = -EFAULT;

    atdl_log_info( hostadap, ATTO_LOG_PCI, "atde_pcidev_resume(): %d\n", rslt );

    atdl_start_services( hostadap );

    atci_serdbg_resume();

    return rslt;
}

#endif


/* PCI driver template */

static struct pci_driver g_atdl_pci_driver_templ = {
    .name     = ATTO_DRVR_NAME,
    .id_table = g_atdl_pci_id_table,
    .probe    = atde_pcidev_probe_one,
    .remove   = atde_pcidev_remove_one,
    .shutdown = atde_pcidev_remove_one,
#ifdef FEAT_SRIOV
    .sriov_configure = atde_sriov_configure,
#endif
#ifdef CONFIG_PM
    .suspend  = atde_pcidev_suspend,
    .resume   = atde_pcidev_resume,
#endif
};


/**
 * atdl_glob_init() - Initialize globals
 * 
 * Returns: 0 for success, non-zero for error
 */

static int atdl_glob_init( void )
{
    struct atsi_globals*    gs = &g_atsi_vars;
    struct atdl_globals*    gd = &g_atdl_vars;

    /* Clear everything first for safe cleanup on failure */

    memset( gd, 0, sizeof(*gd) );
    memset( gs, 0, sizeof(*gs) );

    /* Driver layer globals */

    g_atdl_scsi_templ.module = THIS_MODULE;

    spin_lock_init( &gd->adap_lock );
    spin_lock_init( &gd->dwq_lock );

    init_waitqueue_head( &gd->worker_waitq );

#if defined(FEAT_TRANSPORT)
    if( use_transport_layer )
    {
        gd->xport_templ = atdl_xport_templ_alloc();
        
        if( !gd->xport_templ )
        {
            atto_log_err( ATTO_LOG_XPORT, "Failed to attach to transport.\n" );
            return -1;
        }
    }
#endif

    /* Shared globals */

    gs->page_size = PAGE_SIZE;
    gs->kernel_ver = UTS_RELEASE;
    gs->node_cnt = num_online_nodes();
    gs->cpu_cnt = num_online_cpus();

    gs->dwq.first =
    gs->dwq.last =
    gs->dwq.head = (struct atsi_dev_work_item*) ((u8*) &gs->dwq.first
                            - offsetof( struct atsi_dev_work_item, next ));
    gs->dwq.lock = &gd->dwq_lock;

    /* Create a lookaside cache for device work items */

    gs->dwi_pool = kmem_cache_create( ATTO_DRVR_NAME"_dwi_pool",
                                     sizeof(struct atsi_dev_work_item),
                                     8, 0, NULL );

    if( !gs->dwi_pool )
    {
        atto_log_err( ATTO_LOG_INIT, "Unable to initialize globals!\n" );
        return -1;
    }

    /* Start the worker prior to any adapter arriving */

    atto_log_info( ATTO_LOG_INIT, "Starting worker\n" );

    gd->worker_thread = kthread_run( atde_worker_execute,
                                     NULL,
                                     ATTO_DRVR_NAME );

    return 0;
}


/**
 * atdl_glob_free() - Free globals
 */

static void atdl_glob_free( void )
{
    struct atsi_globals*    gs = &g_atsi_vars;
    struct atdl_globals*    gd = &g_atdl_vars;

    /* Stop the worker and wait for it to stop executing */

    if( gd->worker_thread )
    {
        atto_log_info( ATTO_LOG_INIT, "Stopping worker\n" );

        kthread_stop( gd->worker_thread );
        gd->worker_thread = NULL;
    }

#if defined(FEAT_TRANSPORT)
    if( gd->xport_templ )
    {
         atto_log_info( ATTO_LOG_XPORT, "Detaching transport layer template\n" );
         atdl_xport_templ_free( gd->xport_templ );
         gd->xport_templ = NULL;
    }
#endif

    if( gs->dwi_pool )
    {
        kmem_cache_destroy( gs->dwi_pool );
        gs->dwi_pool = NULL;
    }
}


/**
 * <driver>_init() - Module initialization
 * 
 * Returns: System error code
 */

static int __init ATTO_DRVR_SYM( init )( void )
{
    atto_log_info( ATTO_LOG_INIT, "module init called\n" );

    /* Initialize globals */

    if( atdl_glob_init() )
        goto fail_free_glob;

    /* Attach to attocfg driver */

    memset( &g_atgi_ops, 0, sizeof(g_atgi_ops) );

    g_atgi_ops.version = ATGI_CONN_VERSION;
    
    if( ATTO_CFG_SYM(connect)( &g_atgi_ops ) )
        goto fail_free_glob;

#if defined(FEAT_VSA) || defined(_MP_) || defined(FEAT_BAD_BLK_EVT)
    if( !g_atgi_ops.init_async_event_sockets() )
    {
        atto_log_warn( ATTO_LOG_INIT,
                       "attocfg failed to init async events, "
                       "your adapter may not work as expected.\n" );

#if defined(FEAT_VSA) || defined(_MP_)
        g_atgi_ops.fire_atto_event = NULL;
#endif
#ifdef FEAT_BAD_BLK_EVT
        g_atgi_ops.fire_bb_event = NULL;
#endif
    }
#endif

    if( atci_core_init() )
    {
        atto_log_err( ATTO_LOG_INIT, "Unable to initialize globals!\n" );
        goto fail_disconnect;
    }

#ifdef FEAT_CDEV_IOCTL
    if( atdl_cdev_rsrc_alloc() )
    {
        atto_log_err( ATTO_LOG_INIT, "Failed to allocate IOCTL resources.\n" );
        goto fail_free_core;
    }
#endif

    /* This will call the driver structure's probe method for each device */

    if( pci_register_driver( &g_atdl_pci_driver_templ ) != 0 )
        atto_log_err( ATTO_LOG_PCI, "pci_register_driver FAILED\n" );
    else
        atto_log_info( ATTO_LOG_PCI, "pci_register_driver() OK\n" );

    /* In-tree drivers don't seem to unload if there are no devices to control,
     * and unloading appears to be the reason Thunderbolt devices don't always
     * get assigned a driver instance (see IR17074 and IR17022 example 3).
     * Don't do this for HBAs for now, since it is unknown what the effects are
     * with the conflicting drivers.  (Is there a race for which one probes a
     * newly discovered device?)
     */

#if !defined(THUNDERBOLT)
    if( !g_atsi_vars.adap_cnt )
    {
        atto_log_warn( ATTO_LOG_INIT, "Driver will not be loaded because no ATTO " ATTO_DRVR_NAME " devices were found\n" );
#if defined(ESASHBA4)
        atto_log_warn( ATTO_LOG_INIT, "This can be caused by a conflict with the pm80xx kernel driver.  In order to use the " ATTO_DRVR_NAME " driver, the pm80xx driver must be unloaded.\n" );
#endif
        goto fail_unregister;
    }
    else
#endif  /* !THUNDERBOLT */
        atto_log_info( ATTO_LOG_INIT, "Found %d adapters.\n", g_atsi_vars.adap_cnt );

    return 0;

fail_unregister:

    atto_log_info( ATTO_LOG_PCI, "pci_unregister_driver() called\n" );
    pci_unregister_driver( &g_atdl_pci_driver_templ );

#if defined(FEAT_CDEV_IOCTL)
fail_free_core:
    atdl_cdev_rsrc_free();
#endif

    atci_core_free();

fail_disconnect:

    ATTO_CFG_SYM(disconnect)();

fail_free_glob:

    atdl_glob_free();

    return -1;
}


/**
 * <driver>_exit() - Module exit
 */

static void __exit ATTO_DRVR_SYM( exit )( void )
{
    atto_log_info( ATTO_LOG_INIT, "module exit\n" );

    /* pci_unregister_driver() will call atde_pcidev_remove_one() on each   */
    /* device.  Global resources are cleaned up afterwards.                 */

    atto_log_info( ATTO_LOG_PCI, "pci_unregister_driver() called\n" );

    pci_unregister_driver( &g_atdl_pci_driver_templ );

    ATTO_CFG_SYM(disconnect)();

#ifdef FEAT_CDEV_IOCTL
    atdl_cdev_rsrc_free();
#endif

    atci_core_free();

    atdl_glob_free();
}


/* e.g., celerity16fc_init */
DECLARE_MOD_ENTRY( init );

/* e.g., celerity16fc_exit */
DECLARE_MOD_ENTRY( exit );
