/****************************************************************************

DESCRIPTION: IOCTL functions

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2021
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#include "atdl.h"


/**
 * atdl_ioctl_handler() - Handle ATTO IOCTLs
 * 
 * @hostadap:   SCSI host data
 * @sdev:       SCSI device for device IOCTLs
 * @cmd:        Command number for the IOCTL
 * @userptr:    User buffer for the IOCTL data
 * 
 * Returns: System error code
 */

int atdl_ioctl_handler( struct atdl_adap* hostadap, struct scsi_device* sdev,
                        u32 cmd, void* userptr )
{
    int                     rslt;
    struct atdf_mapd_sgl    msgl;

    atdl_log_dbg( hostadap, ATTO_LOG_IOCTL, "ioctl (%x, %p)\n", cmd, userptr );

    atdf_msgl_init( &msgl, &hostadap->adap );

#ifdef DIRECT2GPU
    if( cmd == ATTO_IOCTL_D2GPU_MEM_PIN
     || cmd == ATTO_IOCTL_D2GPU_MEM_UNPIN )
        return atdl_d2gpu_ioctl_handler( cmd, userptr );
#endif

    if( userptr == NULL )
        return -EINVAL;

    if( cmd == ATTO_IOCTL_CMD_KRNL
     || cmd == ATTO_IOCTL_CMD_USER )
    {
        struct atioctl_hdr* ioctl;
        struct atioctl_hdr  hdr;

        /* Get and validate the IOCTL header */

        if( cmd == ATTO_IOCTL_CMD_KRNL )
            hdr = *(struct atioctl_hdr*) userptr;
        else
        {
            if( __copy_from_user( &hdr, userptr, sizeof(hdr) ) )
            {
                atdl_log_err( hostadap, ATTO_LOG_IOCTL, "Error copying IOCTL header\n" );
                return -EFAULT;
            }
        }

        if( hdr.version > ATIHDR_VER_0 )
        {
            atci_dbg_str( "Invalid version " );
            return -EINVAL;
        }

        if( strcmp( hdr.signature, ATIHDR_SIGNATURE ) )
        {
            atci_dbg_str( "Invalid signature " );
            return -EINVAL;
        }

        if( hdr.ioctlcmd == EXPRESS_IOCTL_GET_MOD_INFO
         || hdr.ioctlcmd == EXPRESS_IOCTL_GET_CHANNELS
         || hdr.ioctlcmd == EXPRESS_IOCTL_CHAN_INFO )
        {
            atci_dbg_str( "Invalid command " );
            return -EINVAL;
        }

        if( hdr.ioctlcmd == ATTO_IOCTL_CMD_SCSI_DEV_INFO
         && !sdev )
        {
            atci_dbg_str( "Device IOCTL sent to adapter " );
            return -EINVAL;
        }

        if( hdr.datalen == 0 )
        {
            atci_dbg_str( "Invalid data length " );
            return -EINVAL;
        }

        /* Service the IOCTL */

        if( cmd == ATTO_IOCTL_CMD_KRNL )
        {
            msgl.kern_addr = userptr;

            ioctl = userptr;

#if defined(HW_SCSI_TGT) || defined(HW_NVME_TGT)
            if( hdr.ioctlcmd == ATTO_IOCTL_CMD_FCTM
             || hdr.ioctlcmd == ATTO_IOCTL_CMD_SASTM )
            {
                rslt = atci_tgt_ioctl_handler( hostadap->coreadap, hdr.ioctlcmd, ioctl + 1 );

                if( rslt )
                    return -EINVAL;

                return 0;
            }
#endif
        }
        else
        {
            rslt = atdf_msgl_map( &msgl, userptr, sizeof(struct atioctl_hdr) + hdr.datalen );

            if( rslt )
                return rslt;

            ioctl = (struct atioctl_hdr*) (msgl.kern_addr + msgl.dma_offset);

            /* Remove the IOCTL header from the DMAable area */

            msgl.dma_offset += sizeof(struct atioctl_hdr);
        }

        if( hdr.ioctlcmd == ATTO_IOCTL_CMD_SCSI_DEV_INFO )
        {
            struct atioctl_scsi_dev_info* sdev_info;

            if( hdr.datalen < sizeof(*sdev_info) )
            {
                atci_dbg_str( "Invalid data length " );
                rslt = -EINVAL;
            }
            else
            {
                sdev_info = (struct atioctl_scsi_dev_info*) (ioctl + 1);

                sdev_info->host_num = hostadap->adap.hostnum;
                sdev_info->channel = atdl_sd_to_port_num( sdev );
                sdev_info->atto_targ_id = atdl_sd_to_targ_id( sdev );
                sdev_info->os_targ_id = sdev->id;
                sdev_info->lun = sdev->lun;

                rslt = 0;
            }
        }
        else
        {
            rslt = atci_ioctl_handler( hostadap->coreadap, hdr.ioctlcmd, ioctl + 1,
                                       hdr.datalen, &msgl );

            if( rslt != IOCTL_SUCCESS )
                rslt = -EINVAL;
        }

        if( cmd == ATTO_IOCTL_CMD_KRNL )
            msgl.kern_addr = NULL;
    }
    else
    {
        u32 size = atci_ioctl_legacy_size( cmd, userptr );

        if( !size )
            return -EINVAL;

        rslt = atdf_msgl_map( &msgl, userptr, size );

        if( rslt )
            return rslt;

        /* Call the IOCTL handler */

        rslt = atdl_ioctl_legacy( hostadap, cmd, &msgl );
    }

    atdf_msgl_unmap( &msgl );

    return rslt;
}


/**
 * atdl_ioctl_legacy() - Handle legacy ATTO IOCTLs
 * 
 * @hostadap:   Scsi host data
 * @cmd:        Command number for the IOCTL
 * @msgl:       User buffer for the IOCTL data
 * 
 * Returns: IOCTL or System error code
 */

int atdl_ioctl_legacy( struct atdl_adap* hostadap, u32 cmd, 
                       struct atdf_mapd_sgl* msgl )
{
#ifdef FEAT_LEGACY_IOCTL
    PEXPRESS_IOCTL_HEADER ioctl;
    
    ioctl = (PEXPRESS_IOCTL_HEADER) (msgl->kern_addr + msgl->dma_offset);

    /* Verify the Signature */

    if( memcmp( ioctl->Signature,
                EXPRESS_IOCTL_SIGNATURE,
                EXPRESS_IOCTL_SIGNATURE_SIZE ) )
    {
        atci_dbg_str( "INVALID SIGNATURE " );
        return -EINVAL;
    }

    /* Get the channel, if requested */

    if( ioctl->Channel != 0xFF
     && cmd != EXPRESS_IOCTL_GET_CHANNELS )
    {
        struct atsi_adap* adap = atsi_get_adap( ioctl->Channel );

        if( !adap )
        {
            ioctl->ReturnCode = IOCTL_BAD_CHANNEL;
            atci_dbg_str( "IOCTL_BAD_CHANNEL " );
            return 0;
        }

        hostadap = adap->hostadap;
    }

    /* Remove the IOCTL header from the DMAable area */

    msgl->dma_offset += sizeof(EXPRESS_IOCTL_HEADER);

    ioctl->ReturnCode = atci_ioctl_handler( hostadap->coreadap, cmd, ioctl + 1, 0, msgl );

    return ioctl->ReturnCode;
#else
    return -EINVAL;
#endif /* FEAT_LEGACY_IOCTL */
}


#ifdef FEAT_CDEV_IOCTL

static struct class*    g_atdl_cdev_class = NULL;
static dev_t            g_atdl_cdev_dev_num = -1;


/**
 * atde_cdev_devnode() - provides devtmpfs for devices in the class
 * 
 * @dev:    Character device.
 * @mode:   Mode that can be used to verify access.
 *
 * Returns: Allocated string describing the devtmpfs structure.
 */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6,2,0)                         \
    || (IS_RHEL_RELEASE_OR_LATER(8,9) && IS_RHEL_RELEASE_BEFORE(9,0))   \
    || IS_RHEL_RELEASE_OR_LATER(9,3)
static char* atde_cdev_devnode( const struct device* dev, umode_t* mode )
#else
static char* atde_cdev_devnode( struct device* dev, umode_t* mode )
#endif
{
    return kasprintf( GFP_KERNEL, ATTO_DRVR_NAME "/%s", dev_name( dev ) );
}


/**
 * atde_cdev_uevent() - provides permissions for devices in the class
 * 
 * @dev:    Character device.
 * @mode:   Mode that can be used to verify access.
 *
 * Returns: Allocated string describing the devtmpfs structure.
 */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6,2,0)                         \
    || (IS_RHEL_RELEASE_OR_LATER(8,9) && IS_RHEL_RELEASE_BEFORE(9,0))   \
    || IS_RHEL_RELEASE_OR_LATER(9,3)
static int atde_cdev_uevent( const struct device* dev, struct kobj_uevent_env* env )
#else
static int atde_cdev_uevent( struct device* dev, struct kobj_uevent_env* env )
#endif
{
    char buf[16];

    sprintf( buf, "DEVMODE=%#o", 0666 );

    g_atgi_ops.add_uevent_var( env, buf );

    return 0;
}


/**
 * atdl_cdev_rsrc_alloc() - Allocate resources for cdev creation
 * 
 * Returns: System error code
 */

int atdl_cdev_rsrc_alloc( void )
{
    dev_t                           devnum;
    struct class*                   cls;
    int                             rslt;
    static struct lock_class_key    key;

    /* Allocate device number range */

    rslt = alloc_chrdev_region( &devnum, 0, ATSI_MAX_ADAPTERS, ATTO_DRVR_NAME );

    if( rslt )
    {
        atto_log_info( ATTO_LOG_INIT, "alloc_chrdev_region failed %d\n", rslt );
        return rslt;
    }

    g_atdl_cdev_dev_num = devnum;

    /* Allocate class */

    cls = g_atgi_ops.class_create_w( THIS_MODULE, ATTO_DRVR_NAME, &key );

    if( IS_ERR( cls ) )
    {
        rslt = PTR_ERR( cls );

        atto_log_info( ATTO_LOG_INIT, "class_create failed %d\n", rslt );
        goto fail;
    }

    cls->devnode = atde_cdev_devnode;
    cls->dev_uevent = atde_cdev_uevent;

    g_atdl_cdev_class = cls;

    return 0;

fail:
    atdl_cdev_rsrc_free();
    return rslt;
}


/**
 * atdl_cdev_rsrc_free() - Free cdev resources
 */

void atdl_cdev_rsrc_free( void )
{
    if( g_atdl_cdev_class )
        g_atgi_ops.class_destroy( g_atdl_cdev_class );

    if( g_atdl_cdev_dev_num != -1 )
        unregister_chrdev_region( g_atdl_cdev_dev_num, ATSI_MAX_ADAPTERS );
}


/**
 * atde_cdev_open() - file_operations.open entry point
 * 
 * @inode:  Device inode associated with this character device.
 * @file:   File pointer for this device.
 *
 * Returns: System error code
 */

static int atde_cdev_open( struct inode* inode, struct file* file )
{
    file->private_data = container_of( inode->i_cdev, struct atdl_adap, ioctl_cdev );

    return 0;
}


/**
 * atde_cdev_ioctl() - file_operations.unlocked_ioctl entry point
 * 
 * @file:   File pointer for this device.
 * @cmd:    IOCTL command.
 * @arg:    Userspace ioctl data structure.
 *
 * Returns: System error code
 */

static long atde_cdev_ioctl( struct file* file, unsigned int cmd, unsigned long arg )
{
    struct atdl_adap* hostadap = file->private_data;
    
    return atdl_ioctl_handler( hostadap, NULL, cmd, (void*) arg );
}


/**
 * atde_cdev_read() - file_operations.read entry point.  Substitute for procfs
 * when not available (CONFIG_SCSI_PROC_FS or CONFIG_PROC_FS not defined).
 * 
 * @file:   File pointer for this device.
 * @buffer: Destination buffer
 * @len:    Destination buffer length
 * @offset: Offset into data to read
 *
 * Returns: Positive length transferred, 0 for end of file, or system error code
 */

static ssize_t atde_cdev_read( struct file *file, char __user *buffer, size_t len, loff_t *offset )
{
    struct atdl_adap*   hostadap = file->private_data;
    unsigned long       lock_flags;
    u32                 src_len;

    spin_lock_irqsave( &hostadap->mem_lock, lock_flags );

    src_len = atci_proc_read( hostadap->coreadap );

    if( *offset >= src_len )
        len = 0;
    else
    {
        if( src_len - *offset < len )
            len = src_len - *offset;

        if( __copy_to_user( buffer, &hostadap->adap.scratch_buf[*offset], len ) )
            len = -EFAULT;
        else
            *offset += len;
    }

    spin_unlock_irqrestore( &hostadap->mem_lock, lock_flags );

    return len;
}


/**
 * atde_cdev_write() - file_operations.write entry point.  Substitute for procfs
 * when not available (CONFIG_SCSI_PROC_FS or CONFIG_PROC_FS not defined).
 * 
 * @file:   File pointer for this device.
 * @buffer: Source buffer
 * @len:    Source buffer length
 * @offset: Offset into data to write
 *
 * Returns: Positive length transferred, 0 for none, or system error code
 */

static ssize_t atde_cdev_write( struct file* file, const char* input, size_t len, loff_t* offset )
{
    struct atdl_adap*   hostadap = file->private_data;
    char                buf[80];

    if( *offset
     || len > sizeof(buf)
     || len == 0 )
        return -EINVAL;

    if( __copy_from_user( buf, input, len ) )
        len = -EFAULT;
    else
    {
        len = atci_proc_write( hostadap->coreadap, buf, len );

        if( len == 0 )
            len = -EINVAL;
    }

    return len;
}


/* Character device file operations */

static const struct file_operations g_atdl_cdev_fops = {
    .owner          = THIS_MODULE,
    .open           = atde_cdev_open,
    .read           = atde_cdev_read,
    .unlocked_ioctl = atde_cdev_ioctl,
    .write          = atde_cdev_write,
};


/**
 * atdl_cdev_create() - Create character device for servicing IOCTLs
 * 
 * Returns: System error code
 */

int atdl_cdev_create( struct atdl_adap* hostadap, u32 instance )
{
    int             major = MAJOR( g_atdl_cdev_dev_num );
    dev_t           devnum = MKDEV( major, instance );
    struct device*  dev;
    int             rslt;
    char            name[32];

    /* Initialize and add the character device */

    cdev_init( &hostadap->ioctl_cdev, &g_atdl_cdev_fops );

    hostadap->ioctl_cdev.owner = THIS_MODULE;

    rslt = cdev_add( &hostadap->ioctl_cdev, devnum, 1 );

    if( rslt )
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT, "cdev_add failed %d\n", rslt );
        return rslt;
    }

    /* Create a /dev entry */

    sprintf( name, "%d", instance );

    dev = g_atgi_ops.device_create( g_atdl_cdev_class, NULL, devnum, NULL, name );

    if( IS_ERR( dev ) )
    {
        rslt = PTR_ERR( dev );

        atdl_log_err( hostadap, ATTO_LOG_INIT, "device_create failed %d\n", rslt );

        cdev_del( &hostadap->ioctl_cdev );

        return rslt;
    }

    hostadap->ioctl_dev = dev;

    return 0;
}


/**
 * atdl_cdev_destroy() - Destroy character device for servicing IOCTLs
 */

void atdl_cdev_destroy( struct atdl_adap* hostadap )
{
    if( hostadap->ioctl_dev )
    {
        g_atgi_ops.device_unregister( hostadap->ioctl_dev );

        cdev_del( &hostadap->ioctl_cdev );
    }
}


#endif /* FEAT_CDEV_IOCTL */

