/****************************************************************************

DESCRIPTION: SCSI initiator mode

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2021
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#include "atdl.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,3,0)
#include <scsi/scsi_dh.h>
#endif


#if LINUX_VERSION_CODE < KERNEL_VERSION(5,15,136) && !IS_RHEL_RELEASE_OR_LATER(9,1) && !IS_SLES_RELEASE_OR_LATER(15,5,0)
static inline void scsi_done(struct scsi_cmnd* cmd)
{
    cmd->scsi_done(cmd);
}
#endif


/**
 * atdi_scsi_req_move_data() - Move data in/out of request's data buffer
 * 
 * Returns: Transferred length, 0 for error
 */

u32 atdi_scsi_req_move_data( struct atsi_req* req, void* buf, u32 buf_len, u32 offset, int read )
{
    struct atdl_adap*   hostadap = req->hostadap;
    struct scsi_cmnd*   cmd = req->scsi.parent;
    struct scatterlist* sgl = scsi_sglist( cmd );
    int                 sge_cnt = scsi_sg_count( cmd );
    u32                 req_len = scsi_bufflen( cmd );
    u32                 len;
    size_t              rslt;

    if( sge_cnt == 0
     || buf_len == 0
     || offset >= req_len )
    {
        atdl_log_err( hostadap, ATTO_LOG_SCSI, "Invalid move data request (%u,%u,%u)\n",
                      sge_cnt, offset, buf_len );
        return 0;
    }

    len = buf_len;

    if( offset + len > req_len )
        len = req_len - offset;

    if( read )
        rslt = sg_pcopy_to_buffer( sgl, sge_cnt, buf, len, offset );
    else
        rslt = sg_pcopy_from_buffer( sgl, sge_cnt, buf, len, offset );

    if( rslt != len )
    {
        atdl_log_err( hostadap, ATTO_LOG_SCSI, "Move request data failed (%u,%u)\n",
                      len, (u32) rslt );
        return 0;
    }

    return rslt;
}


/**
 * atdl_scsi_dev_upd_state() - Update the state for an existing device.
 * 
 * This function uses scsi_device so it can be used with/without the
 * transport layer.
 */ 

void atdl_scsi_dev_upd_state( struct atdl_adap* hostadap,
                              struct scsi_device* sdev,
                              enum atsi_dev_action action )
{
    if( !sdev )
    {
        atdl_log_info( hostadap, ATTO_LOG_DISC, "No device in atdl_scsi_dev_upd_state\n" );
        return;
    }

    atdl_log_info( hostadap, ATTO_LOG_DISC,
                   "%s target %u, LUN %u\n",
                   atci_dev_action_str( action ),
                   atdl_sd_to_targ_id( sdev ),
                   (int) sdev->lun );

    switch( action )
    {
    case ATSI_DEV_ACT_ONLINE:

        scsi_device_set_state( sdev, SDEV_RUNNING );
        break;

    case ATSI_DEV_ACT_OFFLINE:

        scsi_device_set_state( sdev, SDEV_OFFLINE );
        break;

    case ATSI_DEV_ACT_RESCAN:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6,5,7) || IS_SLES_RELEASE_OR_LATER(15,6,0) \
|| (IS_DEB_RELEASE(11,0,0) && LINUX_VERSION_CODE >= KERNEL_VERSION(6,1,0))
        scsi_rescan_device( sdev );
#else
        scsi_rescan_device( &sdev->sdev_gendev );
#endif
        break;

    case ATSI_DEV_ACT_RELOCATE:
    case ATSI_DEV_ACT_REACTIVATE:
    case ATSI_DEV_ACT_ACTIVATE:

        if( sdev->sdev_state != SDEV_RUNNING )
            scsi_device_set_state( sdev, SDEV_RUNNING );

        break;

    case ATSI_DEV_ACT_DEACTIVATE:
        break;

    default:

        atdl_log_info( hostadap, ATTO_LOG_DISC, "Unhandled device action: %d\n", action );
        break;
    }
}


/**
 * atdi_scsi_dev_upd_state() - Update the state for an existing device.
 * 
 * This function uses scsi_device so it can be used with/without the
 * transport layer.
 */ 

void atdi_scsi_dev_upd_state( struct atdl_adap* hostadap, struct atsi_req* req,
                              enum atsi_dev_action action )
{
    if( req->scsi.parent )
        atdl_scsi_dev_upd_state( hostadap, req->scsi.parent->device, action );
}


/**
 * atdl_scsi_dev_set_state() - Set the state for a logical unit.
 * 
 * May create/delete the device or simply updates the state if it already
 * exists.  Note this function cannot be used when the transport layer is
 * enabled.
 */ 

int atdl_scsi_dev_set_state( struct atsi_dev_work_op* dwo, u16 lun, int verbose )
{
    struct atdl_adap*   hostadap = dwo->adap->hostadap;
    struct Scsi_Host*   sh = dwo->adap->scsihost;
    int                 rslt = 0;

    /* NB: This acquires the sh->host_lock spinlock */

    struct scsi_device* sdev = scsi_device_lookup( sh, 0, dwo->targ_id, lun );

    switch( dwo->action )
    {
    case ATSI_DEV_ACT_ADD:

        if( sdev )
        {
            rslt = -EEXIST;

            if( verbose )
            {
                atdl_log_err( hostadap, ATTO_LOG_DISC,
                              "Attempt to add existing device (%u,%u)\n",
                              dwo->targ_id, lun );
            }
        }
        else
        {
            rslt = scsi_add_device( sh, dwo->port_num, dwo->targ_id, lun );

            if( !rslt )
                atdl_log_info( hostadap, ATTO_LOG_DISC, "Created target %u, LUN %u\n", dwo->targ_id, lun );
            else if( verbose )
            {
                atdl_log_err( hostadap, ATTO_LOG_DISC,
                              "scsi_add_device(%u,%u) failed with %d\n",
                              dwo->targ_id, lun, rslt );
            }
        }

        break;

    case ATSI_DEV_ACT_REMOVE:

        if( !sdev )
        {
            rslt = -ENODEV;

            if( verbose )
            {
                atdl_log_err( hostadap, ATTO_LOG_DISC,
                              "Attempt to remove non-existent device (%u,%u)\n",
                              dwo->targ_id, lun );

            }
        }
        else
        {
            atdl_scsi_dev_upd_state( hostadap, sdev, ATSI_DEV_ACT_OFFLINE );

            scsi_remove_device( sdev );

            atdl_log_info( hostadap, ATTO_LOG_DISC, "Destroyed target %u, LUN %u\n", dwo->targ_id, lun );
        }

        break;

    default:

        if( !sdev )
            rslt = -ENODEV;

        atdl_scsi_dev_upd_state( hostadap, sdev, dwo->action );
        break;
    }

    if( sdev )
        scsi_device_put( sdev );

    return rslt;
}


/**
 * atdi_scsi_dev_set_state() - Set the state for a logical unit.
 */ 

int atdi_scsi_dev_set_state( struct atsi_dev_work_op* dwo, u16 lun )
{
    return atdl_scsi_dev_set_state( dwo, lun, TRUE );
}


/**
 * atdi_scsi_host_in_recovery() - Determine if the SCSI host is in recovery.
 * 
 * Returns: Boolean
 */ 

int atdi_scsi_host_in_recovery( struct Scsi_Host* sh )
{
    return sh->shost_state == SHOST_RECOVERY
        || sh->shost_state == SHOST_CANCEL_RECOVERY
        || sh->shost_state == SHOST_DEL_RECOVERY;
}


/**
 * atdi_scsi_bus_reset_detected() - Post bus reset processing
 */

void atdi_scsi_bus_reset_detected( struct atdl_adap* hostadap )
{
    /* Only notify a registered host */

    if( hostadap->adap.flags & ATSI_AF_SCSI_HOST_ADDED )
    {
        atdl_log_info( hostadap, ATTO_LOG_SCSI, "scsi_report_bus_reset() called\n" );
        scsi_report_bus_reset( hostadap->adap.scsihost, 0 );
    }
}


/**
 * atde_show_info() - Handle reading /proc info.
 */

static int atde_show_info( struct seq_file* m, struct Scsi_Host* sh )
{
    struct atdl_adap*   hostadap = atdl_sh_to_hadap( sh );
    u32                 dest_len = m->size - m->count;
    u32                 src_len;
    u32                 len;
    unsigned long       lock_flags;

    spin_lock_irqsave( &hostadap->mem_lock, lock_flags );

    len = src_len = atci_proc_read( hostadap->coreadap );

    if( len > dest_len )
        len = dest_len;

    seq_write( m, hostadap->adap.scratch_buf, len );

    spin_unlock_irqrestore( &hostadap->mem_lock, lock_flags );

    atdl_log_info( hostadap, ATTO_LOG_DEBUG, "show_info (%d, %d)\n",
                   src_len, dest_len );

    return 0;
}


/**
 * atde_write_info() - Handle writing /proc info.
 */

static int atde_write_info( struct Scsi_Host* sh, char* buf, int len )
{
    struct atdl_adap* hostadap = atdl_sh_to_hadap( sh );

    atdl_log_info( hostadap, ATTO_LOG_DEBUG, "write_info (%*s, %d)\n", len, buf, len );

    if( atci_proc_write( hostadap->coreadap, buf, len ) )
        return len;

    return -EINVAL;
}


#ifdef FEAT_PROC_IOCTL

/**
 * atde_attonode_ioctl() - Handle ioctl calls to "/proc/scsi/<driver>/ATTONode"
 */

long atde_attonode_ioctl( struct file* fp, unsigned int cmd, unsigned long arg )
{
    return atdl_ioctl_handler( atdl_sh_to_hadap( g_atdl_vars.proc_host ), NULL, cmd, (void*) arg );
}


/* ATTONode proc operations */

static struct file_operations g_atdl_attonode_fops = {
    .compat_ioctl   = atde_attonode_ioctl,
    .unlocked_ioctl = atde_attonode_ioctl,
};


#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,6,0)
static struct proc_ops g_atdl_attonode_ops = {
    .proc_ioctl = atde_attonode_ioctl,
};
#endif

#endif /* FEAT_PROC_IOCTL */


/**
 * atde_info() - Get a string describing the adapter
 * 
 * Returns: Pointer to retrieved string
 */

static const char* atde_info( struct Scsi_Host* sh )
{
    struct atdl_adap*       hostadap = atdl_sh_to_hadap( sh );
    static char             atdl_info_str[128];
#ifdef FEAT_PROC_IOCTL
    struct atdl_globals*    gd = &g_atdl_vars;

    /* If we haven't done so already, register as a char driver */
    /* and stick a node under "/proc/scsi/<driver>/ATTONode"    */

    if( gd->proc_major <= 0 )
    {
        gd->proc_host = sh;
        gd->proc_major = register_chrdev( 0, ATTO_DRVR_NAME, &g_atdl_attonode_fops );

        atdl_log_info( hostadap, ATTO_LOG_DEBUG, "register_chrdev (major %d)\n", gd->proc_major );

        if( gd->proc_major > 0 )
        {
            struct proc_dir_entry* pde;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,6,0)
            pde = proc_create( ATTONODE_NAME, 0, sh->hostt->proc_dir, &g_atdl_attonode_ops );
#else
            pde = proc_create( ATTONODE_NAME, 0, sh->hostt->proc_dir, &g_atdl_attonode_fops );
#endif
            if( !pde )
            {
                atdl_log_warn( hostadap, ATTO_LOG_DEBUG, "failed to create_proc_entry\n" );
                gd->proc_major = -1;
            }
        }
    }
#endif

    sprintf( atdl_info_str,
             "%s, PCI address %d:%d:%d, Driver " AV_FULL_VERSION_STR,
             hostadap->adap.model_name,
             hostadap->adap.pci.bus,
             hostadap->adap.pci.dev,
             hostadap->adap.pci.func );

    return atdl_info_str;
}


/**
 * atde_ioctl() - Handles IOCTL calls to the adapter's children
 * 
 * @sdev:   Scsi Device to handle the IOCTL
 * @cmd:    Command number for the IOCTL
 * @arg:    Raw buffer for the IOCTL data
 * 
 * Returns: 0 on success, negative errno on failure
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,1,0)
static int atde_ioctl( struct scsi_device* sdev, int cmd, void* arg )
#else
static int atde_ioctl( struct scsi_device* sdev, unsigned int cmd, void* arg )
#endif
{
#ifdef DIRECT2GPU
    if( cmd == ATTO_IOCTL_D2GPU_IO )
        return atdl_d2gpu_ioctl_io( sdev, arg );
#endif

#ifdef PROTO_ATA
    if( cmd >= 0x301
     && cmd <= 0x33F )
    {
        if( arg == NULL
         && cmd != HDIO_SET_MULTCOUNT
         && cmd != HDIO_SET_PIO_MODE
         && cmd != HDIO_DRIVE_RESET )
            return -EINVAL;

        return atdl_ata_ioctl_handler( sdev, cmd, arg );
    }
#endif

    return atdl_ioctl_handler( atdl_sh_to_hadap( sdev->host ), sdev, cmd, arg );
}


/**
 * atdl_dma_unmap() - Unmap the command data buffer
 */

static inline void atdl_dma_unmap( struct atsi_req* req, struct scsi_cmnd* cmd )
{
#ifdef BRCM_SAS
        cmd->sc_data_direction = req->scsi.dma_dir_map;
#endif
        scsi_dma_unmap( cmd );

#ifdef BRCM_SAS
        cmd->sc_data_direction = req->scsi.dma_dir_cmd;
#endif
}


/**
 * atde_queuecommand() - Start a new SCSI command.
 */

static int atde_queuecommand( struct Scsi_Host* host, struct scsi_cmnd* cmd )
{
    struct atdl_adap*   hostadap = atdl_sh_to_hadap( host );
    struct atsi_req*    req;
    int                 use_sg = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,15,0) && !IS_SLES_RELEASE_OR_LATER(15,5,0)
    struct request*     blk_rq = cmd->request;
#else
    struct request*     blk_rq = scsi_cmd_to_rq( cmd );
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0) && !IS_RHEL_RELEASE_OR_LATER(8,0)
    int cpu_id = (blk_rq->cpu < 0) ? atdf_cpu_get() : blk_rq->cpu;
#else
    int cpu_id = blk_mq_rq_cpu( blk_rq );
#endif

#if defined(FEAT_TRANSPORT) && defined(PROTO_FC)
    if( hostadap->flags & ATDL_AF_XPORT_STARTED )
    {
        struct fc_rport*    rport = starget_to_rport( scsi_target( cmd->device ) );
        int                 err;

        /* If initiator mode is disabled, the remote port may be invalid */

        if( !rport )
            err = DID_NO_CONNECT << 16;
        else
        {
            err = fc_remote_port_chkready( rport );

            if( err )
                atdl_log_info( hostadap, ATTO_LOG_DEBUG, "fc rport not ready. failing command\n" );
        }

        if( err )
        {
            cmd->result = err;
            scsi_done( cmd );

            return 0;
        }
    }
#endif /* FEAT_TRANSPORT && PROTO_FC */

#if defined(PROTO_SAS) || defined(PROTO_FC)
    /* Override allowed retry count */

    if( cmd_retry_count != 0xFFFFFFFF )
        cmd->allowed = cmd_retry_count;
#endif

    /* Allocate and init a request */

    req = atdl_req_alloc( hostadap, cpu_id );

    req->scsi.parent = cmd;
    req->scsi.dev_type = cmd->device->type;
    req->timeout = jiffies_to_msecs( blk_rq->timeout );
    req->data_len = cmd->sdb.length;
    req->sgl = cmd->sdb.table.sgl;
    req->sge_cnt = cmd->sdb.table.nents;
    req->targ_id = atdl_sd_to_targ_id( cmd->device );
    req->scsi.lun = (u16) cmd->device->lun;
    req->scsi.sense_buf = cmd->sense_buffer;
    req->scsi.sense_len = SCSI_SENSE_BUFFERSIZE;
    req->scsi.cdb = cmd->cmnd;
    req->scsi.cdb_len = cmd->cmd_len;

    if( cmd->sdb.length )
    {
        if( cmd->sc_data_direction == DMA_FROM_DEVICE )
            req->flags |= ATSI_RF_DATA_DIR_READ;
        else if( cmd->sc_data_direction == DMA_TO_DEVICE )
            req->flags |= ATSI_RF_DATA_DIR_WRITE;

#ifdef BRCM_SAS
        req->scsi.dma_dir_cmd =
        req->scsi.dma_dir_map = cmd->sc_data_direction;

        /* Some firmware versions byte-swap the REPORT ZONES command reply from
        zoned devices by directly accessing in the host buffer. This does
        not respect the default command DMA direction and causes IOMMU page
        faults on some architectures (e.g. AMD hosts).  Avoid such issues by
        making the report zones buffer mapping bi-directional.
        */

        if( cmd->cmnd[0] == 0x95 /* ZBC_IN */
         && cmd->cmnd[1] == 0x00 /* ZI_REPORT_ZONES */ )
        {
            req->scsi.dma_dir_map =
            cmd->sc_data_direction = DMA_BIDIRECTIONAL;
        }
#endif

        use_sg = scsi_dma_map( cmd );

#ifdef BRCM_SAS
        cmd->sc_data_direction = req->scsi.dma_dir_cmd;
#endif
        if( use_sg < 0 )
            goto Busy;

        if( use_sg )
            req->flags |= ATSI_RF_DMA_MAPPING;
    }

    /* Forward it to the core */

    if( !atci_scsi_req_start( hostadap->coreadap, req->rq, req ) )
    {
        /* All good */

        return 0;
    }

    /* Busy, unmap the SGL if present */

    if( use_sg )
        atdl_dma_unmap( req, cmd );

Busy:
    atdl_req_free( req );

    return SCSI_MLQUEUE_HOST_BUSY;
}


/**
 * atdi_scsi_req_complete() - Complete a SCSI command to Linux.
 */

void atdi_scsi_req_complete( struct atsi_req* req, int scsi_stat, u8 hw_sts, int resid )
{
    struct scsi_cmnd*   cmd = req->scsi.parent;
    int                 sts;

    if( likely( req->flags & ATSI_RF_DMA_MAPPING ) )
        atdl_dma_unmap( req, cmd );

    if( likely( hw_sts == HWR_STS_SUCCESS ) )
        sts = DID_OK;
    else
    {
        /* Sometimes the HW layer fills in the sense buffer early, and then */
        /* returns something other than CHECK CONDITION.  Linux always looks*/
        /* at the sense buffer anyway.                                      */

        if( hw_sts != HWR_STS_SCSI_ERROR )
            memset( cmd->sense_buffer, 0, SCSI_SENSE_BUFFERSIZE );

        switch( hw_sts )
        {
            /* NOTE: Linux wants a good status for a SCSI error, because   */
            /* it will check the scsi_stat value in the completion anyway. */

            case HWR_STS_DATA_OVERRUN:
            case HWR_STS_DATA_UNDERRUN:
            case HWR_STS_SUCCESS:
            case HWR_STS_SCSI_ERROR:
                sts = DID_OK;
                break;

            case HWR_STS_TARG_NOT_PRESENT:
            case HWR_STS_TARG_INV:
                sts = DID_NO_CONNECT;
                break;

            case HWR_STS_BUS_RESET:
                sts = DID_RESET;
                break;

            case HWR_STS_ABORTED:
                sts = DID_ABORT;
                break;

            case HWR_STS_TIMEOUT:
                sts = DID_TIME_OUT;
                break;

            case HWR_STS_BUSY:
                sts = DID_BUS_BUSY;
                break;

            default:
                sts = DID_ERROR;
                break;
        }

        scsi_set_resid( cmd, resid );
    }

    /* Free the request first so it is available for the next command started */

    atdl_req_free( req );

    /* Complete the command to the OS.  Note the status is always set now   */
    /* since Linux may set DID_TIME_OUT during the time the eh_timed_out    */
    /* handler is called.  Bad Linux - you shouldn't be touching the object */
    /* while we have ownership.                                             */

    cmd->result = (sts << 16) | (scsi_stat & STATUS_MASK);
    scsi_done( cmd );
}


/**
 * atde_eh_timed_out() - Handle a timed out SCSI command
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6,2,0) || IS_RHEL_RELEASE_OR_LATER(9,3)
static enum scsi_timeout_action atde_eh_timed_out (struct scsi_cmnd *cmd)
{
    /* Our drivers contain timeouts and retries of all commands.            */
    /* Therefore, just return SCSI_EH_RESET_TIMER to indicate that           */
    /* we are handling this command.                                        */

    atdl_log_info( atdl_sh_to_hadap( cmd->device->host ), ATTO_LOG_SCSI,
                   "atde_eh_timed_out cmd:%p. SCSI_EH_RESET_TIMER\n", cmd );

    return SCSI_EH_RESET_TIMER;
}
#else
static enum blk_eh_timer_return atde_eh_timed_out (struct scsi_cmnd *cmd)
{
    /* Our drivers contain timeouts and retries of all commands.            */
    /* Therefore, just return BLK_EH_RESET_TIMER to indicate that           */
    /* we are handling this command.                                        */

    atdl_log_info( atdl_sh_to_hadap( cmd->device->host ), ATTO_LOG_SCSI,
                   "atde_eh_timed_out cmd:%p. BLK_EH_RESET_TIMER\n", cmd );

    return BLK_EH_RESET_TIMER;
}
#endif

/*
 * atdl_req_to_eh_sts() - Translate HW request status to Linux error handler values.
 *
 * Returns: SUCCESS, FAILED, or FAST_IO_FAIL
 */

static int atdl_req_to_eh_sts( u8 req_sts )
{
    switch( req_sts )
    {
        case HWR_STS_SUCCESS:
            return SUCCESS;

        case HWR_STS_TARG_NOT_PRESENT:
        case HWR_STS_TARG_INV:
            return FAST_IO_FAIL;

        default:
            return FAILED;
    }
}


/**
 * atde_eh_abort() - Abort an outstanding command
 */

static int atde_eh_abort( struct scsi_cmnd* cmd )
{
    struct atdl_adap*   hostadap = atdl_sh_to_hadap( cmd->device->host );
    u8                  req_sts;

    atdl_log_info( hostadap, ATTO_LOG_TASKM, "eh_abort (%p)\n", cmd );

    req_sts = atci_scsi_eh_abort( hostadap->coreadap, cmd );

    atdl_log_info( hostadap, ATTO_LOG_TASKM, "eh_abort (%p) result %d\n", cmd, req_sts );

    return atdl_req_to_eh_sts( req_sts );
}


/**
 * atde_eh_adap_reset() - Reset the adapter (error handler)
 */

static int atde_eh_adap_reset( struct scsi_cmnd* cmd )
{
    struct atdl_adap*   hostadap = atdl_sh_to_hadap( cmd->device->host );
    u8                  req_sts;

    atdl_log_info( hostadap, ATTO_LOG_TASKM, "host_reset (%p)\n", cmd );

    req_sts = atci_scsi_eh_adap_reset( hostadap->coreadap );

    atdl_log_info( hostadap, ATTO_LOG_TASKM, "host_reset (%p) result %d\n", cmd, req_sts );

    return atdl_req_to_eh_sts( req_sts );
}


/**
 * atde_eh_bus_reset() - Reset the bus (error handler)
 */

static int atde_eh_bus_reset( struct scsi_cmnd* cmd )
{
    struct atdl_adap*   hostadap = atdl_sh_to_hadap( cmd->device->host );
    u8                  req_sts;

    atdl_log_info( hostadap, ATTO_LOG_TASKM, "bus_reset (%p)\n", cmd );

    req_sts = atci_scsi_eh_bus_reset( hostadap->coreadap );

    atdl_log_info( hostadap, ATTO_LOG_TASKM, "bus_reset (%p) result %d\n", cmd, req_sts );

    return atdl_req_to_eh_sts( req_sts );
}


#ifndef _MP_

/**
 * atde_eh_lun_reset() - Reset a logical unit (error handler)
 */

static int atde_eh_lun_reset( struct scsi_cmnd* cmd )
{
    struct atdl_adap*   hostadap = atdl_sh_to_hadap( cmd->device->host );
    u8                  req_sts;

    atdl_log_info( hostadap, ATTO_LOG_TASKM, "device_reset (%p)\n", cmd );

    req_sts = atci_scsi_eh_lun_reset( hostadap->coreadap,
                                 atdl_sd_to_targ_id( cmd->device ),
                                 cmd->device->lun );

    atdl_log_info( hostadap, ATTO_LOG_TASKM, "device_reset (%p) result %d\n", cmd, req_sts );

    return atdl_req_to_eh_sts( req_sts );
}


/**
 * atde_eh_target_reset() - Reset a target (error handler)
 */

static int atde_eh_target_reset( struct scsi_cmnd* cmd )
{
    struct atdl_adap*   hostadap = atdl_sh_to_hadap( cmd->device->host );
    u8                  req_sts;

    atdl_log_info( hostadap, ATTO_LOG_TASKM, "target_reset (%p)\n", cmd );

    req_sts = atci_scsi_eh_target_reset( hostadap->coreadap,
                                    atdl_sd_to_targ_id( cmd->device ),
                                    cmd->device->lun );

    atdl_log_info( hostadap, ATTO_LOG_TASKM, "target_reset (%p) result %d\n", cmd, req_sts );

    return atdl_req_to_eh_sts( req_sts );
}

#endif  /* !_MP_ */


/**
 * atde_change_queue_depth() - Process a sysfs queue depth change for a logical unit
 * 
 * Also called internally when a logical unit is created.
 * 
 * Returns: The depth set (or closest supported value).
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,19,0)
static int atde_change_queue_depth( struct scsi_device* sdev, int depth, int reason )
#else
static int atde_change_queue_depth( struct scsi_device* sdev, int depth )
#endif
{
    struct atdl_adap* hostadap = atdl_sh_to_hadap( sdev->host );

    atdl_log_info( hostadap, ATTO_LOG_DEBUG, "Set target ID %d queue depth to %d\n",
                   atdl_sd_to_targ_id( sdev ), depth );

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,19,0)
    scsi_adjust_queue_depth( sdev, scsi_get_tag_type( sdev ), depth );
#else
    scsi_change_queue_depth( sdev, depth );
#endif

    return sdev->queue_depth;
}


/**
 * atde_change_queue_type() - Process a sysfs queue type change for a logical unit
 * 
 * Also called internally when a logical unit is created.
 * 
 * Returns: The type set (as requested or closest supported value)
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,19,0)

int atde_change_queue_type( struct scsi_device* sdev, int type )
{
    struct atdl_adap* hostadap = atdl_sh_to_hadap( sdev->host );

    atdl_log_info( hostadap, ATTO_LOG_DEBUG, "Set target ID %d queue type to %d\n",
                   atdl_sd_to_targ_id( sdev ), type );

    switch( type )
    {
        case MSG_SIMPLE_TAG:
        case MSG_ORDERED_TAG:

            if( sdev->tagged_supported )
                break;

            /* Fall through */

        default:
            type = 0;
    }

    scsi_set_tag_type( sdev, type );

    if( type )
        scsi_activate_tcq( sdev, sdev->queue_depth );
    else
        scsi_deactivate_tcq( sdev, sdev->queue_depth );

    return type;
}

#endif


/**
 * atde_slave_alloc() - Allocate resources prior to scanning a logical unit
 * 
 * Returns: 0 on success, -errno on error
 */

static int atde_slave_alloc( struct scsi_device* dev )
{
    struct atdl_adap* hostadap = atdl_sh_to_hadap( dev->host );

    atdl_log_info( hostadap, ATTO_LOG_DISC, "atde_slave_alloc()\n" );

#if defined (_MP_)

    /* If the scsi device handler is not set, install a NULL one         */
    /* called 'proprietary' to prevent another one (alua, rdac, etc.)    */
    /* from being used and conflicting with our proprietary MP solution. */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,3,0)
    if( !dev->handler )
    {
        dev->handler = kmalloc( sizeof(*dev->handler), GFP_ATOMIC );

        if( dev->handler )
        {
            memset( dev->handler, 0, sizeof(*dev->handler) );

            dev->handler->name = "proprietary";

            atdl_log_info( hostadap, ATTO_LOG_DISC,
                           "proprietary device handler: %p\n",
                           dev->handler );
        }
    }
#else
    if( !dev->scsi_dh_data )
    {
        struct scsi_device_handler* dh = kmalloc( sizeof(*dh), GFP_ATOMIC );

        if( dh )
        {
            struct scsi_dh_data* dh_data = kmalloc( sizeof(*dh_data), GFP_ATOMIC );

            if( dh_data )
            {
                memset( dh, 0, sizeof( *dh ) );
                dh->name = "proprietary";

                memset( dh_data, 0, sizeof( *dh_data ) );
                dh_data->scsi_dh = dh;
                dev->scsi_dh_data = dh_data;

                atdl_log_info( hostadap, ATTO_LOG_DISC,
                               "proprietary scsi_dh_data: %p\n",
                               dev->scsi_dh_data );
            }
            else
                kfree( dh );
        }
    }
#endif
#endif /* defined(_MP_) */


#if defined(FEAT_TRANSPORT)
    if( hostadap->flags & ATDL_AF_XPORT_STARTED )
    {
        if( atdl_xport_slave_alloc( hostadap, dev ) )
            return -ENXIO;
    }
    else
#endif
        dev->hostdata = NULL;

    atdl_log_info( hostadap, ATTO_LOG_DISC, "slave_alloc OK\n" );

    return 0;
}


/**
 * atde_slave_configure() - Configure a newly discovered logical unit
 * 
 * Returns: 0 on success, -errno on error
 */

static int atde_slave_configure( struct scsi_device* sdev )
{
    struct atdl_adap* hostadap = atdl_sh_to_hadap( sdev->host );

    atdl_log_info( hostadap, ATTO_LOG_DISC, "atde_slave_configure()\n" );

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,19,0)
    /* Must be in this order */

    atde_change_queue_type( sdev, MSG_SIMPLE_TAG );
    atde_change_queue_depth( sdev, sdev->host->cmd_per_lun, 0 );
#else
    atde_change_queue_depth( sdev, sdev->host->cmd_per_lun );
#endif

    return 0;
}


/**
 * atde_slave_destroy() - Free logical unit resources
 * 
 * Returns: 0 on success, -errno on error
 */

static void atde_slave_destroy (struct scsi_device *dev)
{
    struct atdl_adap* hostadap = atdl_sh_to_hadap( dev->host );

    atdl_log_info( hostadap, ATTO_LOG_DISC, "atde_slave_destroy()\n" );

#if defined(_MP_)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,3,0)
    if( dev->handler
     && !strncmp( dev->handler->name, "proprietary", 11 ) )
    {
        kfree( dev->handler );
        dev->handler = NULL;
    }
#else
    if( dev->scsi_dh_data
     && dev->scsi_dh_data->scsi_dh
     && !strncmp( dev->scsi_dh_data->scsi_dh->name, "proprietary", 11 ) )
    {
        atdl_log_info( hostadap, ATTO_LOG_DISC, "proprietary scsi_dh_data free'd\n" );

        kfree( dev->scsi_dh_data->scsi_dh );
        kfree( dev->scsi_dh_data );

        dev->scsi_dh_data = NULL;
    }
#endif
#endif /* defined (_MP_) */

#if defined(FEAT_TRANSPORT)
    dev->hostdata = NULL;
#endif
}


/* SCSI host template */

struct scsi_host_template g_atdl_scsi_templ =
{
    queuecommand:            atde_queuecommand,
    module:                  THIS_MODULE,
    name:                    ATDL_ADAP_NAME,
    info:                    atde_info,
    ioctl:                   atde_ioctl,
    eh_abort_handler:        atde_eh_abort,
#ifndef _MP_
    eh_device_reset_handler: atde_eh_lun_reset,
    eh_target_reset_handler: atde_eh_target_reset,
#endif
    eh_bus_reset_handler:    atde_eh_bus_reset,
    eh_host_reset_handler:   atde_eh_adap_reset,
    slave_alloc:             atde_slave_alloc,
    slave_configure:         atde_slave_configure,
    slave_destroy:           atde_slave_destroy,
    change_queue_depth:      atde_change_queue_depth,
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,19,0)
    change_queue_type:       atde_change_queue_type,
#endif
    show_info:               atde_show_info,
    write_info:              atde_write_info,
    eh_timed_out:            atde_eh_timed_out,
    proc_name:               ATTO_DRVR_NAME,
    this_id:                 -1,
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0)
    use_clustering:          ENABLE_CLUSTERING,
#else
    dma_boundary:            PAGE_SIZE - 1,
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,16,0) && !IS_SLES_RELEASE_OR_LATER(15,5,0)
    shost_attrs:             g_atdl_host_attrs,
#else
    shost_groups:            g_atdl_host_groups,
#endif
    /* Defaults for drivers that do not have corresponding module parameter */
    can_queue:               64,
    cmd_per_lun:             64,
    sg_tablesize:            SG_MAX_SEGMENTS,
};


