/****************************************************************************

DESCRIPTION: Linux utility wrap functions

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#include "atdl.h"


void atdf_assert( void )
{
#ifdef ASSERTS_ON
    BUG_ON( 1 );
#endif
}


/**
 * atdf_sgl_get_sge_data() - Get address/length pair from SGE data
 */

u32 atdf_sgl_get_sge_data( struct scatterlist* sge, u64* addr )
{
    *addr = sg_dma_address( sge );

    return sg_dma_len( sge );
}


/**
 * atdf_sgl_get_next_sge() - Advance SGE pointer to next entry
 */

void atdf_sgl_get_next_sge( struct scatterlist** sge )
{
    *sge = sg_next( *sge );
}


/**
 * atdf_cpu_stall() - Stall CPU execution for the specified microseconds
 */

void atdf_cpu_stall( u32 delay_us )
{
    while( delay_us )
    {
        if( delay_us < 1000 )
        {
            udelay( delay_us );
            barrier();
            return;
        }
        else
        {
            mdelay( 1 );
            barrier();
            delay_us -= 1000;
        }
    }
}


/**
 * atdf_cpu_yield() - Yield the CPU to execute other tasks
 */

void atdf_cpu_yield( void )
{
    /* Yield the CPU if we are doing something potentially */
    /* time-consuming.  However, this is not safe to do    */
    /* in certain contexts, because it will schedule().    */

    if( !in_atomic() && !irqs_disabled() )
        cond_resched();
}


/**
 * atdf_cpu_get() - Fetch current CPU while preemption is disabled
 */

int atdf_cpu_get( void )
{
    int cpu;

    preempt_disable();
    cpu = smp_processor_id();
    preempt_enable();

    return cpu;
}


/**
 * atdf_cpu_barrier()
 * atdf_cpu_barrier_read()
 * atdf_cpu_barrier_write()
 * 
 * Memory barrier functions
 */

void atdf_cpu_barrier( void )
{
    mb();
}

void atdf_cpu_barrier_read( void )
{
    rmb();
}

void atdf_cpu_barrier_write( void )
{
    wmb();
}


/* String interface */

u32 atdf_sprintf( char* str, const char* format, ... )
{
    va_list ap;
    int     ret;

    va_start( ap, format );
    ret = vsprintf( str, format, ap );
    va_end( ap );

    return ret > 0 ? (u32) ret : 0;
}


int atdf_sscanf( const char* str, const char* format, ... )
{
    va_list ap;
    int     ret;

    va_start( ap, format );
    ret = vsscanf( str, format, ap );
    va_end( ap );

    return ret;
}


/**
 * atdi_log_print() - Log an event, filtering by ATTO level.
 */

void atdi_log_print( u8 atto_lvl, char* fmt, ... )
{
    char*   lnx_lvl;
    va_list ap;
    u32     lvl_len;
    u32     fmt_len;

    /* Get the kernel level.  Default to info as the highest level so we    */
    /* don't have to turn on all system debug messages to see ours.         */

    switch( atto_lvl )
    {
        case HWLOG_LVL_ALERT:       lnx_lvl = KERN_ALERT;   break;
        case HWLOG_LVL_CRITICAL:    lnx_lvl = KERN_CRIT;    break;
        case HWLOG_LVL_ERROR:       lnx_lvl = KERN_ERR;     break;
        case HWLOG_LVL_WARNING:     lnx_lvl = KERN_WARNING; break;
        case HWLOG_LVL_NOTICE:      lnx_lvl = KERN_NOTICE;  break;
        default:                    lnx_lvl = KERN_INFO;    break;
    }

    /* Insert the kernel level */

    lvl_len = strlen( lnx_lvl );
    fmt_len = strlen( fmt );

    va_start( ap, fmt );

    if( fmt_len + lvl_len >= 150 )
        vprintk( fmt, ap );
    else
    {
        char buf[150];

        strcpy( buf, lnx_lvl );
        strcat( buf, fmt );

        vprintk( buf, ap );
    }

    va_end( ap );
    va_start( ap, fmt );

    atci_serdbg_vprint( fmt, ap );

    va_end( ap );
}


/* Virtual memory API */

void* atdf_mem_virt_alloc( u32 size )
{
    return vzalloc( size );
}


void* atdf_mem_virt_alloc_node( u32 size, u32 node )
{
    return vzalloc_node( size, node );
}


void atdf_mem_virt_free( void* addr )
{
    if( addr )
        vfree( addr );
}


u32 atdf_mem_copy_from_user( void* src, void* dst, u32 size )
{
    return __copy_from_user( dst, src, size );
}


/* DMA memory API - Contiguous (one DMA address) */

void* atdf_mem_dma_alloc_node( struct device* dev, u32 size, u64* dma_addr, int cpu )
{
    int     orig_node = dev_to_node( dev );
    int     node = cpu_to_node( cpu );
    void*   buf;

    set_dev_node( dev, node );

    /* Some 32-bit kernels (eg. RHEL5.7) do not overwrite the high DWORD  */
    /* of the physical address, so make sure that it is zeroed out first. */

    *dma_addr = 0;

    buf = dma_alloc_coherent( dev, (size_t) size, (dma_addr_t*) dma_addr, GFP_KERNEL );

    set_dev_node( dev, orig_node );

    if( !buf )
        buf = dma_alloc_coherent( dev, (size_t) size, (dma_addr_t*) dma_addr, GFP_KERNEL );

    return buf;
}


void* atdf_mem_dma_alloc( struct device* dev, u32 size, u64* dma_addr )
{
    /* Some 32-bit kernels (eg. RHEL5.7) do not overwrite the high DWORD  */
    /* of the physical address, so make sure that it is zeroed out first. */

    *dma_addr = 0;

    return dma_alloc_coherent( dev, (size_t) size, (dma_addr_t*) dma_addr, GFP_KERNEL );
}


void atdf_mem_dma_free( struct device* dev, u32 size, void* virt_addr, u64 dma_addr )
{
    dma_free_coherent( dev, (size_t) size, virt_addr, (dma_addr_t) dma_addr );
}

void atdf_mem_dma_flush( struct device* dev, u32 size, void* virt_addr, u64 dma_addr )
{
    dma_sync_single_for_device( dev, (dma_addr_t) dma_addr, (size_t) size, DMA_BIDIRECTIONAL );
}

void atdf_mem_dma_invalidate( struct device* dev, u32 size, void* virt_addr, u64 dma_addr )
{
    dma_sync_single_for_cpu( dev, (dma_addr_t) dma_addr, (size_t) size, DMA_BIDIRECTIONAL );
}


/* DMA API wrappers */

void* atdi_mem_dma_alloc_node( struct atdl_adap* hostadap, u32 size, u64* dma_addr, u32 vector )
{
    /* Default to the node containing our PCI slot */

    int node = dev_to_node( &hostadap->adap.pcidev->dev );

    if( node >= 0
     && hostadap->msi_vectors )
    {
        struct atdl_msi_vector* msi_vect = hostadap->msi_vectors + vector;
        const cpumask_t*        mask = cpumask_of_node( node );

        if( mask )
        {
            int cpu = cpumask_first( mask );

            /* Override if we did load balancing */

            if( !cpumask_empty( &msi_vect->aff_mask ) )
            {
                /* On the off chance multiple nodes are mapped to the same  */
                /* vector (more nodes than vectors), give preference to ours*/

                if( !cpumask_intersects( mask, &msi_vect->aff_mask ) )
                    cpu = cpumask_first( &msi_vect->aff_mask );
            }

            return atdf_mem_dma_alloc_node( &hostadap->adap.pcidev->dev,
                                            size,
                                            dma_addr,
                                            cpu );
        }
    }

    return atdf_mem_dma_alloc( &hostadap->adap.pcidev->dev, size, dma_addr );
}

void* atdi_mem_dma_alloc( struct atdl_adap* hostadap, u32 size, u64* dma_addr )
{
    return atdf_mem_dma_alloc( &hostadap->adap.pcidev->dev, size, dma_addr );
}

void atdi_mem_dma_free( struct atdl_adap* hostadap, u32 size, void* virt_addr, u64 dma_addr )
{
    atdf_mem_dma_free( &hostadap->adap.pcidev->dev, size, virt_addr, dma_addr );
}

void atdi_mem_dma_flush( struct atdl_adap* hostadap, u32 size, void* virt_addr, u64 dma_addr )
{
    atdf_mem_dma_flush( &hostadap->adap.pcidev->dev, size, virt_addr, dma_addr );
}

void atdi_mem_dma_invalidate( struct atdl_adap* hostadap, u32 size, void* virt_addr, u64 dma_addr )
{
    atdf_mem_dma_invalidate( &hostadap->adap.pcidev->dev, size, virt_addr, dma_addr );
}


/**
 * atdf_msgl_get_sge() - Get DMA address/length pair for the given MSGL offset
 */

u32 atdf_msgl_get_sge( struct atdf_mapd_sgl* msgl, u32 offset, u64* dma_addr )
{
    /* The caller is requesting an offset into the user buffer, */
    /* but if the user buffer is not page aligned then we also  */
    /* need to account for where the buffer starts within the   */
    /* first page - which is what msgl->dma_offset is.          */

    u32 total_offset   = offset + msgl->dma_offset;
    u32 offset_in_page = total_offset % PAGE_SIZE;
    u32 page_ix = total_offset / PAGE_SIZE;

    if( !msgl->dma_maps[page_ix] )
    {
        msgl->dma_maps[page_ix] = dma_map_page( msgl->dev,
                                                msgl->dma_pages[page_ix],
                                                0,
                                                PAGE_SIZE,
                                                DMA_BIDIRECTIONAL );

        if( !msgl->dma_maps[page_ix] )
        {
            *dma_addr = 0;
            return 0;
        }
    }

    *dma_addr = msgl->dma_maps[page_ix] + offset_in_page;

    return PAGE_SIZE - offset_in_page;
}


/**
 * atdf_msgl_alloc() - Allocate an MSGL and assign/map pages to fulfill specified size
 */

struct atdf_mapd_sgl* atdf_msgl_alloc( struct atsi_adap* adap, u32 size, void** virt )
{
    u32 num_pages = (size + PAGE_SIZE) / PAGE_SIZE;
    u32 i;
    struct atdf_mapd_sgl* msgl;

    msgl = kzalloc( (num_pages * sizeof(struct page*))
                  + (num_pages * sizeof(dma_addr_t))
                  + sizeof(*msgl),
                    GFP_KERNEL );

    if( !msgl )
        return NULL;

    msgl->dev = &adap->pcidev->dev;
    msgl->dma_pages = (struct page **) (msgl + 1);
    msgl->dma_maps = (dma_addr_t*) (msgl->dma_pages + num_pages);
    msgl->num_pages = num_pages;

    for( i = 0; i < num_pages; i++ )
    {
        msgl->dma_pages[i] = alloc_page( GFP_KERNEL );

        if( !msgl->dma_pages[i] )
            goto fail;
    }

    msgl->kern_addr = *virt = vmap( msgl->dma_pages,
                                    num_pages,
                                    VM_MAP,
                                    pgprot_noncached(PAGE_KERNEL) );

    if( *virt )
        return msgl;

fail:

    atdf_msgl_free( msgl );

    return NULL;
}


/**
 * atdf_msgl_free() - Unmap/free MSGL pages and free MGL
 */

void atdf_msgl_free( struct atdf_mapd_sgl* msgl )
{
    int i;

    if( msgl )
    {
        if( msgl->kern_addr )
            vunmap( msgl->kern_addr );

        for( i = 0; i < msgl->num_pages; i++ )
        {
            if( msgl->dma_maps[i] )
                dma_unmap_page( msgl->dev, msgl->dma_maps[i], PAGE_SIZE, DMA_BIDIRECTIONAL );

            if( msgl->dma_pages[i] )
                __free_page( msgl->dma_pages[i] );
        }

        kfree( msgl );
    }
}


/**
 * atdf_msgl_init() - Initialize a mapped SGL context (allocated on the stack)
 */

void atdf_msgl_init( struct atdf_mapd_sgl* msgl, struct atsi_adap* adap )
{
    memset( msgl, 0, sizeof(*msgl) );

    msgl->dev = &adap->pcidev->dev;
}


/**
 * atdf_msgl_unmap() - Release mapped SGL resources
 */

void atdf_msgl_unmap( struct atdf_mapd_sgl* msgl )
{
    if( msgl->kern_addr )
        vunmap( msgl->kern_addr );

    if( msgl->dma_pages )
    {
        int i;

        for( i = 0; i < msgl->num_pages; i++ )
        {
            if( msgl->dma_maps[i] )
                dma_unmap_page( msgl->dev, msgl->dma_maps[i], PAGE_SIZE, DMA_BIDIRECTIONAL );

            if( msgl->dma_pages[i] )
            {
                set_page_dirty( msgl->dma_pages[i] );
                put_page( msgl->dma_pages[i] );
            }
        }

        kfree( msgl->dma_pages );
    }
}


/**
 * atdf_msgl_map() - Map the supplied user data into kernel space
 * 
 * Returns: System error code
 */

int atdf_msgl_map( struct atdf_mapd_sgl* msgl, void* userptr, u32 size )
{
    unsigned long   data = (unsigned long) userptr;
    unsigned long   first_page = (data & PAGE_MASK) >> PAGE_SHIFT;
    unsigned long   last_page = ((data + size - 1) & PAGE_MASK) >> PAGE_SHIFT;
    int             req_pages = last_page - first_page + 1;
    int             rcv_pages;

    msgl->dma_offset = data - (data & PAGE_MASK);

    msgl->dma_pages = kzalloc( req_pages * sizeof(struct page*)
                             + req_pages * sizeof(dma_addr_t), GFP_KERNEL );

    if( !msgl->dma_pages )
    {
        atto_log_err( ATTO_LOG_IOCTL,
                      "Failed to allocate MSGL memory (%d)\n",
                      req_pages );

        return -ENOMEM;
    }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0)
    down_read( &current->mm->mmap_lock );
#else
    down_read( &current->mm->mmap_sem );
#endif

    msgl->dma_maps = (dma_addr_t*) (msgl->dma_pages + req_pages);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6,5,0)  || IS_SLES_RELEASE_OR_LATER(15,6,0)
    rcv_pages = get_user_pages( data & PAGE_MASK, req_pages,
                                FOLL_WRITE | FOLL_FORCE, msgl->dma_pages );
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,9,0) || IS_SLES_RELEASE(12,3,0)
    rcv_pages = get_user_pages( data & PAGE_MASK, req_pages,
                                FOLL_WRITE | FOLL_FORCE, msgl->dma_pages, NULL );
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,6,0)
    rcv_pages = get_user_pages( data & PAGE_MASK, req_pages, 1, 1, msgl->dma_pages,
                                NULL );
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,4,168) && \
    LINUX_VERSION_CODE < KERNEL_VERSION(4,5,0)
    rcv_pages = get_user_pages( current, current->mm, data & PAGE_MASK,
                                req_pages, FOLL_WRITE | FOLL_FORCE,
                                msgl->dma_pages, NULL );
#else
    rcv_pages = get_user_pages( current, current->mm, data & PAGE_MASK,
                                req_pages, 1, 1, msgl->dma_pages, NULL );
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0)
    up_read( &current->mm->mmap_lock );
#else
    up_read( &current->mm->mmap_sem );
#endif

    if( rcv_pages > 0 )
        msgl->num_pages = rcv_pages;

    if( req_pages != rcv_pages )
    {
        atto_log_err( ATTO_LOG_IOCTL,
                      "MSGL failed get_user_pages (%d, %d)\n",
                      req_pages, rcv_pages );

        goto fail;
    }

    /* Map the pages */

    msgl->kern_addr = vmap( msgl->dma_pages, msgl->num_pages, VM_MAP,
                            pgprot_noncached( PAGE_KERNEL ) );

    if( !msgl->kern_addr )
    {
        atto_log_err( ATTO_LOG_IOCTL, "MSGL failed vmap (%d)\n", msgl->num_pages );

        goto fail;
    }

    return 0;

fail:

    atdf_msgl_unmap( msgl );

    return -ENOMEM;
}


/* PCI API */

int atdf_pci_cfg_read( struct pci_dev* pcidev, u16 offset, void* ptr, int size )
{
    switch( size )
    {
        case 1: return pci_read_config_byte( pcidev, offset, (u8*) ptr );
        case 2: return pci_read_config_word( pcidev, offset, (u16*) ptr );
        case 4: return pci_read_config_dword( pcidev, offset, (u32*) ptr );
    }

    return 0;
}


int atdf_pci_cfg_write( struct pci_dev* pcidev, u16 offset, u32 val, int size )
{
    switch( size )
    {
        case 1: return pci_write_config_byte( pcidev, offset, (u8) val );
        case 2: return pci_write_config_word( pcidev, offset, (u16) val );
        case 4: return pci_write_config_dword( pcidev, offset, (u32) val );
    }

    return 0;
}


void atdi_pci_mem_unmap( struct atdl_adap* hostadap, int* bars, int num_bars, void** maps )
{
    struct pci_dev* pcidev = hostadap->adap.pcidev;
    int             cur_bar;

    for( cur_bar = 0; cur_bar < num_bars; cur_bar++ )
    {
        if( maps[cur_bar] )
            iounmap( maps[cur_bar] );

        maps[cur_bar] = NULL;

        pci_release_region( pcidev, bars[cur_bar] );
    }
}


int atdi_pci_mem_map( struct atdl_adap* hostadap, int* bars, int num_bars, void** maps )
{
    struct pci_dev* pcidev = hostadap->adap.pcidev;
    int             cur_bar;

    /* First make sure the maps are cleared */

    for( cur_bar = 0; cur_bar < num_bars; cur_bar++ )
        maps[cur_bar] = NULL;

    /* Reserve all requested regions */

    for( cur_bar = 0; cur_bar < num_bars; cur_bar++ )
    {
        int rslt;

        if( (rslt = pci_request_region( pcidev, bars[cur_bar], hostadap->rsrc_name )) != 0 )
        {
            atdl_log_err( hostadap, ATTO_LOG_INIT,
                          "Unable to reserve MMIO region (%d, %d), error %d.\n",
                          cur_bar, bars[cur_bar], rslt );

            /* Release previously reserved regions */

            while( --cur_bar > -1 )
                pci_release_region( pcidev, bars[cur_bar] );

            return -1;
        }
    }

    /* Now, map everything */

    for( cur_bar = 0; cur_bar < num_bars; cur_bar++ )
    {
        ulong   reg = pci_resource_start( pcidev, bars[cur_bar] );
        u32     size = pci_resource_len( pcidev, bars[cur_bar] );

#ifdef CELERITY5
        if( cur_bar == 2
         && (maps[cur_bar] = ioremap_wc( reg, size )) )
        {
            /* Mapped DPP aperture with Write Combining */
        }
        else
#endif
        if( !(maps[cur_bar] = ioremap( reg, size )) )
        {
            atdl_log_err( hostadap, ATTO_LOG_INIT,
                          "Mem region at %lx (size %u, %d [%d]) unavailable - already in use?\n",
                          reg, size, bars[cur_bar], cur_bar );

            /* Unmap and release previous maps */

            atdi_pci_mem_unmap( hostadap, bars, num_bars, maps );

            return -1;
        }
    }

    return 0;
}


/**
 * atde_intr_msix_work_q() - Deferred MSI-X work queue entry point
 */

static void atde_intr_msix_work_q( struct work_struct* work )
{
    struct atdl_msi_vector* msi_vect = container_of( work, struct atdl_msi_vector, q_work );

    atci_intr_msi_handler( msi_vect->hostadap->coreadap, msi_vect->vector );
}


/**
 * atde_intr_msix_handler() - MSI-X entry point
 */

static irqreturn_t atde_intr_msix_handler( int irq, void* intr_ctx )
{
    struct atdl_msi_vector *msi_vect = (struct atdl_msi_vector *) intr_ctx;

    queue_work_on( smp_processor_id(), msi_vect->wq, &msi_vect->q_work );

    return IRQ_HANDLED;
}


/**
 * atde_intr_msi_handler() - MSI entry point
 */

static irqreturn_t atde_intr_msi_handler( int irq, void* intr_ctx )
{
    struct atdl_adap* hostadap = intr_ctx;

    atci_intr_msi_handler( hostadap->coreadap, 0 );

    return IRQ_HANDLED;
}


#ifndef HW_NO_LEGACY_INTS

/**
 * atde_intr_legacy_handler() - Legacy ISR entry point
 */

static irqreturn_t atde_intr_legacy_handler( int irq, void* intr_ctx )
{
    struct atdl_adap* hostadap = intr_ctx;

    if( atci_intr_legacy_handler( hostadap->coreadap ) )
        return IRQ_NONE;

    return IRQ_HANDLED;
}


/**
 * atdl_intr_legacy_enable() - Enable legacy interrupts.
 *
 * Returns: Zero on success or non-zero on failure.
 */

int atdl_intr_legacy_enable( struct atdl_adap* hostadap )
{
    int irq = hostadap->adap.pcidev->irq;

    atdl_log_info( hostadap, ATTO_LOG_INIT,
                   "Enable legacy interrupts irq=%d (%s)\n",
                   irq, hostadap->rsrc_name );

    if( request_irq( irq, atde_intr_legacy_handler, IRQF_SHARED, hostadap->rsrc_name, hostadap ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT, "Unable to request IRQ %d\n", irq );
        return 1;
    }

    atdl_log_info( hostadap, ATTO_LOG_INIT, "Claimed IRQ %d\n", irq );

    return 0;
}


/**
 * atdl_intr_legacy_disable() - Disable legacy interrupts.
 */

void atdl_intr_legacy_disable( struct atdl_adap* hostadap )
{
    struct pci_dev *pcidev = hostadap->adap.pcidev;

    atdl_log_info( hostadap, ATTO_LOG_PCI, "Disable legacy interrupts(%d)\n", pcidev->irq );

    free_irq( pcidev->irq, hostadap );
}

#endif /* !HW_NO_LEGACY_INTS */


/**
 * atdl_intr_msi_enable() - Enable MSI.
 *
 * Returns: The number of vectors allocated or zero on failure.
 */

u8 atdl_intr_msi_enable( struct atdl_adap* hostadap )
{
    struct pci_dev* pcidev = hostadap->adap.pcidev;

    atdl_log_info( hostadap, ATTO_LOG_PCI, "irq before enable msi: %d\n", pcidev->irq );

    if( pci_enable_msi( pcidev ) )
        return 0;

    atdl_log_info( hostadap, ATTO_LOG_PCI, "irq after enable msi: %d\n", pcidev->irq );

    if( request_irq( pcidev->irq, atde_intr_msi_handler, 0, hostadap->rsrc_name, hostadap ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_INIT, "Failed to claim MSI interrupts!\n" );

        pci_disable_msi( pcidev );

        return 0;
    }

    return 1;
}


/**
 * atdl_intr_msi_disable() - Disable MSI.
 */

void atdl_intr_msi_disable( struct atdl_adap* hostadap )
{
    struct pci_dev *pcidev = hostadap->adap.pcidev;

    atdl_log_info( hostadap, ATTO_LOG_PCI, "Disable MSI interrupts(%d)\n", pcidev->irq );

    free_irq( pcidev->irq, hostadap );

    pci_disable_msi( pcidev );
}


/**
 * pci_enable_msix_range() - Enable a range of MSI-X vectors.
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0) \
  && (!IS_RHEL_RELEASE_OR_LATER(6,6) || IS_RHEL_RELEASE(7,0))

static int pci_enable_msix_range( struct pci_dev* pcidev, struct msix_entry* entries,
                                  int min, int max )
{
    int cnt = max;
    int err = max;

    if( min > max )
        return -EINVAL;

    while( err )
    {
        err = pci_enable_msix( pcidev, entries, cnt );

        if( err < 0 )
            return err;

        if( err > 0 )
        {
            if( err < min )
                return -ENOSPC;

            cnt = err;
        }
    }

    return cnt;
}

#endif  /* LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0) */


/**
 * atdl_intr_msix_enable() - Enable MSI-X.
 *
 * Returns: The number of vectors allocated or zero on failure.
 */

u8 atdl_intr_msix_enable( struct atdl_adap* hostadap, u8 num_vect )
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,10,0)
    struct msix_entry*  msix_entries;
#endif
    int                 msicnt = 0;
    int                 vect;
    int                 rc;

    atdl_log_info( hostadap, ATTO_LOG_PCI, "LinuxEnableMsix: enabling %d vectors\n", num_vect );

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,10,0)

    /* Older kernels require this argument, which returns the IRQ to us */

    msix_entries = kcalloc( num_vect, sizeof( struct msix_entry ), GFP_KERNEL );

    if( !msix_entries )
    {
        atdl_log_info( hostadap, ATTO_LOG_PCI, "LinuxEnableMsix: not enough memory!" );
        return 0;
    }

    for( vect = 0; vect < num_vect; vect++ )
        msix_entries[vect].entry = vect;

    rc = pci_enable_msix_range( hostadap->adap.pcidev, msix_entries, num_vect, num_vect );

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_enable_msix_range(): %d\n", rc );
#else
    /* TODO: Figure out if we can use MSI-X with fewer than the number of requested vectors */

    rc = pci_alloc_irq_vectors( hostadap->adap.pcidev, num_vect, num_vect, PCI_IRQ_MSIX );

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_alloc_irq_vectors(): %d\n", rc );
#endif

    for( vect = 0; vect < rc; vect++ )
    {
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,10,0)
        int irq = msix_entries[vect].vector;
#else
        int irq = pci_irq_vector( hostadap->adap.pcidev, vect );
#endif
        int                     result;
        struct atdl_msi_vector* msi_vect;
        char                    workqueue_name[64];

        atdl_log_info( hostadap, ATTO_LOG_PCI, "MSI-X[%d] irq:%d\n", vect, irq );

        msi_vect = hostadap->msi_vectors + vect;

        /* Initialize the work queue */

        INIT_WORK( &msi_vect->q_work, atde_intr_msix_work_q );
        sprintf( workqueue_name, "%s-%d", hostadap->rsrc_name, irq );

        msi_vect->wq = g_atgi_ops.alloc_workqueue_w( workqueue_name, WQ_HIGHPRI, 1 );

        if( !msi_vect->wq )
        {
            atdl_log_err( hostadap, ATTO_LOG_PCI, "Error allocating interrupt workqueue\n" );
            break;
        }

        atdl_log_info( hostadap, ATTO_LOG_INIT, "Claim IRQ %d (%s)\n", irq, hostadap->rsrc_name );

        /* No balancing must be requested now to get what we set later */

        result = request_irq( irq,
                              atde_intr_msix_handler,
                              cpu_affinity < 2 ? IRQF_NOBALANCING : 0,
                              hostadap->rsrc_name,
                              msi_vect );

        if( result )
        {
            atdl_log_err( hostadap, ATTO_LOG_PCI,
                          "Error claiming MSI-X vector %d (error %d)\n",
                          irq, result );

            /* Free this now in case the caller accepts what we have allocated so far */

            g_atgi_ops.destroy_workqueue( msi_vect->wq );
            msi_vect->wq = NULL;
            break;
        }

        msi_vect->irq = irq;
        msicnt++;

        atdl_log_info( hostadap, ATTO_LOG_INIT, "Claimed IRQ %d\n", irq );
    }

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,10,0)
    kfree( msix_entries );
#endif

    return msicnt;
}


/**
 * atdl_intr_msix_disable() - Disable MSI-X.
 */

void atdl_intr_msix_disable( struct atdl_adap* hostadap )
{
    struct pci_dev* pcidev = hostadap->adap.pcidev;
    u32             vect;

    for( vect = 0; vect < hostadap->msi_cnt; vect++ )
    {
        struct atdl_msi_vector* msi_vect = hostadap->msi_vectors + vect;
        int                     irq = msi_vect->irq;

        if( irq != -1 )
        {
            synchronize_irq( irq );

            /* Clear the affinity */

            g_atgi_ops.irq_set_affinity_hint( irq, NULL );

            atdl_log_info( hostadap, ATTO_LOG_PCI, "free_irq(%d) called\n", irq );

            free_irq( irq, msi_vect );

            msi_vect->irq = -1;
        }

        if( msi_vect->wq )
        {
            g_atgi_ops.destroy_workqueue( msi_vect->wq );
            msi_vect->wq = NULL;
        }
    }

    atdl_log_info( hostadap, ATTO_LOG_PCI, "pci_disable_msix() called\n" );

    pci_disable_msix( pcidev );
}


/**
 * atde_timer_execute() - Shim to extract the real context data and callback,
 *                        which may live in the core, and service the timer.
 */

static void atde_timer_execute( struct timer_list* os_timer )
{ 
    struct atdf_timer *timer = container_of(os_timer, struct atdf_timer, os_timer);

    timer->handler( timer->data );
}


/**
 * atdf_timer_init() - Initialize a timer with context data and callback.
 */

void atdf_timer_init( struct atdf_timer* timer, void* context, void(*handler)(void*) )
{
    struct timer_list* os_timer = &timer->os_timer;

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,15,0)
    init_timer( os_timer );
    os_timer->function = (void (*)(unsigned long)) atde_timer_execute;
    os_timer->data = (unsigned long) os_timer;
#else
    timer_setup( os_timer, atde_timer_execute, 0 );
#endif

    timer->handler = handler;
    timer->data = context;
}


/**
 * atdf_timer_start() - Schedule a timer with the specified expiration.
 */

void atdf_timer_start( struct atdf_timer* timer, u32 expire_ms )
{
    timer->period_jif = (HZ * expire_ms) / 1000;
    timer->started = 1;
    timer->stopped = 0;

    atdf_timer_restart( timer );
}


/**
 * atdf_timer_restart() - Restart a timer (for periodic operation).
 */

void atdf_timer_restart( struct atdf_timer* timer )
{
    /* We only start/restart a timer, i.e. never modify a scheduled timer.  */
    /* So use add_timer() which does a little less work than mod_timer()    */

    if( timer->started )
    {
        timer->os_timer.expires = jiffies + timer->period_jif;

        add_timer( &timer->os_timer );
    }
    else
        atdf_timer_done( timer );
}


/**
 * atdf_timer_done() - Signal a one-shot timer is done.
 */

void atdf_timer_done( struct atdf_timer* timer )
{
    timer->stopped = 1;
}


/**
 * atdf_timer_cancel() - Cancel a timer, but do not wait for it to stop.
 */

void atdf_timer_cancel( struct atdf_timer* timer )
{
    if( timer->started )
    {
        /* Block a restart of periodic timer */

        timer->started = 0;

        if( !del_timer( &timer->os_timer ) )
            timer->stopped = 1;
    }
}


/**
 * atdf_timer_stop() - Cancel a timer and wait for execution to stop.
 */

void atdf_timer_stop( struct atdf_timer* timer )
{
    if( timer->started )
    {
        /* Block a restart of periodic timer */

        timer->started = 0;

        del_timer_sync( &timer->os_timer );
    }

    timer->stopped = 1;
}


/* Lookaside list API */

void* atdf_kmem_cache_alloc( struct kmem_cache* cache )
{
    return kmem_cache_alloc( cache, GFP_ATOMIC );
}

void* atdf_kmem_cache_alloc_node( struct kmem_cache* cache, int cpu_id )
{
    return kmem_cache_alloc_node( cache, GFP_ATOMIC, cpu_to_node( cpu_id ) );
}

void atdf_kmem_cache_free( struct kmem_cache* cache, void* object )
{
    kmem_cache_free( cache, object );
}


/* System time interface */

u32 atdf_curr_time_secs( void )
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,17,0)
    struct timeval now;
    
    do_gettimeofday( &now );

    return (u32) now.tv_sec;
#else
    struct timespec64 ts;

    ktime_get_real_ts64( &ts );
    
    return (u32) ts.tv_sec;
#endif
}


u32 atdf_curr_time_ms( void )
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,17,0)
    struct timeval now;

    do_gettimeofday( &now );

    return (u32) (now.tv_sec * 1000L + (now.tv_usec / 1000L));
#else
    struct timespec64 ts;

    ktime_get_real_ts64( &ts );

    return (u32) (ts.tv_sec * MSEC_PER_SEC + (ts.tv_nsec / NSEC_PER_MSEC));
#endif
}


u64 atdf_curr_time_us( void )
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,17,0)
    struct timeval now;

    do_gettimeofday( &now );

    return (u64) now.tv_sec * 1000000L + (unsigned long long) now.tv_usec;
#else
    struct timespec64 ts;

    ktime_get_real_ts64( &ts );

    return ts.tv_sec * USEC_PER_SEC + (ts.tv_nsec / NSEC_PER_USEC);
#endif
}


/**
 * atdi_rpt_targ_change() - Report a device state change to the OS
 */

void atdi_rpt_targ_change( struct atsi_dev_work_op* dwo )
{
    struct atdl_adap* hostadap = dwo->adap->hostadap;

    atdl_log_info( hostadap, ATTO_LOG_DISC, "Report state change: %s target %d(%d)\n",
                   atci_dev_action_str( dwo->action ), dwo->targ_id, dwo->type );

#if defined(FEAT_NVME_INI)

    /* Report an NVMe target */

    if( dwo->type & ATSI_DEV_TYPE_NVME )
        atdl_nvme_rpt_targ_change( dwo );
#endif

    /* Bail out if not a SCSI device */

    if( !(dwo->type & (ATSI_DEV_TYPE_SCSI | ATSI_DEV_TYPE_EXPANDER)) )
        return;

    /* Report a SCSI target, or expander when the transport layer is enabled */

#if defined(FEAT_TRANSPORT)
    if( hostadap->flags & ATDL_AF_XPORT_STARTED )
        atdl_xport_rpt_targ_change( dwo );
    else
#endif
    if( dwo->type & ATSI_DEV_TYPE_SCSI )
    {
        struct atsi_dev_work_op dwo2 = *dwo;

        int lun;

        if( dwo->action == ATSI_DEV_ACT_REMOVE
         || dwo->action == ATSI_DEV_ACT_RESCAN )
        {
            int found = 0;

            dwo2.action = ATSI_DEV_ACT_REMOVE;

            for( lun = 0; lun < ATSI_MAX_LUN; lun++ )
            {
                if( !atdl_scsi_dev_set_state( &dwo2, lun, FALSE ) )
                    found = 1;
            }

            if( !found )
                atdl_log_info( hostadap, ATTO_LOG_DISC, "No LUNs found for target %d\n", dwo->targ_id );
        }

        if( dwo->action == ATSI_DEV_ACT_ADD
         || dwo->action == ATSI_DEV_ACT_RESCAN )
        {
            dwo2.action = ATSI_DEV_ACT_ADD;

            for( lun = 0; lun < ATSI_MAX_LUN; lun++ )
                atdl_scsi_dev_set_state( &dwo2, lun, FALSE );
        }
    }
}


/**
 * atde_worker_execute() - Adapter worker thread callback to do the work
 */

int atde_worker_execute( void* data )
{
    struct atdl_globals* gd = &g_atdl_vars;

    atto_log_info( ATTO_LOG_INIT, "Worker thread running\n" );

    while( !kthread_should_stop() )
    {
        u32 delay;

        if( wait_event_interruptible( gd->worker_waitq,
                                      atci_worker_has_work() || kthread_should_stop() ) )
            break;

        delay = atci_worker_execute();

        if( delay )
        {
            /* Honor the delay, but allow the wait to be interrupted for    */
            /* thread termination.                                          */

            ulong jif = msecs_to_jiffies( delay );

            while( jif )
            {
                set_current_state( TASK_INTERRUPTIBLE );

                jif = schedule_timeout( jif );

                if( kthread_should_stop() )
                    break;
            }
        }
        else
        {
            /* We can spend a lot of time in here, so give up the CPU   */
            /* so we don't get caught by the watchdog.                  */

            schedule();
        }
    }

    gd->worker_thread = NULL;

    atto_log_info( ATTO_LOG_INIT, "worker thread stopping\n" );

    return 0;
}


void atdi_worker_schedule( u32 delay_ms )
{
    /* TODO: Need to honor delay */

    wake_up( &g_atdl_vars.worker_waitq );
}


/* DPC API */

void atdi_dpc_schedule( struct atdl_adap* hostadap )
{
    if( likely( hostadap->flags & ATDL_AF_ALLOW_DPC ) )
        queue_work_on( smp_processor_id(), hostadap->dpc_work_q, &hostadap->dpc_work_s );
}

/**
 * atde_dpc_work_q() - DPC work queue entry point
 */

void atde_dpc_work_q( struct work_struct* work )
{
    struct atdl_adap *hostadap = container_of( work, struct atdl_adap, dpc_work_s );

    atci_dpc_execute( hostadap->coreadap );
}


/* Spinlock API */

void atdf_spinlock_acquire( spinlock_t* lock, unsigned long* flags )
{
    spin_lock_irqsave( lock, *flags );
}


void atdf_spinlock_release( spinlock_t* lock, unsigned long* flags )
{
    spin_unlock_irqrestore( lock, *flags );
}


spinlock_t* atdf_spinlock_init( void )
{
    spinlock_t* lock;

    if( (lock = kmalloc( sizeof(*lock), GFP_KERNEL )) )
        spin_lock_init( lock );

    return lock;
}


void atdf_spinlock_free( spinlock_t* lock )
{
    if( lock )
        kfree( lock );
}


/* Completion API */

void atdf_compl_init( struct completion* cmpl )
{
    init_completion( cmpl );
}


void atdf_compl_reinit( struct completion* cmpl )
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,13,0) || IS_RHEL_RELEASE_OR_LATER(7,5))
    reinit_completion( cmpl );
#else
    cmpl->done = 0;
#endif
}


void atdf_compl_wait( struct completion* cmpl )
{
    wait_for_completion( cmpl );
}


void atdf_compl_signal( struct completion* cmpl )
{
    complete( cmpl );
}


int atdf_compl_wait_tmo( struct completion* cmpl, u32 tmo_ms )
{
    if( wait_for_completion_timeout( cmpl, msecs_to_jiffies( tmo_ms ) ) )
        return 1;

    return 0;
}

int atdf_compl_wait_tmo_intr( struct completion* cmpl, u32 tmo_ms )
{
    if( wait_for_completion_interruptible_timeout( cmpl, msecs_to_jiffies( tmo_ms ) ) )
        return 1;

    return 0;
}


/**
 * atdi_req_alloc() - Allocate a request
 * 
 * Returns: Pointer to the atsi_req (which is followed by PIOREQ)
 */

struct atsi_req* atdi_req_alloc( struct atdl_adap* hostadap )
{
    return atdl_req_alloc( hostadap, atdf_cpu_get() );
}


/**
 * atdi_req_free() - Free a request
 */

void atdi_req_free( struct atsi_req* req )
{
    atdl_req_free( req );
}


#if defined(FEAT_VSA) || defined(_MP_)

/**
 * atdi_fire_atto_evt() - Fire an async event to the app layer
 */

void atdi_fire_atto_evt( struct atevent_hba_hdr* data, u32 len )
{
    if( !g_atgi_ops.fire_atto_event )
        atto_log_warn( ATTO_LOG_EVENT, "Event handler not registered.\n" );
    else if( !g_atgi_ops.fire_atto_event( data, len ) )
        atto_log_err( ATTO_LOG_EVENT, "Error sending event message.\n" );
}

#endif /* FEAT_VSA || _MP_*/


#ifdef FEAT_BAD_BLK_EVT

/**
 * atdi_fire_bad_blk_evt() - Fire a bad block event to the app layer
 */

void atdi_fire_bad_blk_evt( void* data, u32 len )
{
    if( !g_atgi_ops.fire_bb_event )
        atto_log_warn( ATTO_LOG_EVENT, "BB event handler not registered.\n" );
    else if( !g_atgi_ops.fire_bb_event( data, len ) )
        atto_log_err( ATTO_LOG_EVENT, "Error sending BB event message.\n" );
}

#endif /* FEAT_BAD_BLK_EVT */



