/****************************************************************************

DESCRIPTION: Linux driver SAS transport interface

THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
PURSUANT TO WHICH IT WAS FURNISHED.

COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2017
ALL RIGHTS RESERVED.
UNPUBLISHED CONFIDENTIAL INFORMATION OF ATTO TECHNOLOGY, INC.

*****************************************************************************/


#include "atdl.h"


#if defined(FEAT_TRANSPORT) && defined(PROTO_SAS)


int atdl_dev_get_phy_info( struct atdl_adap* hostadap,
                           struct atdl_sas_dev* asdev,
                           u8 phy_id, PHWIHBA_SAS_PHY_INFO info );


static inline u64 atdl_sasaddr_to_u64( u8* sasaddr )
{
    return be64_to_cpu( *(u64*) sasaddr );
}

static const char* atdl_linkrate_to_str( u8 atto_rate )
{
    switch( atto_rate )
    {
        case SAS_LINK_RATE_1_5GB:
            return "1.5Gb/s";
        case SAS_LINK_RATE_3GB:
            return "3Gb/s";
        case SAS_LINK_RATE_6GB:
            return "6Gb/s";
        case SAS_LINK_RATE_12GB:
            return "12Gb/s";
        default:
            return "Unknown";
    }
}

static const char* atdl_prot_to_str( u8 prot )
{
    /* More than one could be set, but unlikely so for simplicity... */

    if( prot & SAS_PROTOCOL_SATA )
        return "sata";

    if( prot & SAS_PROTOCOL_SMP )
        return "smp";

    if( prot & SAS_PROTOCOL_STP )
        return "stp";

    if( prot & SAS_PROTOCOL_SSP )
        return "ssp";

    return "none";
}


/**
 * atdl_port_find_phy_by_id() - Get a port's PHY by PHY ID
 * 
 * Sync: Topology mutex must be held
 */

static struct atdl_sas_phy* atdl_port_find_phy_by_id( struct atdl_sas_port* asport, u8 phy_id )
{
    struct sas_phy* phy;

    if( phy_id >= asport->parent_dev->num_phys )
        return NULL;

    list_for_each_entry( phy, &asport->port->phy_list, port_siblings )
    {
        if( phy->identify.phy_identifier == phy_id )
            return &asport->parent_dev->phy_list[phy_id];
    }

    return NULL;
}


/**
 * atdl_port_create() - Create a SAS port and add it to the topology
 * 
 * Sync: Topology mutex must be held
 */

static struct atdl_sas_port* atdl_port_create( struct atdl_adap* hostadap,
                                               struct atdl_sas_dev* attach_dev )
{
    struct atdl_sas_port*   asport;
    struct atdl_sas_dev*    parent_dev;

    parent_dev = attach_dev->parent_dev;

    asport = kzalloc( sizeof(*asport), GFP_KERNEL );

    if( !asport )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT, "Failed to allocate SAS port\n" );
        return NULL;
    }

    INIT_LIST_HEAD( &asport->link_dev );

    asport->parent_dev = parent_dev;
    asport->attach_dev = attach_dev;

    /* Allocate a SAS Transport Layer port */

    atdl_log_dbg( hostadap, ATTO_LOG_XPORT, "sas_port_alloc_num() called.\n" );

    asport->port = sas_port_alloc_num( parent_dev->dev );

    if( !asport->port )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "sas_port_alloc_num failed on 0x%llx.\n",
                      parent_dev->sasaddr );
        goto error;
    }

    /* Put the SAS Transport port into the mid-layer */

    atdl_log_dbg( hostadap, ATTO_LOG_XPORT, "sas_port_add() called.\n" );

    if( sas_port_add( asport->port ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT, "sas_port_add failed on 0x%llx\n",
                      parent_dev->sasaddr );
        goto error;
    }

    /* The port goes on the parent device */

    list_add_tail( &asport->link_dev, &parent_dev->port_list );

    return asport;

error:
    /* NB: The release method is different after a successful sas_port_add()! */

    if( asport->port )
        sas_port_free( asport->port );

    kfree( asport );

    return NULL;
}


/**
 * atdl_port_destroy() - Destroy a SAS port
 * 
 * Sync: Topology mutex must be held
 */

static void atdl_port_destroy( struct atdl_adap* hostadap, struct atdl_sas_port* asport )
{
    if( asport )
    {
        atdl_log_info( hostadap, ATTO_LOG_XPORT, "Destroying port %p\n", asport );

        /* Remove the port from the parent device's port list */

        list_del( &asport->link_dev );

        kfree( asport );
    }
}


/**
 * atdl_port_add_phys() - Add PHYs to a port. 
 */

static int atdl_port_add_phys( struct atdl_adap* hostadap, struct atdl_sas_port* asport )
{
    int                     phycount = 0;
    struct atdl_sas_dev*    parent_dev = asport->parent_dev;
    struct atdl_sas_phy*    asphy;
    int                     i;
    HWIHBA_SAS_DEV_INFO     devinfo;
    HWIHBA_SAS_PHY_INFO     phyinfo;

    atdl_log_info( hostadap, ATTO_LOG_XPORT,
                   "Add phys to port 0x%llx<-->0x%llx\n",
                   parent_dev->sasaddr,
                   asport->attach_dev->sasaddr );

    /* Get the list of PHY IDs for the attached device */

    if( atci_get_dev_info( hostadap->coreadap, asport->attach_dev->target.target_id, &devinfo ) )
    {
        atdl_log_info( hostadap, ATTO_LOG_XPORT,
                       "Failed to get PHY IDs for device 0x%llx\n",
                       asport->attach_dev->sasaddr );

        return 0;
    }

    /* Run the list and add each PHY */

    for( i = 0; i < HWI_SDI_MAX_PHYS_WIDE_PORT; ++i )
    {
        u8 phy_id = devinfo.PhyId[i];

        /* Validate the PHY ID */

        if( phy_id == HWI_SAS_PHY_ID_INV
         || phy_id >= parent_dev->num_phys )
            continue;

        asphy = &parent_dev->phy_list[phy_id];

        if( !asphy->phy )
        {
            /* Previous effort to allocate/configure this PHY failed */

            atdl_log_err( hostadap, ATTO_LOG_XPORT,
                            "Unable to add PHY %d to port at device 0x%llx\n",
                            phy_id, parent_dev->sasaddr );
            continue;
        }

        atdl_log_info( hostadap, ATTO_LOG_XPORT, "Adding phy(%d) to port\n", phy_id );

        if( !atdl_dev_get_phy_info( hostadap, parent_dev, phy_id, &phyinfo ) )
        {
            /* Update the PHY info */

            enum sas_linkrate r = phyinfo.NegLinkRate;

            asphy->phy->negotiated_linkrate = r;
            asphy->attach_sasaddr = atdl_sasaddr_to_u64( phyinfo.AttachIdentFrame.SasAddr );
            asphy->phy->enabled = (r != SAS_PHY_DISABLED);
        }

        /* Add the phy to the port, even if we could not update the */
        /* phy info with current values. At least the topology may  */
        /* look right.                                              */

        sas_port_add_phy( asport->port, asphy->phy );

        ++phycount;
    }

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "Added %d phys\n", phycount );

    return phycount;
}


/**
 * atdl_port_name() - Fetch the sysfs name for the given port.
 */

static const char* atdl_port_name( struct atdl_sas_port* asport )
{
    return dev_name( &asport->port->dev );
}


/**
 * atdl_phy_name() - Fetch the sysfs name for the given PHY.
 */

static const char* atdl_phy_name( struct atdl_sas_phy* asphy )
{
    return dev_name( &asphy->phy->dev );
}


/**
 * atdl_dev_alloc_phys() - Allocate and initialize a device's PHY array.
 * 
 * Returns: The number of PHYs initialized or 0 on error.
 */

static int atdl_dev_alloc_phys( struct atdl_adap* hostadap, struct atdl_sas_dev* asdev, u8 num_phys )
{
    const char*             dev_str = asdev->dev_type == ATDL_SAS_DT_HOST ? "host" : "expander";
    int                     init_cnt = 0;
    u8                      phy_id;
    struct atdl_sas_phy*    asphy;

    atdl_log_info( hostadap, ATTO_LOG_XPORT,
                   "Adding %d PHYs to SAS %s 0x%llx\n",
                   num_phys,
                   dev_str,
                   asdev->sasaddr );

    asdev->phy_list = kcalloc( num_phys, sizeof(*asphy), GFP_KERNEL );

    if( !asdev->phy_list )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Failed to allocate PHY list for SAS %s 0x%llx\n",
                      dev_str,
                      asdev->sasaddr );
        return -1;
    }

    asdev->num_phys = num_phys;

    for( phy_id = 0; phy_id < num_phys; phy_id++ )
    {
        HWIHBA_SAS_PHY_INFO phyinfo;

        asphy = &asdev->phy_list[phy_id];

        /* Allocate the transport layer component */

        asphy->phy = sas_phy_alloc( asdev->dev, phy_id );

        if( !asphy->phy )
        {
            atdl_log_err( hostadap, ATTO_LOG_XPORT, "sas_phy_alloc(%d) failed.\n", phy_id );
            break;
        }

        /* Set default values as though the PHY is not used */

        asphy->phy->identify.phy_identifier = phy_id;
        asphy->phy->identify.device_type = SAS_PHY_UNUSED;
        asphy->phy->negotiated_linkrate =
        asphy->phy->maximum_linkrate_hw =
        asphy->phy->maximum_linkrate =
        asphy->phy->minimum_linkrate_hw =
        asphy->phy->minimum_linkrate = SAS_LINK_RATE_UNKNOWN;

        /* Set the actual PHY data */

        if( !atdl_dev_get_phy_info( hostadap, asdev, phy_id, &phyinfo ) )
        {
            asphy->attach_sasaddr = atdl_sasaddr_to_u64( phyinfo.AttachIdentFrame.SasAddr );
            asphy->phy->identify.device_type = phyinfo.IdentFrame.Type >> 4;
            asphy->phy->identify.initiator_port_protocols = phyinfo.IdentFrame.IniFlags;
            asphy->phy->identify.target_port_protocols = phyinfo.IdentFrame.TgtFlags;
            asphy->phy->identify.sas_address = atdl_sasaddr_to_u64( phyinfo.IdentFrame.SasAddr );

            asphy->phy->negotiated_linkrate = phyinfo.NegLinkRate;
            asphy->phy->minimum_linkrate_hw = phyinfo.MinLinkRateHw;
            asphy->phy->minimum_linkrate = phyinfo.MinLinkRateProg;
            asphy->phy->maximum_linkrate_hw = phyinfo.MaxLinkRateHw;
            asphy->phy->maximum_linkrate = phyinfo.MaxLinkRateProg;

            asphy->phy->enabled = (phyinfo.NegLinkRate != SAS_PHY_DISABLED);
        }
        else
        {
            atdl_log_err( hostadap, ATTO_LOG_XPORT, "Failed to get PHY ID %d info on 0x%llx\n",
                          phy_id, asdev->sasaddr );
        }

        /* Add the PHY whether we got info about it, or not */

        atdl_log_dbg( hostadap, ATTO_LOG_XPORT, "sas_phy_add(%d) called.\n", phy_id );

        if( sas_phy_add( asphy->phy ) )
        {
            atdl_log_err( hostadap, ATTO_LOG_XPORT, "sas_phy_add(%d) failed for 0x%llx.\n",
                          phy_id, asdev->sasaddr );

            sas_phy_free( asphy->phy );
            break;
        }

        atdl_log_info( hostadap, ATTO_LOG_XPORT,
                       " PHY ID %d: 0x%llx 0x%02x 0x%llx\n",
                       phy_id,
                       asphy->phy->identify.sas_address,
                       asphy->phy->negotiated_linkrate,
                       asphy->attach_sasaddr );

        init_cnt++;
    }

    return init_cnt;
}


/**
 * atdl_dev_create() - Allocate a device and initialize defaults
 */

static struct atdl_sas_dev* atdl_dev_create( struct atdl_adap* hostadap, PSASADDR sas_address )
{
    struct atdl_sas_dev* d = kzalloc( sizeof(*d), GFP_KERNEL );

    if( !d )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT, "Failed to allocate device\n" );
        return NULL;
    }

    INIT_LIST_HEAD( &d->link_topo );
    INIT_LIST_HEAD( &d->port_list );

    d->sasaddr = atdl_sasaddr_to_u64( sas_address->c );

    return d;
}


/**
 * atdl_dev_destroy() - Destroy a device
 * 
 * Sync: Topology utex must be held
 */

static void atdl_dev_destroy( struct atdl_adap* hostadap, struct atdl_sas_dev* asdev )
{
    struct atdl_sas_port* asport;
    struct atdl_sas_port* tempport;

    if( !asdev )
        return;

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "Destroying device 0x%llx\n", asdev->sasaddr );

    /* For expanders and hosts, free ports to children */

    list_for_each_entry_safe( asport, tempport, &asdev->port_list, link_dev )
    {
        atdl_port_destroy( hostadap, asport );
    }

    /* Free own PHYs */

    if( asdev->phy_list )
        kfree( asdev->phy_list );

    /* Remove from tracked device list */

    if( !list_empty( &asdev->link_topo ) )
        list_del( &asdev->link_topo );

    kfree( asdev );
}


/**
 * atdl_dev_name() - Fetch the sysfs name for the given devcic.
 */

static const char* atdl_dev_name( struct atdl_sas_dev* asdev )
{
    return dev_name( asdev->dev );
}


/**
 * atdl_dev_find_port() - Find a port in the devices port list by SAS address
 *
 * Returns: Pointer to port, if found; otherwise NULL.
 * 
 * Sync: Topology mutex must be held
 */

static struct atdl_sas_port* atdl_dev_find_port( struct atdl_sas_dev* asdev, u64 sasaddr )
{
    struct atdl_sas_port* p;

    list_for_each_entry( p, &asdev->port_list, link_dev )
    {
        if( p->attach_dev
         && p->attach_dev->sasaddr == sasaddr )
            return p;
    }

    return NULL;
}


/**
 * atdl_dev_init_as_exp() - Initialize a device as an expander
 * 
 * Returns: 0 on success, non-zero on failure
 */

static int atdl_dev_init_as_exp( struct atdl_adap* hostadap,
                                 struct atdl_sas_dev* asdev,
                                 struct atdl_sas_port* asport )
{
    struct atcl_adap*           a = hostadap->coreadap;
    u8                          low_phy_id = 0;
    HWIHBA_EXP_INFO             sei;
    struct sas_expander_device* sed;
    int                         phy_count;
    HWIHBA_SAS_DEV_INFO         devinfo;
    int                         phy_id;

    atdl_log_info( hostadap, ATTO_LOG_XPORT,
                   "Make expander at target ID %d\n", asdev->target.target_id );

    asdev->dev_type = ATDL_SAS_DT_EXP;

    /* Get the expander info */

    if( atci_get_exp_info( a, asdev->target.target_id, &sei ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Failed to get expander 0x%llx info.\n",
                      asdev->sasaddr );
        return -1;
    }

    /* Get the lowest parent PHY ID */

    if( atci_get_dev_info( a, asport->attach_dev->target.target_id, &devinfo ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Failed to get near PHY info for 0x%llx\n",
                      asdev->sasaddr );
        return -1;
    }

    for( phy_id = 0; phy_id < HWI_SDI_MAX_PHYS_WIDE_PORT; phy_id++ )
    {
        if( devinfo.PhyId[phy_id] != HWI_SAS_PHY_ID_INV )
        {
            low_phy_id = devinfo.PhyId[phy_id];
            break;
        }
    }

    /* Alloc expander device off the parent port */

    atdl_log_dbg( hostadap, ATTO_LOG_XPORT, "sas_expander_alloc() called.\n" );

    asdev->rphy = sas_expander_alloc( asport->port, SAS_EDGE_EXPANDER_DEVICE );

    if( !asdev->rphy )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT, "sas_expander_alloc failed for device 0x%llx\n",
                      asdev->sasaddr );
        return -1;
    }

    /* Set rphy attributes */

    asdev->rphy->identify.target_port_protocols = SAS_PROTOCOL_SMP;
    asdev->rphy->identify.initiator_port_protocols = SAS_PROTOCOL_SMP;
    asdev->rphy->identify.sas_address = asdev->sasaddr;
    asdev->rphy->identify.phy_identifier = low_phy_id;
    asdev->rphy->scsi_target_id = asdev->target.target_id;

    /* Set expander attributes: vendor, component, etc. */

    sed = rphy_to_expander_device( asdev->rphy );

    memcpy( sed->vendor_id, sei.VendorId, SAS_EXPANDER_VENDOR_ID_LEN );
    memcpy( sed->product_id, sei.ProductId, SAS_EXPANDER_PRODUCT_ID_LEN );
    memcpy( sed->product_rev, sei.Revision, SAS_EXPANDER_PRODUCT_REV_LEN );
    memcpy( sed->component_vendor_id, sei.ComponentVendor,
            SAS_EXPANDER_COMPONENT_VENDOR_ID_LEN );
    sed->component_id = sei.ComponentId;
    sed->component_revision_id = sei.ComponentRev;

    /* Add rphy to sysfs */
    
    atdl_log_dbg( hostadap, ATTO_LOG_XPORT, "sas_rphy_add() called.\n" );

    if( sas_rphy_add( asdev->rphy ) )
    {
        sas_rphy_free( asdev->rphy );
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Failed to add remote PHY 0x%llx to device 0x%llx.\n",
                      asdev->sasaddr, asport->parent_dev->sasaddr );
        return -1;
    }

    /* Save a type-agnostic shortcut to the device structure */

    asdev->dev = &asdev->rphy->dev;

    /* Add PHYs to the expander - must have asdev->dev first */

    phy_count = atdl_dev_alloc_phys( hostadap, asdev, sei.NumPhys );

    if( phy_count != sei.NumPhys )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT, "Missing PHYs (%d of %d) for expander 0x%llx.\n",
                      sei.NumPhys - phy_count,
                      sei.NumPhys, asdev->sasaddr );
    }

    atdl_log_dbg( hostadap, ATTO_LOG_XPORT, "Init expander device done\n" );

    return 0;
}


/**
 * atdl_dev_init_as_end_dev() - Initialize a device as an end device
 * 
 * Returns: 0 on success, non-zero on failure
 */

static int atdl_dev_init_as_end_dev( struct atdl_adap* hostadap,
                                     struct atdl_sas_dev* asdev,
                                     struct atdl_sas_port* asport )
{
    struct atcl_adap*   a = hostadap->coreadap;
    HWIHBA_SAS_DEV_INFO devinfo;
    int phy_ix;

    atdl_log_info( hostadap, ATTO_LOG_XPORT,
                   "Make end device at target ID %d\n", asdev->target.target_id );

    asdev->dev_type = ATDL_SAS_DT_END_DEV;

    /* Get the device info */

    if( atci_get_dev_info( a, asdev->target.target_id, &devinfo ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Failed to get device0x%llx info.\n",
                      asdev->sasaddr );
        return -1;
    }

    /* Alloc an end device off the parent port */

    atdl_log_dbg( hostadap, ATTO_LOG_XPORT, "sas_end_device_alloc() called.\n" );

    asdev->rphy = sas_end_device_alloc( asport->port );

    if( !asdev->rphy )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "sas_end_device_alloc failed for device 0x%llx.\n",
                      asdev->sasaddr );
        return -1;
    }

    /* Set rphy attributes */

    asdev->rphy->identify.target_port_protocols = devinfo.TgtFlags;
    asdev->rphy->identify.sas_address = asdev->sasaddr;

    for( phy_ix = 0; phy_ix < HWI_SDI_MAX_PHYS_WIDE_PORT; phy_ix++ )
    {
        if( devinfo.PhyId[phy_ix] != HWI_SAS_PHY_ID_INV )
        {
            asdev->rphy->identify.phy_identifier = devinfo.PhyId[phy_ix];
            break;
        }
    }

    atdl_log_info( hostadap, ATTO_LOG_XPORT,
                   "sas_rphy:target_port_protocols:0x%x\n",
                   asdev->rphy->identify.target_port_protocols );

    /* This is a waste of time, but do it anyway in case Linux is ever      */
    /* fixed to allow us to sync the target IDs.                            */

    asdev->rphy->scsi_target_id = asdev->target.target_id;

    /* Add rphy to sysfs */

    atdl_log_dbg( hostadap, ATTO_LOG_XPORT, "sas_rphy_add() called.\n" );

    if( sas_rphy_add( asdev->rphy ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT, "sas_rphy_add failed for device 0x%llx.\n",
                      asdev->sasaddr );
        sas_rphy_free( asdev->rphy );
        return -1;
    }

    atdl_log_info( hostadap, ATTO_LOG_XPORT,
                   "sas_rphy scsi_target_id:0x%x\n",
                   asdev->rphy->scsi_target_id );

    /* Save a type-agnostic shortcut to the device structure */

    asdev->dev = &asdev->rphy->dev;

    atdl_log_dbg( hostadap, ATTO_LOG_XPORT, "Make end device done\n" );

    return 0;
}


/**
 * atdl_dev_get_phy_info() - Get PHY information for device's PHY ID
 * 
 * Returns: 0 on success, non-zero on failure
 */

int atdl_dev_get_phy_info( struct atdl_adap* hostadap,
                           struct atdl_sas_dev* asdev,
                           u8 phy_id, PHWIHBA_SAS_PHY_INFO info )
{
    struct atcl_adap*   a = hostadap->coreadap;
    SASADDR             sasaddr;

    sasaddr.q = be64_to_cpu( asdev->sasaddr );

    /* 'Do the right thing' based on device type */

    if( asdev->dev_type == ATDL_SAS_DT_HOST )
        return atci_get_phy_info( a, phy_id, info );

    return atci_smp_discover( a, &sasaddr, phy_id, info );
}


/**
 * atdl_dev_find_by_rphy() - Find a device by remote PHY
 *
 * Returns: Pointer to device if found; otherwise NULL
 * 
 * Sync: Topology mutex must be held
 */

static struct atdl_sas_dev* atdl_dev_find_by_rphy( struct atdl_adap* hostadap, struct sas_rphy* rphy )
{
    struct atdl_sas_dev* d;

    list_for_each_entry( d, &hostadap->sas_topo.dev_list, link_topo )
    {
        if( d->rphy == rphy )
            return d;
    }

    return NULL;
}


/**
 * atdl_dev_find_by_targ_id() - Find a device by target ID
 *
 * Returns: Pointer to device if found; otherwise NULL
 * 
 * Sync: Topology mutex must be held
 */

static struct atdl_sas_dev* atdl_dev_find_by_targ_id( struct atdl_adap* hostadap, u16 targ_id )
{
    struct atdl_sas_dev* d;

    list_for_each_entry( d, &hostadap->sas_topo.dev_list, link_topo )
    {
        if( d->target.target_id == targ_id )
            return d;
    }

    return NULL;
}


/**
 * atdl_targ_validate() - Handle device info/location changes
 *
 * Validate the PHYs of the parent device related to the port connecting the
 * indicated device.
 * 
 * Sync: Topology mutex must be held
 */

static void atdl_targ_validate( struct atdl_adap* hostadap, struct atdl_sas_dev* asdev )
{
    u16                     targid = asdev->target.target_id;
    struct atdl_sas_dev*    parent_dev;
    struct atdl_sas_port*   asport;
    u8                      phy_id;

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "Validate device at target %d\n", targid );

    parent_dev = asdev->parent_dev;

    asport = atdl_dev_find_port( asdev->parent_dev, asdev->sasaddr );

    if( !asport )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Failed to find port 0x%llx<-->0x%llx for target %d validation.\n",
                      parent_dev->sasaddr,
                      asdev->sasaddr, targid );
        return;
    }

    /* Port membership may have changed for one or more PHYs, so query the
     * current state of each PHY.  Only PHYs that are now in or were previously
     * in the port are of interest.  Other events will be delivered to handle
     * changes in other PHYs.
     */

    for( phy_id = 0; phy_id < parent_dev->num_phys; phy_id++ )
    {
        HWIHBA_SAS_PHY_INFO     phyinfo;
        struct atdl_sas_phy*    asphy;
        enum sas_linkrate       rate;
        u64                     att_sasaddr;

        asphy = atdl_port_find_phy_by_id( asport, phy_id );

        if( atdl_dev_get_phy_info( hostadap, parent_dev, phy_id, &phyinfo ) )
        {
            /* If an error occurred and the PHY is known, remove it from the
             * port.  If it's a transient error and something ends up trying to
             * add it to another port later, the kernel will BUG().  More
             * likely, it is a chained removal.
             */

            if( asphy )
            {
                atdl_log_info( hostadap, ATTO_LOG_XPORT,
                               "%s() failed for %s on %s (parent of target %d).\n",
                               __func__, atdl_phy_name( asphy ),
                               atdl_dev_name( parent_dev ), targid );

                /* Remove PHY from this port */

                sas_port_delete_phy( asport->port, asphy->phy );
                asphy->attach_sasaddr = 0ULL;
            }

            continue;
        }

        rate = phyinfo.NegLinkRate;
        att_sasaddr = atdl_sasaddr_to_u64( phyinfo.AttachIdentFrame.SasAddr );

        /* An existing PHY needs to be updated or removed */

        if( asphy )
        {
            atdl_log_info( hostadap, ATTO_LOG_XPORT,
                           "Updating %s on %s (phy 0x%x on 0x%llx)\n",
                           atdl_phy_name( asphy ), atdl_dev_name( parent_dev ),
                           asphy->phy->identify.phy_identifier,
                           parent_dev->sasaddr );

            if( rate == SAS_LINK_RATE_UNKNOWN
             || asdev->sasaddr != att_sasaddr )
            {
                atdl_log_info( hostadap, ATTO_LOG_XPORT,
                               "Dropping %s from %s (0x%llx<-->0x%llx). "
                               "(Now linked to 0x%llx @ %s)\n",
                               atdl_phy_name( asphy ), atdl_port_name( asport ),
                               parent_dev->sasaddr, asdev->sasaddr,
                               att_sasaddr,
                               atdl_linkrate_to_str( rate ) );

                /* Remove PHY from this port */

                sas_port_delete_phy( asport->port, asphy->phy );
                asphy->attach_sasaddr = 0ULL;
            }

            asphy->phy->negotiated_linkrate = rate;
            asphy->phy->enabled = (rate != SAS_PHY_DISABLED);

            continue;
        }

        /* A non member PHY may need to be added to the port */

        asphy = &parent_dev->phy_list[phy_id];

        if( !asphy->phy )
        {
            atdl_log_info( hostadap, ATTO_LOG_XPORT,
                           "phy 0x%x is missing from %s.  Skipping...\n", phy_id,
                           atdl_dev_name( parent_dev ) );
            continue;
        }

        if( rate == SAS_LINK_RATE_UNKNOWN
         || asdev->sasaddr != att_sasaddr )
        {
            /* Not linked up, or linked to somebody else -> skip it. */

            continue;
        }

        atdl_log_info( hostadap, ATTO_LOG_XPORT,
                       "Adding %s (phy 0x%x) to %s (0x%llx<-->0x%llx).\n",
                       atdl_phy_name( asphy ), phy_id, atdl_port_name( asport ),
                       parent_dev->sasaddr, asdev->sasaddr );

        /* This PHY might be newly enabled, so update the link rate too */

        asphy->attach_sasaddr = att_sasaddr;
        asphy->phy->enabled = (rate != SAS_PHY_DISABLED);
        asphy->phy->negotiated_linkrate = rate;
        sas_port_add_phy( asport->port, asphy->phy );
    }
}


/**
 * atdl_targ_add() - Handle state change to add a target
 *
 * Sync: Topology mutex must be held
 */

static void atdl_targ_add( struct atsi_dev_work_op* dwo )
{
    struct atdl_adap*       hostadap = dwo->adap->hostadap;
    struct atdl_sas_topo*   topo = &hostadap->sas_topo;
    struct atcl_adap*       a = hostadap->coreadap;
    struct atdl_sas_dev*    parent_dev = NULL;
    struct atdl_sas_port*   asport = NULL;
    struct atdl_sas_dev*    asdev = NULL;
    HWIHBA_SAS_DEV_INFO     devinfo;
    SASADDR                 sas_addr;

    /* Get the device info */

    if( atci_get_dev_info( a, dwo->targ_id, &devinfo ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Could not get device info for target ID %d\n", dwo->targ_id );
        return;
    }

    if( devinfo.LocFlags & HWI_SDI_LF_DIRECT )
        parent_dev = topo->host_dev;
    else if( devinfo.LocFlags & HWI_SDI_LF_EXPANDER )
    {
        /* Make sure the parent device is in the topology */

        parent_dev = atdl_dev_find_by_targ_id( hostadap, devinfo.ExpTargetId );
    }

    if( !parent_dev )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Add device failure - parent of target ID %d not found\n", dwo->targ_id );
        return;
    }

    atdl_log_info( hostadap, ATTO_LOG_XPORT,
                   "Parent address: 0x%llx\n", parent_dev->sasaddr );

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "Parent type: %d\n", parent_dev->dev_type );

    /* Get the SAS address */

    if( atci_get_dev_addr( a, dwo->targ_id, &sas_addr ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Add device failure - target ID %d device addr failure\n", dwo->targ_id );
        return;
    }

    /* Create a new device */

    asdev = atdl_dev_create( hostadap, &sas_addr );

    if( !asdev )
        return;

    /* Set the device attributes */

    asdev->target.target_id = dwo->targ_id;
    asdev->target.port_num = dwo->port_num;
    asdev->parent_dev = parent_dev;

    /* Create a new port to link between parent device and this device */

    asport = atdl_port_create( hostadap, asdev );

    if( !asport )
        goto error_device;

    /* Add PHYs from parent device to the port */

    atdl_port_add_phys( hostadap, asport );

    /* Put this device in the topology before making an end device */
    /* or expander device, as the topology may be searched during  */
    /* the next phase.                                             */

    list_add_tail( &asdev->link_topo, &topo->dev_list );

    /* Initialize device type specific data */

    if( dwo->type & ATSI_DEV_TYPE_EXPANDER )
    {
        if( atdl_dev_init_as_exp( hostadap, asdev, asport ) )
            goto error_added_device;
    }
    else if( dwo->type & ATSI_DEV_TYPE_SCSI )
    {
        if( atdl_dev_init_as_end_dev( hostadap, asdev, asport ) )
            goto error_added_device;
    }
    else
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Unknown device type %d for device 0x%llx\n",
                      dwo->type, asdev->sasaddr );
        goto error_added_device;
    }

    /* Assign the persistent SG ID */

    if( asdev->dev_type == ATDL_SAS_DT_END_DEV )
    {
        if( dwo->targ_id < ATDL_XPORT_NUM_SG_IDS
         && topo->sg_ids[dwo->targ_id] != ATDL_XPORT_INV_SG_ID )
            asdev->sgid_pers = topo->sg_ids[dwo->targ_id];
        else
        {
            asdev->sgid_pers = topo->last_sg_id++;

            if( dwo->targ_id < ATDL_XPORT_NUM_SG_IDS )
                topo->sg_ids[dwo->targ_id] = asdev->sgid_pers;
        }
    }

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "Added device 0x%llx\n", asdev->sasaddr );

    atdl_log_dbg( hostadap, ATTO_LOG_XPORT, "Added %s device, %s\n",
                  atdl_prot_to_str( asdev->rphy->identify.target_port_protocols ),
                  dev_name( &asdev->rphy->dev ) );
    return;

error_added_device:

    sas_port_delete( asport->port );
    atdl_port_destroy( hostadap, asport );

error_device:

    atdl_dev_destroy( hostadap, asdev );
}


/**
 * atdl_targ_remove() - Handle state change to remove a target
 *
 * Sync: Topology mutex must be held
 */

static void atdl_targ_remove( struct atdl_adap* hostadap, struct atdl_sas_dev* asdev )
{
    u16                     targid = asdev->target.target_id;
    struct atdl_sas_dev*    parent_dev;
    struct atdl_sas_port*   asport;

    if( asdev->dev_type == ATDL_SAS_DT_EXP )
    {
        struct atdl_sas_port* p;
        struct atdl_sas_port* tp;

        /* Remove sub-tree beneath this expander */

        list_for_each_entry_safe( p, tp, &asdev->port_list, link_dev )
        {
            atdl_targ_remove( hostadap, p->attach_dev );
        }
    }

    /* Log this now, as the device name will go away when the device */
    /* is removed.                                                   */

    atdl_log_err( hostadap, ATTO_LOG_XPORT, "removed %s device, %s\n",
                  atdl_prot_to_str( asdev->rphy->identify.target_port_protocols ),
                  dev_name( &asdev->rphy->dev ) );

    /* Save this device's parent device pointer */

    parent_dev = asdev->parent_dev;

    /* Now update the parent's PHYs in this port */

    atdl_log_info( hostadap, ATTO_LOG_XPORT,
                   "Updating PHYs on %s after detaching %s.\n",
                   atdl_dev_name( parent_dev ),
                   atdl_dev_name( asdev ) );

    atdl_targ_validate( hostadap, asdev );

    /* Find the port from the parent device's port list */

    asport = atdl_dev_find_port( parent_dev, asdev->sasaddr );

    if( !asport )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Error removing device - could not find port for device 0x%llx\n",
                      parent_dev->sasaddr );
        return;
    }

    /* Delete the sas port; sas_port_delete() deletes the end device (rphy) */
    /* and unlinks the sas phys from the port. It also deletes the PHYs     */
    /* on the rphy instance.                                                */

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "sas_port_delete() called.\n" );

    sas_port_delete( asport->port );

    /* Free the port */

    atdl_port_destroy( hostadap, asport );

    /* Free this device */

    atdl_dev_destroy( hostadap, asdev );

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "Device removed at target %d\n", targid );
}


/**
 * atdl_targ_rescan() - Rescan an existing target
 *
 * Sync: Topology mutex must be held
 */

static void atdl_targ_rescan( struct atdl_adap* hostadap, struct atdl_sas_dev* asdev )
{
    u16 targid = asdev->target.target_id;
    
    if( asdev->dev_type != ATDL_SAS_DT_END_DEV )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "Attempt to invoke rescan on non-target device at target %d.\n",
                      targid );
        return;
    }

    if( !asdev->rphy )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT, "No rphy for device at target %d rescan.\n",
                      targid );
        return;
    }

    /* Call mid-layer to invoke a scan on the port */

    atdl_log_dbg( hostadap, ATTO_LOG_SCSI, "scsi_scan_target() called.\n" );

    scsi_scan_target( asdev->dev, 0, asdev->rphy->scsi_target_id,
                      SCAN_WILD_CARD, 1 );

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "Scan device done\n" );
}


/**
 * atdl_xport_rpt_targ_change() - Report a target change to the OS
 */

void atdl_xport_rpt_targ_change( struct atsi_dev_work_op* dwo )
{
    struct atdl_adap*       hostadap = dwo->adap->hostadap;
    struct atdl_sas_topo*   topo = &hostadap->sas_topo;
    struct atdl_sas_dev*    asdev;

    mutex_lock( &topo->topo_mutex );

    /* Find the associated device for the state handlers (may be NULL) */

    asdev = atdl_dev_find_by_targ_id( hostadap, dwo->targ_id );

    /* Handle the state (perform the specified action) */

    switch( dwo->action )
    {
        case ATSI_DEV_ACT_ADD:

            /* Sanity check for existing device with given target ID */

            if( asdev )
            {
                atdl_log_err( hostadap, ATTO_LOG_XPORT,
                              "Add device failure - duplicate target ID %d\n", dwo->targ_id );
            }
            else
                atdl_targ_add( dwo );

            break;

        case ATSI_DEV_ACT_REMOVE:

            if( !asdev )
                goto error_target;

            atdl_targ_remove( hostadap, asdev );
            break;

        case ATSI_DEV_ACT_RESCAN:

            if( !asdev )
                goto error_target;

            atdl_targ_rescan( hostadap, asdev );
            break;

        case ATSI_DEV_ACT_REACTIVATE:

            if( !asdev )
                goto error_target;

            atdl_targ_validate( hostadap, asdev );
            break;

        case ATSI_DEV_ACT_RELOCATE:

            if( !asdev )
                goto error_target;

            atdl_targ_remove( hostadap, asdev );
            atdl_targ_add( dwo );
            break;

        case ATSI_DEV_ACT_ACTIVATE:
        case ATSI_DEV_ACT_DEACTIVATE:

            if( !asdev )
                goto error_target;

            break;

        default:
            break;
    }

    mutex_unlock( &topo->topo_mutex );
    return;

error_target:

    mutex_unlock( &topo->topo_mutex );

    atdl_log_info( hostadap, ATTO_LOG_DISC,
                   "Failed to find device at target ID %d for %s action\n",
                   dwo->targ_id, atci_dev_action_str( dwo->action ) );
}


/**
 * atdl_xport_slave_alloc() - Attach transport data to scsi_device
 * 
 * Returns: 0 for success, non-zero on error
 */

int atdl_xport_slave_alloc( struct atdl_adap* hostadap, struct scsi_device* sdev )
{
    struct atdl_sas_dev*    asdev;
    struct sas_rphy*        rphy;

    rphy = dev_to_rphy( sdev->sdev_target->dev.parent );

    asdev = atdl_dev_find_by_rphy( hostadap, rphy );

    if( !asdev )
    {
        atdl_log_err( hostadap, ATTO_LOG_DISC,
                      "Unable to find sas device 0x%llx\n",
                      rphy->identify.sas_address );
        return -1;
    }

    /* Link to our target data to get the target ID in the I/O path */

    sdev->hostdata = &asdev->target;

    return 0;
}


/* Template entry points */

static int atde_xport_get_linkerrors( struct sas_phy* phy )
{
    return -ENXIO;
}


static int atde_xport_phy_reset( struct sas_phy* phy, int param )
{
    return -ENXIO;
}


static int atde_xport_get_bay_identifier( struct sas_rphy* rphy )
{
    int                     bay = 0;
    struct Scsi_Host*       sh;
    struct atdl_adap*       hostadap;
    struct atdl_sas_dev*    asdev;
    HWIHBA_SAS_DEV_INFO     devinfo;

    atto_log_dbg( ATTO_LOG_XPORT,
                  "atde_xport_get_bay_identifier(rphy:0x%llx) called\n",
                  rphy->identify.sas_address );

    sh = rphy_to_shost( rphy );

    if( !sh )
    {
        atto_log_warn( ATTO_LOG_XPORT,
                       "atde_xport_get_bay_identifier: invalid scsi host for rphy 0x%llx\n",
                       rphy->identify.sas_address );
        return -ENXIO;
    }

    hostadap = atdl_sh_to_hadap( sh );

    atdl_log_dbg( hostadap, ATTO_LOG_XPORT,
                  "atde_xport_get_bay_identifier: found host%d for rphy\n",
                  sh->host_no );

    asdev = atdl_dev_find_by_rphy( hostadap, rphy );

    if( !asdev )
    {
        atdl_log_warn( hostadap, ATTO_LOG_XPORT,
                       "atde_xport_get_bay_identifier: unknown device for rphy 0x%llx\n",
                       rphy->identify.sas_address );
        return -ENXIO;
    }

    atdl_log_dbg( hostadap, ATTO_LOG_XPORT,
                  "atde_xport_get_bay_identifier: found device at HW target %d for rphy\n",
                  asdev->target.target_id );

    if( !atci_get_dev_info( hostadap->coreadap, asdev->target.target_id, &devinfo ) )
        bay = devinfo.SlotNum;

    atdl_log_info( hostadap, ATTO_LOG_XPORT,
                   "atde_xport_get_bay_identifier: bay %d found for rphy 0x%llx\n",
                   bay, rphy->identify.sas_address );
    return bay;
}


static int atde_xport_set_phy_speed( struct sas_phy* phy, struct sas_phy_linkrates* rate )
{
    return -ENXIO;
}


static int atde_xport_phy_enable( struct sas_phy* phy, int enable )
{
    return -ENXIO;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,17,0) || IS_SLES_RELEASE_OR_LATER(15, 1, 0)

static void atde_xport_smp_handler( struct bsg_job* job, struct Scsi_Host* sh,
                                    struct sas_rphy* rphy )
{
    struct request* req = blk_mq_rq_from_pdu( job );

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)

static void atde_xport_smp_handler( struct bsg_job* job, struct Scsi_Host* sh,
                                    struct sas_rphy* rphy )
{
    struct request* req = job->req;

#else

static int atde_xport_smp_handler( struct Scsi_Host* sh, struct sas_rphy* rphy,
                                   struct request* req )
{
    struct request*         rsp = req->next_rq;
#endif
    struct atdl_adap*       hostadap = atdl_sh_to_hadap( sh );
    int                     ret;
    int                     type;
    struct atsi_smp_task    task;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,18,0) || IS_SLES_RELEASE_OR_LATER(15, 5, 0)
    struct scsi_cmnd *scmd = blk_mq_rq_to_pdu( req );
#endif

    task.Response = NULL;

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0) && ! IS_SLES_RELEASE_OR_LATER(15, 1, 0)
    if( !rsp )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT, "%s: no space for smp response\n", __func__ );
        ret = -EINVAL;
        goto out;
    }
#endif

    if( !rphy )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "%s: SMP commands only supported on remote phys.\n", __func__ );
        ret = -EINVAL;
        goto out;
    }

    type = rphy->identify.device_type;

    if( type != SAS_EDGE_EXPANDER_DEVICE
     && type != SAS_FANOUT_EXPANDER_DEVICE )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "%s: SMP commands only supported by expanders.\n", __func__ );
        ret = -EINVAL;
        goto out;
    }

    task.SasAddress = rphy->identify.sas_address;
    task.Request = bio_data( req->bio );
    task.RequestLen = blk_rq_bytes( req );

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)  || IS_SLES_RELEASE_OR_LATER(15, 1, 0)
    task.Response = kmalloc( job->reply_payload.payload_len, GFP_KERNEL );
    task.ResponseLen = job->reply_payload.payload_len;

    if( !task.Response )
    {
        ret = -ENOMEM;
        goto out;
    }
#else
    task.Response = bio_data( rsp->bio );
    task.ResponseLen = blk_rq_bytes( rsp );
#endif

    if( atci_smp_command( hostadap->coreadap, &task ) )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT,
                      "%s: SMP request failed, result %d\n", __func__, task.Result );
        ret = -ENXIO;
        goto out;
    }
    else
        ret = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0) || IS_SLES_RELEASE_OR_LATER(15, 1, 0)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,18,0) || IS_SLES_RELEASE_OR_LATER(15, 5, 0)
    scmd->resid_len = task.RequestLen;
#else
    scsi_req( req )->resid_len = task.RequestLen;
#endif
    sg_copy_from_buffer( job->reply_payload.sg_list,
                         job->reply_payload.sg_cnt, task.Response,
                         job->reply_payload.payload_len );

out:
    if( g_atgi_ops.bsg_job_done )
        g_atgi_ops.bsg_job_done( job, ret, task.ResponseLen );
    else
        WARN_ONCE( 1, "Incompatible shim driver loaded" );

    kfree( task.Response );

    return;

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
    scsi_req( req )->resid_len = task.RequestLen;
    scsi_req( rsp )->resid_len = task.ResponseLen;
out:
    return ret;
#else
    req->resid_len = task.RequestLen;
    rsp->resid_len = task.ResponseLen;
out:
    return ret;
#endif
}


/* SAS Transport Layer function template */

struct sas_function_template g_atdl_sas_xport_templ = {
    .get_linkerrors           = atde_xport_get_linkerrors,
    .phy_reset                = atde_xport_phy_reset,
    .get_bay_identifier       = atde_xport_get_bay_identifier,
    .get_enclosure_identifier = NULL,
    .set_phy_speed            = atde_xport_set_phy_speed,
    .phy_enable               = atde_xport_phy_enable,
    .smp_handler              = atde_xport_smp_handler,
};


/**
 * atdl_xport_templ_alloc() - Allocate an FC transport layer template
 * 
 * Returns: Pointer to the associated SCSI template or NULL.
 */

struct scsi_transport_template* atdl_xport_templ_alloc( void )
{
    return sas_attach_transport( &g_atdl_sas_xport_templ );
}


/**
 * atdl_xport_templ_free() - Free the FC transport layer template
 */

void atdl_xport_templ_free( struct scsi_transport_template* stt )
{
    sas_release_transport( stt );
}


/**
 * atdl_xport_adap_init() - Initialize transport layer data and start operation
 * 
 * Returns: 0 for success, non-zero on error
 */

int atdl_xport_adap_init( struct atdl_adap* hostadap )
{
    struct Scsi_Host*       sh = hostadap->adap.scsihost;
    struct atcl_adap*       a = hostadap->coreadap;
    struct atdl_sas_topo*   topo = &hostadap->sas_topo;
    SASADDR                 sas_addr;
    struct atdl_sas_dev*    asdev;
    u16                     i;

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "Initializing SAS topology\n" );

    /* Initialize topology defaults */

    INIT_LIST_HEAD( &topo->dev_list );
    mutex_init( &topo->topo_mutex );

    for( i = 0; i < ATDL_XPORT_NUM_SG_IDS; i++ )
        topo->sg_ids[i] = ATDL_XPORT_INV_SG_ID;

    /* Create the host device and add it to the topology */

    atci_get_host_addr( a, &sas_addr );

    asdev = atdl_dev_create( hostadap, &sas_addr );

    if( !asdev )
    {
        atdl_log_err( hostadap, ATTO_LOG_XPORT, "Failed to allocate SAS Host Device\n" );

        return -1;
    }

    atdl_log_info( hostadap, ATTO_LOG_XPORT,
                   "Adding SAS Host Device (0x%llx) to topology\n",
                   asdev->sasaddr );

    /* Set host attributes */

    asdev->dev_type = ATDL_SAS_DT_HOST;
    asdev->dev = &sh->shost_gendev;

    /* Add PHYs to host - must set asdev->dev first */

    if( atdl_dev_alloc_phys( hostadap, asdev, atci_get_host_phys( a ) ) == -1 )
        return -1;

    /* Store the host in the topology */

    topo->host_dev = asdev;

    atdl_log_info( hostadap, ATTO_LOG_XPORT,
                   "SAS Topology created for host: %d PHYs  0x%llx\n",
                   asdev->num_phys, asdev->sasaddr );

    return 0;
}


/**
 * atdl_xport_adap_stop() - Stop adapter transport layer operation
 */

void atdl_xport_adap_stop( struct atdl_adap* hostadap )
{
    struct Scsi_Host* sh = hostadap->adap.scsihost;

    /* Starting in kernel 4.12, sas_remove_host() calls scsi_remove_host()
     * after removing children.  Since we have common code that calls
     * scsi_remove_host() to cover other protocols, just remove children here.
     * This recursively destroys the sas_ports and sas_phys.
     */

    sas_remove_children( &sh->shost_gendev );

    /* Topology structures are freed later after scsi_remove_host() is called */
}


/**
 * atdl_xport_adap_free() - Free transport layer data.
 * 
 * Somewhere around 3.10, the scsi_device objects started surviving
 * sas_remove_host().  They do not survive scsi_remove_host().  As a last
 * gasp, attached disks will issue SYNC CACHE commands, and
 * atto_queuecommand() expects (by way of atdl_sd_to_targ_id()) that the
 * device's 'hostdata' field is valid.  Since atdl_sas_dev holds the scsi_device
 * data, topology objects are not freed until now.
 */

void atdl_xport_adap_free( struct atdl_adap* hostadap )
{
    struct atdl_sas_topo*   topo = &hostadap->sas_topo;
    struct atdl_sas_dev*    d;
    struct atdl_sas_dev*    td;

    /* Make sure atdl_xport_adap_init() at least initialized defaults */

    if( !topo->dev_list.next )
        return;

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "Destroy topology\n" );

    mutex_lock( &topo->topo_mutex );

    /* Safely traverse the topology - free the memory for each device.  Note
     * that the order is not necessarily bottom up, and the underlying sas_xxx
     * objects in the kernel have already been removed.
     */

    list_for_each_entry_safe( d, td, &topo->dev_list, link_topo )
    {
        atdl_dev_destroy( hostadap, d );
    }

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "Destroying host device\n" );

    /* Free the host device */

    atdl_dev_destroy( hostadap, topo->host_dev );

    topo->host_dev = NULL;

    mutex_unlock( &topo->topo_mutex );

    mutex_destroy( &topo->topo_mutex );

    atdl_log_info( hostadap, ATTO_LOG_XPORT, "Topology destroyed.\n" );
}


#ifdef FEAT_SAS_DEV_ORDER

/**
 * atdi_xport_print_dev_list() - /proc listing helper to print device table.
 * 
 * Returns: The number of characters printed to the buffer or -1 if sas_dev_order
 *          is disabled/invalid.
 */

int atdi_xport_print_dev_list( struct atdl_adap* hostadap, char* buf )
{
    struct atdl_sas_topo*   topo = &hostadap->sas_topo;
    char*                   p = buf;
    int                     cnt = 0;
    int                     sgid = 0;
    struct atdl_sas_dev*    asdev;

    if( sas_dev_order == 0
     || sas_dev_order > 3 )
        return -1;

    p += sprintf( p,
                  "\n"
                  "Discovered Devices:\n"
                  "\n"
                  "   #   sgID  SAS Address        TargID\n"
                  "--------------------------------------\n" );

    if( !(hostadap->flags & ATDL_AF_XPORT_STARTED) )
        goto done;

    mutex_lock( &topo->topo_mutex );

    if( sas_dev_order == 2 )
    {
        /* Assign the 0-based target index in Linux order.  Note it's saved */
        /* in a separate location to not clobber the original value.  This  */
        /* facilitates dynamically changing sas_dev_order.                  */

        list_for_each_entry( asdev, &topo->dev_list, link_topo )
        {
            if( asdev->dev_type == ATDL_SAS_DT_END_DEV )
                asdev->sgid_rel = sgid++;
        }
    }

    if( sas_dev_order == 2 || sas_dev_order == 3 )
    {
        u16 targ_id;

        /* List the devices in HW target order, but include the Linux       */
        /* SG ID (3rd column in /proc/scsi/sg/devices). That is not the     */
        /* global sg number, just the relative order of the device within   */
        /* the SCSI host.                                                   */

        for( targ_id = 0; targ_id < num_targets; ++targ_id )
        {
            SASADDR sa;

            /* This is essentially a presence check to avoid topology       */
            /* traversal below for every target ID.                         */

            if( atci_get_dev_addr( hostadap->coreadap, targ_id, &sa ) )
                continue;

            asdev = atdl_dev_find_by_targ_id( hostadap, targ_id );

            if( !asdev
             || asdev->dev_type != ATDL_SAS_DT_END_DEV )
                continue;

            /* The SG ID should be the same as the order in which the       */
            /* target device was added to the mid-layer, which is the same  */
            /* as its relative order in the topology device list.           */

            if( sas_dev_order == 3 )
                sgid = asdev->sgid_pers;
            else
                sgid = asdev->sgid_rel;

            p += sprintf( p, "%4d  %4d   %08X:%08X  %4d\n",
                          ++cnt,
                          sgid,
                          (u32) (asdev->sasaddr >> 32),
                          (u32) asdev->sasaddr,
                          targ_id );
        }
    }
    else if( sas_dev_order == 1 )
    {
        /* List the devices as they were reported to the SCSI layer */

        list_for_each_entry( asdev, &topo->dev_list, link_topo )
        {
            if( asdev->dev_type != ATDL_SAS_DT_END_DEV )
                continue;

            p += sprintf( p, "%4d  %4d   %08X:%08X  %4d\n",
                          ++cnt,
                          asdev->sgid_pers,
                          (u32) (asdev->sasaddr >> 32),
                          (u32) asdev->sasaddr,
                          asdev->target.target_id );
        }
    }

    mutex_unlock( &topo->topo_mutex );

done:
    return p - buf;
}

#endif /* FEAT_SAS_DEV_ORDER */

#endif /* FEAT_TRANSPORT && PROTO_SAS */

