/****************************************************************************

DESCRIPTION: Driver to discover ATTO HBAs for the ATTO ConfigTool

Copyright (C) 2006-2018  ATTO TECHNOLOGY, INC.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*****************************************************************************/

#define ATTO_DRVR_RAWNAME  ATTO_CFG_MOD_RAWNAME    /* Set by makefile */

#include "atto_cfg2.h"

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/slab.h>
#include <linux/workqueue.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <net/sock.h>
#include <linux/bsg-lib.h>

#include "atto_features.h"
#include "atto_log.h"
#include "atto.h"
#include "atto_ioctl.h"


MODULE_DESCRIPTION ("ATTO ConfigTool driver");
MODULE_AUTHOR ("ATTO Technology");
MODULE_LICENSE("GPL");
MODULE_VERSION(AV_VERSION_STR);


unsigned int atto_log_mask = 0;
module_param (atto_log_mask, uint, S_IRUGO | S_IRUSR);

unsigned int log_level = 0;
module_param (log_level, uint, S_IRUGO | S_IRUSR);


struct attocfg_host
{
    struct list_head   list;      /* List of hosts */
    struct kobject    *shost;     /* The Scsi_Host object from the driver */
    struct attohba_ops hba_ops;
};


/* Structure and functions to handle async event notification. */

typedef struct AsyncEventQueue ASYNC_EVENT_Q, *PASYNC_EVENT_Q;
#define AE_MAX_LENGTH    0x100

struct AsyncEventQueue
{
    PASYNC_EVENT_Q next;
    int            length;
    int            port;
    unsigned char  data [AE_MAX_LENGTH];
};

spinlock_t      async_event_lock;
PASYNC_EVENT_Q  async_event_queue = NULL;
PASYNC_EVENT_Q  async_event_free = NULL;
void           *async_event_memory = NULL;
int             async_events_initialized = 0;

/* Arbitrary count of the max async events we can handle at once. */

#define ASYNC_EVENT_COUNT   64

struct socket *event_socket [3] = {NULL, NULL, NULL};


/****************************************************************************

    FUNCTION:   void send_async_event (PASYNC_EVENT_Q event)

    PURPOSE:    Send the specified data to the async event ports.  Cannot
                be run from interrupt context!

****************************************************************************/

static void send_async_event (PASYNC_EVENT_Q event)
{
    int res;
    struct kvec vec;
    struct msghdr msg;
    int sock_num;

    memset ( &vec, 0, sizeof ( vec ) );
    memset ( &msg, 0, sizeof ( msg ) );

    vec.iov_base = event->data;
    vec.iov_len  = event->length;

    msg.msg_flags = MSG_DONTWAIT | MSG_NOSIGNAL;

    sock_num = event->port - ATTO_EVT_PORT_HBA1;

Resend:
       	if (event_socket [sock_num])
            res = kernel_sendmsg(event_socket [sock_num],
                                &msg,
                                &vec,
                                1,
                                event->length);
        else
            res = -ENOTSOCK;

    if( sock_num == 0 )
    {
        /* VDA events sent to port ATTO_EVT_PORT_HBA1 also */
        /* need to be sent to port ATTO_EVT_PORT_HBA2.     */

        sock_num = 1;
        goto Resend;
    }

    if( res )
    {
#ifdef HW_DEBUG
        /* Note: if there is no listener for the given socket, */
        /* you'll get -ECONNREFUSED (-111) from kernel_sendmsg.  */

        atto_log_err( ATTO_LOG_EVENT, "error sending event message (%d)\n", res );
#endif
    }
}


/****************************************************************************

    FUNCTION:   void handle_async_event_queue (unsigned long unused)

    PURPOSE:    Clean off the async event queue.  This is a task that does
                not run in interrupt context.

****************************************************************************/

static void handle_async_event_queue( struct work_struct *work )
{
    unsigned long flags;

    /* Take the events off the queue in order and send them. */

    spin_lock_irqsave(&async_event_lock, flags);

    while( 1 )
    {
        PASYNC_EVENT_Q curr;

        curr = async_event_queue;

        if( curr == NULL )
        {
            spin_unlock_irqrestore( &async_event_lock, flags );
            break;
        }

        async_event_queue = curr->next;

        spin_unlock_irqrestore( &async_event_lock, flags );

        send_async_event( curr );

        spin_lock_irqsave( &async_event_lock, flags );
        curr->next = async_event_free;
        async_event_free = curr;
    }
}


/****************************************************************************

    FUNCTION:   int init_async_event_queue (void)

    PURPOSE:    Initialize the async event queue

    RETURNS:    1 on success, 0 on failure

****************************************************************************/

static int init_async_event_queue( void )
{
    unsigned long flags;
    PASYNC_EVENT_Q queue;
    int i;

    if( async_events_initialized )
        return 1;

    queue = kcalloc( ASYNC_EVENT_COUNT, sizeof (ASYNC_EVENT_Q), GFP_KERNEL );

    if( !queue )
    {
        atto_log_err( ATTO_LOG_INIT, "could not alloc async event queue\n" );

        return 0;
    }

    spin_lock_irqsave( &async_event_lock, flags );

    /* Sanity check, all queues should be empty */

    if( async_event_free || async_event_queue )
    {
        atto_log_err( ATTO_LOG_INIT, "Async event queue error during initialization (%p, %p)\n",
                      async_event_free, async_event_queue );
        spin_unlock_irqrestore( &async_event_lock, flags );
        kfree( queue );
        return 0;
    }

    async_events_initialized = 1;
    async_event_memory = queue;

    /* Fill in the free async event queue */

    async_event_free = queue;

    for( i = 0; i < ASYNC_EVENT_COUNT; i++, queue++ )
        queue->next = queue + 1;

    queue->next = NULL;

    spin_unlock_irqrestore( &async_event_lock, flags );

    return 1;
}


/****************************************************************************

    FUNCTION:   int free_async_event_queue (void)

    PURPOSE:    Free (uninitialize) the async event queue

    RETURNS:    1 on success, 0 on failure

****************************************************************************/

int free_async_event_queue( void )
{
    unsigned long flags;
    void *addr_to_free;

    spin_lock_irqsave( &async_event_lock, flags );

    if( !async_events_initialized )
    {
        /* This is ok, we could get called even if the queue is uninited .*/
error_exit:
        spin_unlock_irqrestore( &async_event_lock, flags );
        return 0;
    }

    /* Sanity check, all queues should be empty */

    if( async_event_memory == NULL )
    {
        atto_log_err( ATTO_LOG_INIT,
                      "Async event queue memory NULL even though queue is initialized\n" );
        goto error_exit;
    }

    /* Cannot vfree with interrupts disabled, so wait until we release spinlock. */

    addr_to_free = async_event_memory;

    async_event_memory = NULL;
    async_event_free   = NULL;
    async_event_queue  = NULL;
    async_events_initialized = 0;

    spin_unlock_irqrestore( &async_event_lock, flags );

    kfree( addr_to_free );

    return 1;
}


/****************************************************************************

    FUNCTION:   fire_async_event()

    PURPOSE:    Queue the specified data to be sent out the specified port

    RETURNS:    1 on success, 0 on failure

****************************************************************************/

DECLARE_WORK( async_event_task, handle_async_event_queue );

static int fire_async_event( void *data, int len, int port )
{
    unsigned long flags;
    PASYNC_EVENT_Q event;

    if( len > AE_MAX_LENGTH )
    {
        atto_log_err( ATTO_LOG_EVENT, "Illegal async event length of %d\n", len );
        return 0;
    }

    spin_lock_irqsave ( &async_event_lock, flags );

    if( !async_events_initialized )
    {
        spin_unlock_irqrestore( &async_event_lock, flags );
        atto_log_err( ATTO_LOG_EVENT, "Fire_async_event called with an uninitialized queue\n" );
        return 0;
    }

    event = async_event_free;

    if( !event )
    {
        spin_unlock_irqrestore(&async_event_lock, flags);
        atto_log_err( ATTO_LOG_EVENT, "Out of async events\n" );
        return 0;
    }

    async_event_free = event->next;

    event->next   = NULL;
    event->length = len;
    event->port   = port;
    memcpy( event->data, data, len );

    /* To preserve event ordering, place this at the end of the queue */

    if( async_event_queue == NULL )
        async_event_queue = event;
    else
    {
        PASYNC_EVENT_Q curr = async_event_queue;

        while( curr->next )
            curr = curr->next;

        curr->next = event;
    }

    spin_unlock_irqrestore( &async_event_lock, flags );

    schedule_work( &async_event_task );

    return 1;
}


/****************************************************************************

    FUNCTION:   fire_atto_event()

    PURPOSE:    Queue the specified data to be sent out the generic async
                event ports

    RETURNS:    1 on success, 0 on failure

****************************************************************************/

static int fire_atto_event( void *data, int len )
{
    atto_log_info( ATTO_LOG_EVENT, "ATTO async event firing (len %d)\n", len );

    return fire_async_event( data, len, ATTO_EVT_PORT_HBA1 );
}


#ifdef FEAT_BAD_BLK_EVT

/****************************************************************************

    FUNCTION:   fire_bb_event()

    PURPOSE:    Queue the specified data to be sent out the BB event port

    RETURNS:    1 on success, 0 on failure

****************************************************************************/

int fire_bb_event( void* data, int len )
{
    atto_log_info( ATTO_LOG_EVENT, "BB async event firing (len %d)\n", len );

    return fire_async_event( data, len, ATTO_EVT_PORT_TARG );
}

#endif


/****************************************************************************

    FUNCTION:   void cleanup_async_event_sockets (void)

    PURPOSE:    Cleans up the sockets for async event notification

****************************************************************************/

static void cleanup_async_event_sockets( void )
{
    int i;
    int num_sockets = sizeof(event_socket)/sizeof(event_socket[0]);

    for( i = 0; i < num_sockets; i++ )
    {
        if( event_socket [i] )
        {
            sock_release( event_socket [i] );
            event_socket [i] = NULL;
            atto_log_info( ATTO_LOG_INIT, "event socket %d released\n", i );
        }
    }
}


/****************************************************************************

    FUNCTION:   int init_async_event_sockets (void)

    PURPOSE:    Initializes the sockets for async event notification

    RETURNS:    1 on success, 0 on failure

****************************************************************************/

static int init_async_event_sockets( void )
{
    int sock_num;
    int reuse = 1;
#if LINUX_VERSION_CODE > KERNEL_VERSION(5,9,0)
    sockptr_t optval = KERNEL_SOCKPTR(&reuse);
#endif

    int num_sockets = sizeof(event_socket)/sizeof(event_socket[0]);

    if( !init_async_event_queue() )
        return 0;

    for( sock_num = 0; sock_num < num_sockets; sock_num++ )
    {
        int res;
        struct sockaddr_in sin;

        if( event_socket [sock_num] )
            continue;

        res = sock_create( AF_INET, SOCK_DGRAM, 0, &event_socket [sock_num] );

        if( res < 0 )
        {
            atto_log_err( ATTO_LOG_INIT, "Could not create event socket %d (%d)\n",
                          sock_num, res );
            cleanup_async_event_sockets();
            return 0;
        }

        memset( &sin, 0, sizeof (sin) );
        sin.sin_family = AF_INET;
        sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
        sin.sin_port = htons(ATTO_EVT_PORT_HBA1 + sock_num);

        res = event_socket [sock_num]->ops->connect( event_socket [sock_num],
                                                     (struct sockaddr *) &sin,
                                                     sizeof (sin),
                                                     0 );

        if( res < 0 )
        {
            atto_log_err( ATTO_LOG_INIT, "Could not bind event socket %d (%d)\n",
                          sock_num, res );
            cleanup_async_event_sockets();
            return 0;
        }
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,9,0)
        res = event_socket [sock_num]->ops->setsockopt( event_socket [sock_num],
                                                        0,
                                                        SO_REUSEADDR,
                                                        (char*)&reuse,
                                                        sizeof(reuse) );
#else
        res = event_socket [sock_num]->ops->setsockopt( event_socket [sock_num],
                                                        0,
                                                        SO_REUSEADDR,
                                                        optval,
                                                        sizeof(reuse) );
#endif
        if( res < 0 )
        {
            atto_log_info( ATTO_LOG_INIT, "Could not set SO_REUSEADDR bit %d (%d)\n",
                           sock_num, res );
        }
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,9,0)
        res = event_socket [sock_num]->ops->setsockopt( event_socket [sock_num],
                                                        0,
                                                        SO_REUSEPORT,
                                                        (char*)&reuse,
                                                        sizeof(reuse) );
#else
        res = event_socket [sock_num]->ops->setsockopt( event_socket [sock_num],
                                                        0,
                                                        SO_REUSEPORT,
                                                        optval,
                                                        sizeof(reuse) );
#endif
        if( res < 0 )
        {
            atto_log_info( ATTO_LOG_INIT, "Could not set SO_REUSEPORT bit %d (%d)\n",
                           sock_num, res );
        }

        atto_log_info( ATTO_LOG_INIT, "event socket %d created\n", sock_num );
    }

    return 1;
}


#if defined(FEAT_NVME_INI) || defined(FEAT_NVME_TGT)
#include <linux/scatterlist.h>
#include <scsi/fc/fc_fs.h>
#include <linux/nvme.h>
#include <linux/nvme-fc.h>
#include <linux/nvme-fc-driver.h>
#endif

#ifdef FEAT_NVME_INI

static int gew_nvme_fc_register_localport( struct nvme_fc_port_info* info,
                                           struct nvme_fc_port_template* tmpl,
                                           struct device* dev,
                                           struct nvme_fc_local_port** lport )
{
    return nvme_fc_register_localport( info, tmpl, dev, lport );
}

static int gew_nvme_fc_unregister_localport( struct nvme_fc_local_port* lport )
{
    return nvme_fc_unregister_localport( lport );
}

static int gew_nvme_fc_register_remoteport( struct nvme_fc_local_port *lport,
                                            struct nvme_fc_port_info *info,
                                            struct nvme_fc_remote_port **rport )
{
    return nvme_fc_register_remoteport( lport, info, rport );
}

static int gew_nvme_fc_unregister_remoteport( struct nvme_fc_remote_port *rport )
{
    return nvme_fc_unregister_remoteport( rport );
}

static int gew_nvme_fc_set_remoteport_devloss( struct nvme_fc_remote_port* remoteport,
                                               u32 dev_loss_tmo )
{
    return nvme_fc_set_remoteport_devloss( remoteport, dev_loss_tmo );
}

#endif  /* FEAT_NVME_INI */


#if defined (FEAT_NVME_TGT)

static int gew_nvmet_fc_register_targetport( struct nvmet_fc_port_info* info,
                                             struct nvmet_fc_target_template* tmpl,
                                             struct device* dev,
                                             struct nvmet_fc_target_port** port )
{
    return nvmet_fc_register_targetport( info, tmpl, dev, port );
}

static int gew_nvmet_fc_unregister_targetport( struct nvmet_fc_target_port* port )
{
    return nvmet_fc_unregister_targetport( port );
}

static int gew_nvmet_fc_rcv_ls_req( struct nvmet_fc_target_port* port,
                                    void* hosthandle,
                                    struct nvmefc_ls_rsp* req,
                                    void* reqbuf,
                                    u32 buflen )
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0) || IS_SLES_RELEASE_OR_LATER(15,2,0) || IS_RHEL_RELEASE_OR_LATER(8,4)
    return nvmet_fc_rcv_ls_req( port, hosthandle, req, reqbuf, buflen );
#else
    return nvmet_fc_rcv_ls_req( port, req, reqbuf, buflen );
#endif
}

static int gew_nvmet_fc_rcv_fcp_req( struct nvmet_fc_target_port* port,
                                     struct nvmefc_tgt_fcp_req* req,
                                     void* buf,
                                     u32 buflen )
{
    return nvmet_fc_rcv_fcp_req( port, req, buf, buflen );
}


static void gew_nvmet_fc_rcv_fcp_abort( struct nvmet_fc_target_port *port,
                                        struct nvmefc_tgt_fcp_req *req )
{
    nvmet_fc_rcv_fcp_abort( port, req );
}

#endif  /* FEAT_NVME_TGT */


static const struct cpumask* gew_get_cpu_mask( unsigned int cpu )
{
    return get_cpu_mask( cpu );
}


static int gew_irq_set_affinity_hint( unsigned int irq, const struct cpumask* mask )
{
    return irq_set_affinity_hint( irq, mask );
}


static int gew_irq_set_affinity_notifier( unsigned int irq,
                                          struct irq_affinity_notify* notify )
{
    return irq_set_affinity_notifier( irq, notify );
}


static void gew_bsg_job_done( struct bsg_job *job, int result,
                              unsigned int reply_payload_rcv_len )
{
    bsg_job_done( job, result, reply_payload_rcv_len );
}


static struct workqueue_struct* gew_alloc_workqueue( const char* name,
                                                     unsigned int flags,
                                                     int max_active )
{
    return alloc_workqueue( name, flags, max_active );
}


static void gew_destroy_workqueue( struct workqueue_struct* wq )
{
    destroy_workqueue( wq );
}


static struct device* gew_device_create( struct class* cls, struct device* parent,
                                         dev_t devt, void* drvdata, const char* name )
{
    return device_create( cls, parent, devt, drvdata, name );
}


static void gew_device_unregister( struct device* dev )
{
    device_unregister( dev );
}


static struct class* gew_class_create( struct module* owner, const char* name,
                                       struct lock_class_key* key )
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6,4,0) || IS_RHEL_RELEASE_OR_LATER(9,4)
    return class_create( name );
#else
    return __class_create( owner, name, key );
#endif
}


static void gew_class_destroy( struct class *cls )
{
    class_destroy( cls );
}


static int gew_add_uevent_var( struct kobj_uevent_env* env, const char* str )
{
    return add_uevent_var( env, "%s", str );
}

static int gew_pci_enable_sriov( struct pci_dev* dev, int nr_virtfn )
{
    return pci_enable_sriov( dev, nr_virtfn );
}

static void gew_pci_disable_sriov( struct pci_dev* dev )
{
    pci_disable_sriov( dev );
}

/**
 * Allow the HBA driver to signal that an adapter has been discovered, and
 * provide service routines to access it.
 *
 * @hba:   The service routines to access the HBA.  The caller must fill out the
 *         full structure.
 * @shost: The Scsi_Host object for the adapter.
 * @cfg_host: atto_cfg host filled in by the atto_cfg.
 *
 * Returns: 0 on success, negative errno on failure.
 */
static int attocfg_register_host( struct attohba_ops *hba,
                                  struct kobject *shost,
                                  struct attocfg_host **cfg_host )
{
    struct attocfg_host *host;

    host = kzalloc( sizeof(*host), GFP_KERNEL );

    if( !host )
        return -ENOMEM;

    INIT_LIST_HEAD( &host->list );
    host->shost = shost;
    host->hba_ops = *hba;

    *cfg_host = host;

    return 0;
}


/**
 * Allow the HBA driver to signal that the indicated host is going away.  No
 * further calls will be made using the attohba_ops provided when the given host
 * was created.
 *
 * @host: The host created by this driver when the HBA driver registered the
 *        adapter.
 *
 * Returns: 0 on success, negative errno on failure.
 */
static int attocfg_unregister_host( struct attocfg_host *host )
{
    kfree( host );

    return 0;
}


/**
 * Allow any global initialization to be performed when the HBA driver first
 * loads.
 *
 * Returns: 0 on success, negative errno on failure.
 */
int ATTO_CFG_SYM(connect)( struct atgi_connect_ops* ops )
{
    if( ops->version != ATGI_CONN_VERSION )
    {
        atto_log_err( ATTO_LOG_INIT,
                      "Detected incompatibility with HBA driver.  (expected %d)\n",
                      ATGI_CONN_VERSION );
        return -EINVAL;
    }

    ops->register_host = attocfg_register_host;
    ops->unregister_host = attocfg_unregister_host;
    ops->init_async_event_sockets = init_async_event_sockets;
    ops->fire_atto_event = fire_atto_event;
#ifdef FEAT_BAD_BLK_EVT
    ops->fire_bb_event = fire_bb_event;
#endif
    ops->bsg_job_done = gew_bsg_job_done;
#ifdef FEAT_NVME_INI
    ops->nvme_fc_register_localport     = gew_nvme_fc_register_localport;
    ops->nvme_fc_unregister_localport   = gew_nvme_fc_unregister_localport;
    ops->nvme_fc_register_remoteport    = gew_nvme_fc_register_remoteport;
    ops->nvme_fc_unregister_remoteport  = gew_nvme_fc_unregister_remoteport;
    ops->nvme_fc_set_remoteport_devloss = gew_nvme_fc_set_remoteport_devloss;
#endif
#ifdef FEAT_NVME_TGT
    ops->nvmet_fc_register_targetport   = gew_nvmet_fc_register_targetport;
    ops->nvmet_fc_unregister_targetport = gew_nvmet_fc_unregister_targetport;
    ops->nvmet_fc_rcv_ls_req            = gew_nvmet_fc_rcv_ls_req;
    ops->nvmet_fc_rcv_fcp_req           = gew_nvmet_fc_rcv_fcp_req;
    ops->nvmet_fc_rcv_fcp_abort         = gew_nvmet_fc_rcv_fcp_abort;
#endif
    ops->get_cpu_mask               = gew_get_cpu_mask;
    ops->irq_set_affinity_hint      = gew_irq_set_affinity_hint;
    ops->irq_set_affinity_notifier  = gew_irq_set_affinity_notifier;
    ops->alloc_workqueue_w          = gew_alloc_workqueue;
    ops->destroy_workqueue          = gew_destroy_workqueue;
    ops->device_create              = gew_device_create;
    ops->device_unregister          = gew_device_unregister;
    ops->class_create_w             = gew_class_create;
    ops->class_destroy              = gew_class_destroy;
    ops->add_uevent_var             = gew_add_uevent_var;
    ops->pci_enable_sriov           = gew_pci_enable_sriov;
    ops->pci_disable_sriov          = gew_pci_disable_sriov;

    return 0;
}
EXPORT_SYMBOL(ATTO_CFG_SYM(connect));


/**
 * Allow any global cleanup to be performed when the HBA driver module unloads.
 *
 * Returns: 0 on success, negative errno on failure.
 */

int ATTO_CFG_SYM(disconnect)( void )
{
    return 0;
}
EXPORT_SYMBOL(ATTO_CFG_SYM(disconnect));


/**
 * The module initialization entry point.
 *
 * Returns: 0 on success, negative errno on failure.
 */
static int __init ATTO_CFG_SYM(init)( void )
{
    atto_log_info( ATTO_LOG_INIT, "module init called (%s)\n", AV_VERSION_STR );

    return 0;
}

/**
 * The module teardown entry point.
 */
static void __exit ATTO_CFG_SYM(exit)( void )
{
    atto_log_info( ATTO_LOG_INIT, "module exit called (%s)\n", AV_VERSION_STR );
}


/* e.g.  module_init(attocfg_init) */
DECLARE_MOD_ENTRY(init);
DECLARE_MOD_ENTRY(exit);
