/****************************************************************************

DESCRIPTION: API for communicating between ATTO drivers

Copyright (C) 2006-2018  ATTO TECHNOLOGY, INC.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*****************************************************************************/

#if !defined(ATTO_CFG2_H)
#define ATTO_CFG2_H

#ifndef ATTO_DRVR_RAWNAME
#error  "Define ATTO_DRVR_RAWNAME (without quotes) before including this file"
#endif


#include <linux/kobject.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/pci.h>
#include "atto_distro.h"


/*  INTERNAL MACROS (don't use directly)  */

/* Build family specific symbol names.  Do not use these first two directly! */
#define ATTO_CONCAT(prefix, name)       prefix ## _ ## name
#define ATTO_EVAL_NAME(prefix, name)    ATTO_CONCAT(prefix, name)

/* This double indirection is needed so that ATTO_DRVR_SYM is fully unrolled
 * by the time it is passed to the module_xxx() macro.
 */
#define DECLARE_MOD_ENTRY2(name, decl)  module_ ## name( decl )

/*  END INTERNAL MACROS  */


/* e.g. "celerity16fc" */
#if !defined(ATTO_DRVR_NAME)
/* This is also defined in atsi.h, which is included in the core. */
#define ATTO_DRVR_NAME  __stringify( ATTO_DRVR_RAWNAME )
#endif

/* e.g. ATTO_DRVR_SYM(init) -> celerity16fc_init */
#define ATTO_DRVR_SYM(name)  ATTO_EVAL_NAME(ATTO_DRVR_RAWNAME, name)

/* e.g. ATTO_CFG_SYM(connect) -> celerity16fc_cfg_connect */
#define ATTO_CFG_SYM(name)   ATTO_EVAL_NAME(ATTO_CFG_MOD_RAWNAME, name)

/* e.g. DECLARE_MOD_ENTRY(init) -> module_init(celerity16fc_init) */
#define DECLARE_MOD_ENTRY(name)   DECLARE_MOD_ENTRY2(name, ATTO_DRVR_SYM(name))

typedef ssize_t (*hba_ops_handler) (struct kobject *kobj,
                                    char           *buf,
                                    loff_t          off,
                                    size_t          count,
                                    int             operation);

typedef int (*attocfg_fire_async_event) (void* data, int len);


struct attocfg_host;
struct attohba_host;

struct device;

struct nvme_fc_port_info;
struct nvme_fc_port_template;
struct nvme_fc_local_port;
struct nvme_fc_remote_port;

struct nvmet_fc_port_info;
struct nvmet_fc_target_template;
struct nvmet_fc_target_port;

struct nvmefc_tgt_fcp_req;
struct nvmefc_ls_rsp;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,8,0) && !IS_SLES_RELEASE_OR_LATER(15,2,0) && !IS_RHEL_RELEASE_OR_LATER(8,4)
#define nvmefc_ls_rsp nvmefc_tgt_ls_req
struct nvmefc_tgt_ls_req;
#endif

struct irq_affinity_notify;
struct bsg_job;
struct workqueue_struct;


/**
 * Services provided by the HBA driver, called by the attocfg driver.
 *
 * @attohba:  The cookie provided by the HBA driver, to be passed in each
 *            operation.
 */

struct attohba_ops {
    struct attohba_host *attohba;
};


/**
 * Module-level services exchanged between attocfg and HBA driver.  Connection
 * is initiated by the HBA driver during module init.  Mismatched version
 * results in failure to load.
 *
 * @version:  The structure version, initialized by the HBA driver.  Increased
 *            each time the API changes.
 */

struct atgi_connect_ops {
    int version;
        #define ATGI_CONN_VERSION    7

    /* HBA registration */

    int (*register_host)( struct attohba_ops* hba,
                          struct kobject* shost,
                          struct attocfg_host** cfg_host );

    int (*unregister_host)( struct attocfg_host* host );

    /* Wrappers */

    int (*nvme_fc_register_localport)( struct nvme_fc_port_info* info,
                                       struct nvme_fc_port_template* tmpl,
                                       struct device* dev,
                                       struct nvme_fc_local_port** lport );

    int (*nvme_fc_unregister_localport)( struct nvme_fc_local_port* lport );

    int (*nvme_fc_register_remoteport)( struct nvme_fc_local_port* lport,
                                        struct nvme_fc_port_info* info,
                                        struct nvme_fc_remote_port** rport );

    int (*nvme_fc_unregister_remoteport)( struct nvme_fc_remote_port* rport );

    int (*nvme_fc_set_remoteport_devloss)( struct nvme_fc_remote_port* remoteport,
                                           u32 dev_loss_tmo );

    int (*nvmet_fc_register_targetport)( struct nvmet_fc_port_info* info,
                                         struct nvmet_fc_target_template* tmpl,
                                         struct device* dev,
                                         struct nvmet_fc_target_port** port );

    int (*nvmet_fc_unregister_targetport)( struct nvmet_fc_target_port* port );

    int (*nvmet_fc_rcv_ls_req)( struct nvmet_fc_target_port* port,
                                void* hosthandle,
                                struct nvmefc_ls_rsp* req,
                                void* reqbuf,
                                u32 buflen );

    int (*nvmet_fc_rcv_fcp_req)( struct nvmet_fc_target_port* port,
                                 struct nvmefc_tgt_fcp_req* req,
                                 void* buf,
                                 u32 buflen );

    void (*nvmet_fc_rcv_fcp_abort)( struct nvmet_fc_target_port *port,
                                    struct nvmefc_tgt_fcp_req *req );

    const struct cpumask* (*get_cpu_mask)( unsigned int cpu );

    int (*irq_set_affinity_hint)( unsigned int irq, const struct cpumask* mask );

    int (*irq_set_affinity_notifier)( unsigned int irq,
                                      struct irq_affinity_notify *notify );

    void (*bsg_job_done)( struct bsg_job *job,
                          int result, unsigned int reply_payload_rcv_len );

    int (*init_async_event_sockets)( void );

    int (*fire_atto_event)( void *data, int len );

    int (*fire_bb_event)( void *data, int len );

    struct workqueue_struct* (*alloc_workqueue_w)( const char* name,
                                                   unsigned int flags,
                                                   int max_active );

    void (*destroy_workqueue)( struct workqueue_struct* wq );

    struct device* (*device_create)( struct class* cls, struct device* parent,
                                     dev_t devt, void* drvdata, const char* name );

    void (*device_unregister)( struct device* dev );

    struct class* (*class_create_w)( struct module* owner, const char* name,
                                     struct lock_class_key* key );

    void (*class_destroy)( struct class *cls );

    int (*add_uevent_var)( struct kobj_uevent_env *env, const char* str );

    int (*pci_enable_sriov)( struct pci_dev* dev, int nr_virtfn );

    void (*pci_disable_sriov)( struct pci_dev* dev );
};


int ATTO_CFG_SYM(connect)( struct atgi_connect_ops* ops );
int ATTO_CFG_SYM(disconnect)( void );


#endif  /* ATTO_CFG2_H */
